/*****************************************************************************/
/*
                                 yahMAIL.c

Yet Another Hyper-MAIL :^)  With acknowlegements for seminal ideas, other
VMS incarnations, etc., etc., including Jonathan Boswell (jsb@early.com). 

WASD yahMAIL is for interfacing with OpenVMS MAIL files.  It is NOT ONLY for
providing access to public/archived mail messages.  It can provide on-line,
authenticated, Web access to VMS users' private mail (with much of the
functionality available with DECwindows mail).  It can provide "postmaster"
(superuser) access to a VMS systems mail system, including mail profiles.  It
can also make public areas/archives of mail messages available for
non-athenticated access via the Web.

YAHMAIL makes some limited use of JavaScript and client-side cookies.  The
JavaScript ensures there is only ever one folder browse page in the history
list as well as keeping some state information allowing some semi-intelligent
reloading of browse pages as necessary (i.e. after deleting, move or sending
message(s)).  The pages will work OK on a non-JavaScript-enabled browser but
these features will not work and the user must keep track of when to reload
folder pages, etc.

It is possible to localize the language used in this script.  See LANG_EN.C and
any other language-specific files that may be in the source directory.


ISSUES
------
There have been a number of issues with the Mail system from 5.5-2 through to
the present (that I know of).  First, please ensure you have the latest MAIL
and C RTL patches installed if behaviour problems exist.


PATHS TO MAIL AND ARCHIVES
--------------------------
The request path is not necesarily fixed and can have various components
present or not, and dependent on whether it's a request for private Mail or
public archives.  A tilde should be the first character of a private request
and cause the server to authenticate the request.  The query string can vary
considerably but always contains some of those elements described below in the
"CGI Variables" section.

  /~/                          open NEWMAIL folder
  /~/folder/                   open folder
  /~/file/folder/              open folder found in file
  /~/999                       get item from NEWMAIL folder
  /~/folder/999                get item from folder
  /~/file/folder/999           get item from folder found in file
  /~~                          cancel current authentication
  
  /name/                       public mail file
  /name/999                    get item from NEWMAIL in public file
  /name/folder/                open folder found in public file
  /name/folder/999             get item from folder in public file


CONFIG FILE
-----------
The configuration file (amongst other functions) maps authenicated user names
to VMS usernames and archive paths to archive Mail files. It can be protected
against world access.  SYSPRV is used when opening it. Several free-form text
sections indicated by square-bracket delimited keywords.  The left-most square
bracket must be on the left-most margin of the line.  Do not place any other
than a section keyword left square bracket on this margin (indent it at least
one space if necessary).  The section keywords are as follows:

  [anchor]           integer, zero disables conversion of URLs into anchors
  [body]             a localized HTML <BODY...> tag
  [charset]          default character set (overrides any RFC822 header)
  [footer]           free form, HTML marked-up text for the bottom of the page
  [header]           free form, HTML marked-up text for the top of the page
  [list]             names of "global" address lists with file specifications
  [newmail]          attention seeking message for new mail check
  [postmaster]       a localized warning message when postmaster access active
  [private]          mappings for access to the VMS mail system (authenticated)
  [public]           mappings for archive access requests (non-authenticated)
  [publicfooter]     for a public page (archive) overrides [header]
  [publicheader]     for a public page (archive) overrides [footer]
  [smtp]             transport for smtp mail (for pre-VMS6.2 Mail)
  [window]           number of items in browse window (an integer)
  [!yahmail!]        a non-default public config message string (see below)
  [#]                comment directive

Sections may be present in any order, EXCEPT the [public] and [private]
sections MUST be the last two.  The text for each section is considered to span
from the right square bracket of the section keyword to the left square bracket
of the next section keyword.  Leading and trailing white-space is trimmed.

The "[private]" section.  Each single line in this section comprises an entry
in the following format:

  authenticated-name\realm\equivalent-VMS-username

The authenticated name is the string in WWW_REMOTE_USER.  The realm is the
string in the WASD WWW_AUTH_REALM.  The eqivalent VMS user name is the VMS user
that YAHMAIL allows mail access to.

Any of these three elements may be asterisks.  If the authenticated name is an
asterisk, any name will be mapped. If the realm is an asterisk, names from
any (and all) realms will be mapped.  If the equivalent VMS username is an
asterisk then the authenticated name is mapped as the VMS user account.

The "[public]" section.  Each line in this section comprises an entry in the
following format:

  public-name\VMS-file-path\optional-folder-name

The first element of the path is the first '/' delimited string of the path
(following the script portion).  The VMS file path is a full file specification
to the mail file (e.g. USER_DISK:[USERS.EXAMPLE]MAIL.MAI).  The optional folder
name allows only a single folder to be accesses (otherwise it's the entire mail
file).

Example YAHMAIL configuration file.

  [#] this is an example yahMAIL configuration file
  [body]
  <BODY BGCOLOR="#ffffff" TEXT="#000000" LINK="#ff0000" VLINK="#990000">
  [header]
  <FONT COLOR="#ff0000">
  Unauthorized access via <B>yahMAIL</B> is prohibited
  by Federal and State law.
  </FONT>
  [footer]
  <A HREF="/yahmail/-/help.html" TARGET=_parent>yahMAIL help</A>
  [#] this is a pre-VMS6.2 system, provide default Internet mail transport
  [smtp] MX%
  [list]
  all_users WORLD:[LISTS]ALL_USERS.DIS
  postmaster WORLD:[LISTS]POSTMASTER.DIS
  [private]
  daniel\VMS\postmaster
  *\VMS\*
  [public]
  info-WASD\WEB:[INFO-WASD]MAIL.MAI\


CAUTION!!
---------
CARE SHOULD BE EXERCISED WITH THE CONFIGURATION FILE.  Misconfiguration could
conceivably leave the mail system open to inappropriate access.  For instance,
mapping an authenticated name to the wrong VMS username.  The code has been
designed to minimise the risk of catastrophic errors, for instance mapping all
authenticated users as POSTMASTERS.  Also, unless a username is explicitly
mapped in some way it will be rejected.  But, DO NOT MAP all authenticated
users to VMS usernames unless the source of the authentication is the SYSUAF,
or unless the realm usernames always correspond the appropriate VMS username. 
That is, DO NOT DO SOMETHING LIKE:

  *\not-from-SYSUAF\*

For the same reasons do not map all realms with a wildcard unless you have only
the one authentication realm, from the SYSUAF, or all realm names are
authenticated from the SYSUAF, or all usernames in the realm correspond to the
appropriate VMS usernames.  That is, DO NOT DO:

  *\*\*


PUBLIC CONFIG MESSAGE
---------------------
When requesting from a public area, if the folder contains a message with the
subject string "!yahMAIL!" (case sensitive), or another specified using the
[!yahmail!] configuration directive, it's contents are used to generate custom
titles, etc., on the public folder browse page.  Section keywords:

  [body]             a localized HTML <BODY...> tag
  [title]            title of the folder
  [header]           free form, HTML marked-up text for the top of the page
  [footer]           free form, HTML marked-up text for the bottom of the page

These keywords must be positioned on the left-most edge of the page (the '['
character should not occur here otherwise). In-between these keywords is
free-form text, which can contain HTML markup tags.  Each section of text, up
until the next '[' delimited keyword or end-of-message, is used for the purpose
specified by the preceding keyword.  Note that [header] and [footer] here
overrides any [header], [footer], [publicheader] or [publicfooter] in the
configuration file.  This is an example:

  # not going to customize the <BODY...> tag for this public area!
  [title]
  Example Archive
  [header]
  This is an <U>example archive</U>!
  Note that HTML <I>markup</I> can be used.
  This is just a descriptive paragraph.
  [footer]
  For more information see the
  <A HREF="/yahmail/-/help.html">yahMAIL Help</A>.


CGI VARIABLES
-------------
AUTH_REALM          authentication realm name
PATH_INFO           request path information
REMOTE_USER         authenticated user name
REQUEST_METHOD      GET or POST
REQUEST_SCHEME      "http:" or "https:"
SCRIPT_NAME         name of script component in path information
SERVER_NAME         host name of server
SERVER_PORT         request port
SERVER_SOFTWARE     software ID string for server

FORM_ACT            form action string (usually from a submit button)
FORM_ANC            allows the user to override the 'MakeAnchors' setting
FORM_ANCHOR         same as FORM_ANC (for user convenience)
FORM_BWC            private browse window configuration, message create button
FORM_BWF            private browse window configuration, folder select/action
FORM_BWM            private browse window configuration, standard move buttons
FORM_BWP            private browse window configuration, pick fields
FORM_BWU            private browse window configuration, user move buttons
FORM_CMC            current message count (at time folder was listed)
FORM_CNM            refresh (check) NEWMAIL page every so many minutes
FORM_NEWMAIL        same as FORM_CNM (for user convenience)
FORM_FSE            folder selected from a <SELECT></SELECT>
FORM_FTE            folder name entered in text ENTRY box
FORM_HEA            when displaying mail include VMS header
FORM_IDn            field names 1..n form each message ID checkbox checked
FORM_INC            include the message text with a create reply or forward
FORM_LSE            address list name from a <SELECT></SELECT>
FORM_LTE            address list name entered in text ENTRY box
FORM_MCT            checkbox selected "application/octet-stream"
FORM_MUN            when MUNPACKing MIME content provides the part (file) name
FORM_NMS            new message seconds (used during NEWMAIL checking)
FORM_NNN            simple number based on time, to make folder request unique
FORM_PAG            browse folder page number (1 == most recent, nnn == least)
FORM_PCC            pick CC (select on substring of "CC:")
FORM_PFR            pick from (select on substring of "From:")
FORM_PSU            pick subject (select on substring of "Subj:")
FORM_PTO            pick to (select on substring of "To:")
FORM_QUO            when reply/forward message quote ('>') body
FORM_RFC            when displaying mail include the RFC822 header
FORM_SBD            sending mail, message body
FORM_SFR            sending mail, postmaster (only) may specify who it's from!
FORM_STO            sending mail, message to
FORM_SCC            sending mail, message CC
FORM_SCS            sending mail, CC to self
FORM_SPS            show private browse page setup selectors and buttons
FORM_SSU            sending mail, message subject
FORM_TAW            <TEXTAREA> to automatically wrap at integer
FORM_TXT            text from <TEXTAREA> when editing list or sig file
FORM_UAP            setting user profile, auto-purge
FORM_UCP            setting user profile, CC prompt
FORM_UCF            setting user profile, copy forward
FORM_UCR            setting user profile, copy reply
FORM_UCS            setting user profile, copy send
FORM_UED            setting user profile, editor
FORM_UFW            setting user profile, forwarding
FORM_UNM            setting user profile, new messages
FORM_UPN            setting user profile, personal name
FORM_USF            setting user profile, signature file
FORM_UPQ            setting user profile, print queue
FORM_UPF            setting user profile, print queue form
FORM_WHA            differentiates multiple buttons with the same action
FORM_WIN            browse window (number of messages on one page)
FORM_WINDOW         same as FORM_WIN (for user convenience)
FORM_WRA            mail composition edit wraps at integer (also see _TAW)
FORM_WRAP           same as FORM_WRA (for user convenience)


PRIVILEGES REQUIRED
-------------------
SYSPRV              to access and send mail on behalf of others
READALL             (optional) to access mail protected against SYSTEM

if installed as follows:

  $ INSTALL ADD HT_EXE:YAHMAIL /OPEN /HEADER /SHARED /PRIV=SYSPRV

or

  $ INSTALL ADD HT_EXE:YAHMAIL /OPEN /HEADER /SHARED /PRIV=(SYSPRV,READALL)

then should have an ACL attached to prevent unexpected access:

  $ SET SECURITY HT_EXE:YAHMAIL.EXE -
  /ACL=((IDENT=HTTP$SERVER,ACCESS=READ+EXEC),(IDENT=*,ACCESS=NONE))


QUALIFIERS
----------
/CHARSET=           "Content-Type: text/html; charset=..."
/CONFIG=            yahMAIL configuration file (overrides YAHMAIL$CONFIG)
/DBUG               turns on all "if (Debug)" statements
/POSTMASTER         PostMaster functionality can be used with this script


LOGICAL NAMES
-------------
YAHMAIL$CONFIG      file containing configuration directives
YAHMAIL$CHECK       turns on access checking messages in ProcessConfig()
YAHMAIL$DISABLE     disables yahMAIL, provides an HTML message to that effect
YAHMAIL$DBUG        turns on all "if (Debug)" statements
YAHMAIL$PARAM       equivalent to (overrides) the command line
                    parameters/qualifiers (define as a system-wide logical)
                    may also be assigned as a local/global symbol


BUILD DETAILS
-------------
See BUILD_YAHMAIL.COM procedure.


COPYRIGHT
---------
Copyright (c) 1999 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under the conditions of the GNU GENERAL PUBLIC LICENSE, version 2.


VERSION HISTORY (update 'SoftwareID' as well)
---------------
22-DEC-1999  MGD  v1.1.0, private browse setup,
                          standard Mail and "user" folder buttons,
                          copy-to-self from profile now sets checkbox
22-NOV-1999  MGD  v1.0.2, bugfix; revise signature file from 6.2 to 7.0
07-NOV-1999  MGD  v1.0.1, bugfix; MakeAnchors only if isdigit()
31-OCT-1999  MGD  v1.0.0, final non-BETA release
26-SEP-1999  MGD  v1.0.B4, NEWMAIL "biff" (and bugfix),
                           optional auto-wrap on mail-create <TEXTAREA>s,
                           additions to LANG_EN.C
16-SEP-1999  MGD  v1.0.B3, add lang_Response...,
                           [PublicHeader] and [PublicFooter] configuration,
                           replace 302 redirect for MIME with
                           "Content-Disposition: filename="
21-AUG-1999  MGD  v1.0.B2, separate language module,
                           signature file update
22-JUN-1999  MGD  v1.0.B1, initial development
*/

/*****************************************************************************/

#ifdef __ALPHA
   char SoftwareID [] = "YAHMAIL AXP-1.1.0"
#else
   char SoftwareID [] = "YAHMAIL VAX-1.1.0"
#endif
#  ifdef YAHMAIL_MUNPACK
      " (MIME)"
#  endif
;

#ifdef __ALPHA
#   pragma nomember_alignment
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <stat.h>

/* VMS related header files */
#include <descrip.h>
#include <fscndef.h>
#include <libdef.h>
#include <libdtdef.h>
#include <libclidef.h>
#include <maildef.h>
#include <mailmsgdef.h>
#include <rms.h>
#include <prvdef.h>
#include <ssdef.h>
#include <stsdef.h>
#include <syidef.h>

/* mainly external declarations for messages */
#include "lang.h"

/* make up for the lack of these in some header files */
#ifndef MAIL$_USER_SIGFILE
#  define MAIL$_SEND_SIGFILE 23
#  define MAIL$_SEND_NO_SIGFILE 24
#  define MAIL$_USER_SET_SIGFILE 3110
#  define MAIL$_USER_SET_NO_SIGFILE 3111
#  define MAIL$_USER_SIGFILE 3141
#endif

#define VMSok(x) ((x) & STS$M_SUCCESS)
#define VMSnok(x) !(((x) & STS$M_SUCCESS))

#define boolean int
#define true 1
#define false 0

#define FI_LI __FILE__, __LINE__

/* make false to exclude authenticated (private) functionality from image */
#ifndef YAHMAIL_PRIVATE
#  define YAHMAIL_PRIVATE 1
#endif

/* make false to exclude POSTMASTER functionality from image */
#ifndef YAHMAIL_POSTMASTER
#  if YAHMAIL_PRIVATE
#     define YAHMAIL_POSTMASTER 1
#  else
#     define YAHMAIL_POSTMASTER 0
#  endif
#endif

/* size of "chunks" of memory allocated when reading message text */
#define MESSAGE_TEXT_CHUNK 4096

/* size of "chunks" of memory allocated when reading configuration text */
#define CONFIG_TEXT_CHUNK 1024

/* number of messages to show per folder page */
#define DEFAULT_BROWSE_WINDOW 10

/* string used in "Subj:" field to indicate it's the folder config message */
#define DEFAULT_PUBLIC_CONFIG_SUBJ "!yahMAIL!"

#define DEFAULT_BODY_TAG \
"BGCOLOR=\"#ffffff\" TEXT=\"#000000\" LINK=\"#0000ff\" VLINK=\"#000099\""

#define ADDRESS_LIST_WINDOW \
"menubar=1,toolbar=0,location=0,directories=0,width=550,height=500,resizable=1"

/* make HTML anchors out of URLs and/or mail addresses in plain text */
#define DEFAULT_ANCHORS (CGILIB_ANCHOR_WEB | CGILIB_ANCHOR_MAIL)

#define DEFAULT_CONFIG_FILE "YAHMAIL$CONFIG"

struct VmsItemStruct {
   short int  buf_len;
   short int  item;
   unsigned char   *buf_addr;
   unsigned short  *ret_len;
};

struct VmsItemStruct  NullItem = {0,0,0,0};

char  Utility [] = "YAHMAIL";

boolean  CliPostMasterEnabled,
         ConfigPrivate,
         ConfigPublic,
         ConfigPublicBrowseFolders,
         Debug,
         DoAddrListAccess,
         DoAddrListRemove,
         DoAddrListSelect,
         DoAddrListUpdate,
         DoFolderBrowse,
         DoMessageCopy,
         DoMessageCreate,
         DoMessageDelete,
         DoMessageEmptyWasteBasket,
         DoMessageForward,
         DoMessageMove,
         DoMessageRead,
         DoMessageReadMunpack,
         DoMessageReadOctetStream,
         DoMessageReadPlainText,
         DoMessageReply,
         DoMessageSend,
         DoSigFileAccess,
         DoSigFileUpdate,
         DoUserCreate,
         DoUserDelete,
         DoUserList,
         DoUserProfile,
         DoUserSet,
         IsCgiPlus,
         OsuEnvironment,
         PostMasterAuthenticated,
         PrivateSetupHide,
         PrivateSetupShow,
         StdCgiEnvironment;

int  BrowsePage,
     BrowseWindow,
     CgiPlusUsageCount,
     CheckNewMailMinutes,
     CurrentMessageCount,
     MakeAnchors,
     MessageIncrement,
     PublicConfigSubjCount,
     PublicConfigSubjLength,
     VmsVersion,
     WrapAt;

unsigned long  MailFileContext,
               UserContext;

unsigned short  NumTime [7];

unsigned long  BinTime [2];

char  TildeChar;

char  *CliConfigFilePtr = DEFAULT_CONFIG_FILE,
      *CgiAuthRealmPtr,
      *CgiEnvironmentPtr,
      *CgiFormActionPtr,
      *CgiFormFolderNamePtr,
      *CgiFormIncludePtr,
      *CgiFormMunpackContentTypePtr,
      *CgiFormMunpackPtr,
      *CgiFormNnnPtr,
      *CgiFormQuotePtr,
      *CgiFormRefreshPtr,
      *CgiHttpCookiePtr,
      *CgiHttpHostPtr,
      *CgiHttpUserAgentPtr,
      *CgiRemoteUserPtr,
      *CgiRequestSchemePtr,
      *CgiServerNamePtr,
      *CgiServerPortPtr,
      *CgiServerSoftwarePtr,
      *CgiPathInfoPtr,
      *CgiQueryStringPtr,
      *CgiRemoteUserPtr,
      *CgiRequestMethodPtr,
      *CgiScriptNamePtr,
      *CgiTypePtr,
      *CliCharsetPtr = lang_Charset,
      *ConfigAddrListPtr,
      *ConfigBodyTagPtr,
      *ConfigCharsetPtr,
      *ConfigFooterPtr,
      *ConfigHeaderPtr,
      *ConfigNewMailPtr,
      *ConfigPublicFooterPtr,
      *ConfigPublicHeaderPtr,
      *ConfigSmtpTransportPtr,
      *PostMasterWarningPtr,
      *PublicConfigSubjPtr,
      *PublicHeaderPtr,
      *PublicFooterPtr,
      *PublicTitlePtr;

char  ActionPathInfo [256],
      ActionQueryString [512],
      SchemeHostPort [256];

char  PathPart [4][64];

/* global variables, information to VMS mail */

unsigned long  MailFileContext,
               MailFileContext,
               MailFileMessagesDeleted,
               MailFileNameLength,
               MailFolderNameLength,
               MailSubstringCcLength,
               MailSubstringFromLength,
               MailSubstringSubjLength,
               MailSubstringToLength,
               MailUserNameLength,
               CopyFolderNameLength;

/* size of array for keeping message IDs supplied as checkbox selections */
#define MAX_MESSAGE_ID 64
int  MailMessageIdCount;
unsigned long  MailMessageId [MAX_MESSAGE_ID+1];

char  CopyFolderName [256],
      HtmlEscapedMailFolderName [256],
      MailFileName [256],
      MailFolderName [256],
      MailSubstringCc [256],
      MailSubstringFrom [256],
      MailSubstringSubj [256],
      MailSubstringTo [256],
      MailUserName [256],
      PublicFileName [256],
      PublicFolderName [256],
      PublicUserName [256];

/* global variables, information from VMS mail */

int  VmsMailHeaderLength,
     VmsMailRfc822HeaderLength,
     VmsMailRfc822MimeVersion;

unsigned long  VmsMailTextLength,
               VmsMailMessageId,
               VmsMailMessageSize,
               VmsMailMessageSelectedCount,
               VmsMailSigFileTextLength,
               VmsMailTextSize,
               VmsMailUserAutoPurge,
               VmsMailUserCcPrompt,
               VmsMailUserCopyForward,
               VmsMailUserCopyReply,
               VmsMailUserCopySend;

unsigned long  VmsMailBinaryDate [2];

unsigned short  VmsMailToLength,
                VmsMailCcLength,
                VmsMailDateLength,
                VmsMailExtIdLength,
                VmsMailFromLength,
                VmsMailFolderLength,
                VmsMailMessageFlags,
                VmsMailSenderLength,
                VmsMailSubjectLength,
                VmsMailUserEditorLength,
                VmsMailUserFormLength,
                VmsMailUserForwardingLength,
                VmsMailUserFullDirectoryLength,
                VmsMailUserNewMessages,
                VmsMailUserPersonalNameLength,
                VmsMailUserQueueLength,
                VmsMailUserSigFileLength;

char  *VmsMailTextPtr,
      *VmsMailSigFileTextPtr;

char  VmsMailTo [256],
      VmsMailCc [256],
      VmsMailDate [256],
      VmsMailExtId [256],
      VmsMailFrom [256],
      VmsMailFolder [256],
      VmsMailRfc822MimeCharset [256],
      VmsMailRfc822MimeContTransEnc [256],
      VmsMailRfc822MimeContentType [256],
      VmsMailSender [256],
      VmsMailSubject [256],
      VmsMailUserEditor [256],
      VmsMailUserForm [256],
      VmsMailUserFullDirectory [256],
      VmsMailUserForwarding [256],
      VmsMailUserPersonalName [256],
      VmsMailUserQueue [256],
      VmsMailUserSigFile [256];

struct FolderListStruct
{
   struct FolderListStruct  *NextPtr;
   unsigned long  MessageCount;
   char  *FolderNamePtr;
   char  Data [];
};

int  FolderListCount;
struct FolderListStruct  *FolderListHeadPtr,
                         *FolderListTailPtr;

struct MessageListStruct
{
   struct MessageListStruct  *PrevPtr,
                             *NextPtr;
   unsigned long  MessageId;
   unsigned long  BinaryDate [2];
   char  *DatePtr,
         *ToPtr,
         *FromPtr,
         *SubjectPtr;
   char  Data [];
};

int  MessageListCount;
struct MessageListStruct  *MessageListHeadPtr,
                          *MessageListTailPtr;

struct UserListStruct
{
   struct UserListStruct  *NextPtr;
   char  *UserNamePtr;
   char  Data [];
};

int  UserListCount;
struct UserListStruct  *UserListHeadPtr,
                       *UserListTailPtr;

/* required prototypes */
char* CreateMailMessageReFw (char*, char*);
int GenerateFolderListEntry ();
char* VmsEmailMessage (char*, char*, char*, char*, char*, char*, int);

/*****************************************************************************/
/*
Include the CGILIB.H/.C function code from the standard WASD location if
CGILIB_INCLUDE_H/_C has not been defined.  If it has been defined then the
value must be the #include location.
*/

/* indicate required functionality */
#define CGILIB_CGIVAR_SUPPORT   1
#define CGILIB_CGIPLUS_SUPPORT  1
#define CGILIB_OSU_SUPPORT      1
#define CGILIB_GENERAL_SUPPORT  1
#define CGILIB_POST_SUPPORT     1
#define CGILIB_RESPONSE_SUPPORT 1

#define CGILIB_DEBUG (Debug)

#define CGILIB_RESPONSE_BODY          ConfigBodyTagPtr
#define CGILIB_RESPONSE_CHARSET       ConfigCharsetPtr
#define CGILIB_RESPONSE_ERROR_TITLE   lang_ResponseErrorTitle
#define CGILIB_RESPONSE_REPORTED      lang_ResponseReportedByServer
#define CGILIB_RESPONSE_SOFTWAREID    SoftwareID
#define CGILIB_RESPONSE_SUCCESS_TITLE lang_ResponseSuccessTitle

#ifndef CGILIB_INCLUDE_H
#   include "/ht_root/src/misc/cgilib.h"
#   include "/ht_root/src/misc/cgilib.c"
#else
#   include CGILIB_INCLUDE_H
#   include CGILIB_INCLUDE_C
#endif

/*****************************************************************************/
/*
*/

main (int argc, char* argv[])
       
{
   int  status;

   /*********/
   /* begin */
   /*********/

   GetParameters ();

   if (getenv ("YAHMAIL$DBUG") != NULL) Debug = true;

   CgiLibEnvironmentInit (argc, argv, false);
   IsCgiPlus = CgiLibEnvironmentIsCgiPlus ();
   CgiEnvironmentPtr = CgiLibEnvironmentName ();

   VmsVersion = GetVmsVersion ();

#ifdef YAHMAIL_MUNPACK
   if (IsCgiPlus)
   {
      /*
         I suspect the current MUNPACK used for decoding MIME messages
         wouldn't be designed for reuse and therefore shouldn't be!
      */
      CgiLibResponseError (FI_LI, 0, lang_ErrCgiPlusMunpack);
      exit (SS$_NORMAL);
   }
#endif /* YAHMAIL_MUNPACK */

   if (IsCgiPlus)
   {
      for (;;)
      {
         /* block waiting for the next request */
         CgiLibVar ("");
         ProcessRequest ();
         CgiLibCgiPlusEOF ();
      }
   }
   else
      ProcessRequest ();
}

/*****************************************************************************/
/*
Get "command-line" parameters, whether from the command-line or from a
configuration symbol or logical containing the equivalent.
*/

GetParameters ()

{
   static char  CommandLine [256];
   static unsigned long  Flags = 0;

   register char  *aptr, *cptr, *clptr, *sptr;

   int  status;
   unsigned short  Length;
   char  ch;
   $DESCRIPTOR (CommandLineDsc, CommandLine);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "GetParameters()\n");

   if ((clptr = getenv ("YAHMAIL$PARAM")) == NULL)
   {
      /* get the entire command line following the verb */
      if (VMSnok (status =
          lib$get_foreign (&CommandLineDsc, 0, &Length, &Flags)))
         exit (status);
      (clptr = CommandLine)[Length] = '\0';
   }

   aptr = NULL;
   ch = *clptr;
   for (;;)
   {
      if (aptr != NULL && *aptr == '/') *aptr = '\0';
      if (!ch || ch == '!') break;

      *clptr = ch;
      if (Debug) fprintf (stdout, "clptr |%s|\n", clptr);
      while (*clptr && isspace(*clptr)) *clptr++ = '\0';
      aptr = clptr;
      if (*clptr == '/') clptr++;
      while (*clptr && !isspace (*clptr) && *clptr != '/')
      {
         if (*clptr != '\"')
         {
            clptr++;
            continue;
         }
         cptr = clptr;
         clptr++;
         while (*clptr)
         {
            if (*clptr == '\"')
               if (*(clptr+1) == '\"')
                  clptr++;
               else
                  break;
            *cptr++ = *clptr++;
         }
         *cptr = '\0';
         if (*clptr) clptr++;
      }
      ch = *clptr;
      if (*clptr) *clptr = '\0';
      if (Debug) fprintf (stdout, "aptr |%s|\n", aptr);
      if (!*aptr) continue;

      /***********/
      /* process */
      /***********/

      if (strsame (aptr, "/CONFIG=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (!*cptr) continue;
         CliConfigFilePtr = cptr+1;
         continue;
      }

      if (strsame (aptr, "/CHARSET=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (!*cptr) continue;
         CliCharsetPtr = cptr+1;
         continue;
      }

      if (strsame (aptr, "/DBUG", -1))
      {
         Debug = true;
         continue;
      }

      if (strsame (aptr, "/POSTMASTER", 5))
      {
#if YAHMAIL_POSTMASTER
         CliPostMasterEnabled = true;
         continue;
#else
         fprintf (stdout, "%%%s-E-BUILD, build does not support POSTMASTER\n",
                  Utility);
         exit (STS$K_ERROR | STS$M_INHIB_MSG);
#endif
      }

      if (*aptr == '/')
      {
         fprintf (stdout, "%%%s-E-IVQUAL, unrecognized qualifier\n \\%s\\\n",
                  Utility, aptr+1);
         exit (STS$K_ERROR | STS$M_INHIB_MSG);
      }

      fprintf (stdout, "%%%s-E-MAXPARM, too many parameters\n \\%s\\\n",
               Utility, aptr);
      exit (STS$K_ERROR | STS$M_INHIB_MSG);
   }
}

/*****************************************************************************/
/*
Read the contents of the configuration file into memory.  The process the text
from the file.  Keyword delimited elements are parse from the file.  These are
null-terminated and used by being pointed at so the file's contents cannot be
disposed of before reprocessing a new request (for CGIplus).  The [private]
and/or [public] directives MUST follow any others.
*/

ProcessConfig ()

{
   static char  *ConfigTextPtr;

   register char  *bptr, *cptr, *sptr, *tptr;

   boolean  AuthenticatedNameWildcard,
            ConfigCheck;
   int  cnt,
        status,
        ConfigTextLength,
        ConfigTextSize,
        LineLength;
   char  Line [256];
   FILE  *ConfigFile;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProcessConfig()\n");

   if (getenv("YAHMAIL$CHECK") != NULL)
   {
      CgiLibResponseHeader (200, "text/html");
      fputs (lang_CurrentlyUnavailable, stdout);
      fputs ("<!--\n", stdout);
      ConfigCheck = true;
   }
   else
   if (getenv("YAHMAIL$DISABLE") != NULL)
   {
      CgiLibResponseHeader (200, "text/html");
      fputs (lang_CurrentlyUnavailable, stdout);
      exit (SS$_NORMAL);
   }
   else
      ConfigCheck = false;

   BrowseWindow = DEFAULT_BROWSE_WINDOW;
   ConfigBodyTagPtr = DEFAULT_BODY_TAG;
   MakeAnchors = DEFAULT_ANCHORS;
   PublicConfigSubjPtr = DEFAULT_PUBLIC_CONFIG_SUBJ;
   PublicConfigSubjLength = sizeof(DEFAULT_PUBLIC_CONFIG_SUBJ)-1;
   ConfigHeaderPtr = ConfigFooterPtr = ConfigAddrListPtr =
      PostMasterWarningPtr = "";
   ConfigCharsetPtr = ConfigNewMailPtr = ConfigSmtpTransportPtr =
      ConfigPublicHeaderPtr = ConfigPublicFooterPtr = NULL;
   HtmlEscapedMailFolderName[0] =
      PublicFileName[0] = PublicFolderName[0] = PublicUserName[0] =
      MailFileName[0] = MailFolderName[0] = MailUserName[0] = '\0';
   ConfigPrivate = ConfigPublic = ConfigPublicBrowseFolders =
      PostMasterAuthenticated = false;

   /*************************/
   /* read file into memory */
   /*************************/

   if (ConfigTextPtr != NULL) free (ConfigTextPtr);
   ConfigTextLength = 0;
   ConfigTextPtr = calloc (1, CONFIG_TEXT_CHUNK);
   ConfigTextSize = CONFIG_TEXT_CHUNK;

   EnableAccess (1);

   if ((ConfigFile = fopen (CliConfigFilePtr, "r", "shr=get")) == NULL)
   {
      status = vaxc$errno;
      if (Debug) fprintf (stdout, "fopen() %%X%08.08X\n", status);
      if (ConfigCheck)
      {
         fprintf (stdout, "error opening %s: %%X%08.08X\n-->\n",
                  CliConfigFilePtr, status);
         exit (SS$_NORMAL);
      }
      return;
   }

   EnableAccess (0);

   while (fgets (Line, sizeof(Line), ConfigFile) != NULL)
   {
      /** if (Debug) fprintf (stdout, "|%s|\n", Line); **/
      LineLength = strlen (Line);
      if (ConfigTextLength+LineLength > ConfigTextSize)
      {
          ConfigTextSize += CONFIG_TEXT_CHUNK;
          ConfigTextPtr = realloc (ConfigTextPtr, ConfigTextSize);
      }
      /* copy the terminating null as well! */
      memcpy (ConfigTextPtr+ConfigTextLength, Line, LineLength+1);
      ConfigTextLength += LineLength;
   }

   fclose (ConfigFile);

   /********************/
   /* process contents */
   /********************/

   cptr = ConfigTextPtr;
   while (*cptr)
   {
      if (strsame (cptr, "[public]", 8) ||
          strsame (cptr, "[private]", 9))
         break;

      if (strsame (cptr, bptr = "[anchor]", cnt = 8) ||
          strsame (cptr, bptr = "[body]", cnt = 6) ||
          strsame (cptr, bptr = "[charset]", cnt = 9) ||
          strsame (cptr, bptr = "[footer]", cnt = 8) ||
          strsame (cptr, bptr = "[header]", cnt = 8) ||
          strsame (cptr, bptr = "[list]", cnt = 6) ||
          strsame (cptr, bptr = "[newmail]", cnt = 9) ||
          strsame (cptr, bptr = "[postmaster]", cnt = 12) ||
          strsame (cptr, bptr = "[publicfooter]", cnt = 14) ||
          strsame (cptr, bptr = "[publicheader]", cnt = 14) ||
          strsame (cptr, bptr = "[smtp]", cnt = 6) ||
          strsame (cptr, bptr = "[window]", cnt = 8) ||
          strsame (cptr, bptr = "[!yahmail!]", cnt = 11))
      {    
         /* find the start of the next directive or end of text */
         sptr = (cptr += cnt);
         while (*cptr && *(unsigned short*)cptr != '\n[') cptr++;
         if (*cptr) *cptr++ = '\0';

         /* trim leading then trailing white space */
         while (*sptr && isspace(*sptr)) sptr++;
         tptr = sptr;
         while (*sptr) sptr++;
         if (sptr > tptr) sptr--;
         while (sptr > tptr && isspace(*sptr)) sptr--;
         if (sptr > tptr) *++sptr = '\0';         

         /* set the appropriate pointer */
         if (strsame (bptr, "[anchor]", 8))
            MakeAnchors = atoi(tptr);
         else
         if (strsame (bptr, "[body]", 6))
            ConfigBodyTagPtr = tptr;
         else
         if (strsame (bptr, "[charset]", 9))
            ConfigCharsetPtr = tptr;
         else
         if (strsame (bptr, "[footer]", 8))
            ConfigFooterPtr = tptr;
         else
         if (strsame (bptr, "[header]", 8))
            ConfigHeaderPtr = tptr;
         else
         if (strsame (bptr, "[list]", 6))
            ConfigAddrListPtr = tptr;
         else
         if (strsame (bptr, "[newmail]", 9))
            ConfigNewMailPtr = tptr;
         else
         if (strsame (bptr, "[postmaster]", 12))
            PostMasterWarningPtr = tptr;
         else
         if (strsame (bptr, "[publicfooter]", 14))
            ConfigPublicFooterPtr = tptr;
         else
         if (strsame (bptr, "[publicheader]", 14))
            ConfigPublicHeaderPtr = tptr;
         else
         if (strsame (bptr, "[smtp]", 6))
            ConfigSmtpTransportPtr = tptr;
         else
         if (strsame (bptr, "[window]", 8))
            BrowseWindow = atoi(tptr);
         else
         if (strsame (bptr, "[!yahmail!]", 11))
         {
            PublicConfigSubjPtr = tptr;
            PublicConfigSubjLength = strlen(tptr);
         }
      }
      else
      {
         /* find the start of the next line */
         while (*cptr && *cptr != '\n') cptr++;
         if (*cptr) cptr++;
      }
   }

   if (Debug)
      fprintf (stdout, "|%s|\n|%s|\n|%s|\n|%s|\n|%s|\n|%s|\n|%s|\n|%s|\n",
               ConfigBodyTagPtr, ConfigHeaderPtr, ConfigFooterPtr,
               ConfigPublicHeaderPtr, ConfigPublicFooterPtr,
               ConfigAddrListPtr, PostMasterWarningPtr, PublicConfigSubjPtr);

   /******************************/
   /* encountered access control */
   /******************************/

   if (CgiRemoteUserPtr[0])
   {
      /**********************/
      /* authenticated user */
      /**********************/

#if !YAHMAIL_PRIVATE
      fprintf (stdout, "%%%s-E-BUILD, build does not support [private]\n",
               Utility);
      exit (STS$K_ERROR | STS$M_INHIB_MSG);
#endif

      /* find the appropriate section */
      while (!strsame (cptr, "[private]", 9))
      {
         /** if (Debug) fprintf (stdout, "cptr |%s|\n", cptr); **/
         while (*cptr && *(unsigned short*)cptr != '\n[') cptr++;
         if (!*cptr) break;
         cptr++;
      }
      /** if (Debug) fprintf (stdout, "cptr |%s|\n", cptr); **/

      if (ConfigCheck)
         fprintf (stdout, "[private] %s%s%s\n",
                  CgiRemoteUserPtr,
                  CgiAuthRealmPtr[0] ? "\\" : "\\?", CgiAuthRealmPtr);

      /* authenticated-name\realm\equivalent-VMS-username */

      while (*cptr)
      {
         /* find start of next line */
         while (*cptr && *cptr != '\n') cptr++;
         if (!*cptr) break;
         cptr++;
         /** if (Debug) fprintf (stdout, "|%s|\n", cptr); **/

         if (ConfigCheck)
         {
            for (sptr = cptr; *sptr && *sptr != '\n'; sptr++);
            if (*cptr != '\n')
               fprintf (stdout, "|%*.*s|\n", sptr-cptr, sptr-cptr, cptr);
         }

         while (*cptr && isspace(*cptr)) cptr++;
         /* if comment line */
         if (*cptr == '#' || *cptr == '!') continue;
         /* if end of section break */
         if (*cptr == '[') break;

         /* authenticated name */
         sptr = cptr;
         while (*cptr && !isspace(*cptr) && *cptr != '\\') cptr++;
         if (*cptr) *cptr++ = '\0';
         if (*sptr != '*' && !strsame (sptr, CgiRemoteUserPtr, -1))
         {
            if (ConfigCheck) fprintf (stdout, "reject: authenticated name\n");
            continue;
         }
         if (*sptr == '*')
            AuthenticatedNameWildcard = true;
         else
            AuthenticatedNameWildcard = false;

         /* authentication realm */
         sptr = cptr;
         while (*cptr && !isspace(*cptr) && *cptr != '\\') cptr++;
         if (*cptr) *cptr++ = '\0';
         if (*sptr != '*' && !strsame (sptr, CgiAuthRealmPtr, -1))
         {
            if (ConfigCheck) fprintf (stdout, "reject: authentication realm\n");
            continue;
         }

         /* get the resultant VMS user name */
         if (*cptr == '*')
         {
            strzcpy (MailUserName, CgiRemoteUserPtr, sizeof(MailUserName));
            for (sptr = MailUserName; *sptr; sptr++) *sptr = toupper(*sptr);
         }
         else
         {
            sptr = MailUserName;
            while (*cptr && !isspace(*cptr)) *sptr++ = toupper(*cptr++);
            *sptr = '\0';
         }
         if (Debug) fprintf (stdout, "MailUserName |%s|\n", MailUserName);
         if (!isalnum(MailUserName[0]))
         {
            if (ConfigCheck) fprintf (stdout, "reject: VMS username\n");
            continue;
         }

#if YAHMAIL_POSTMASTER
         if (strsame (MailUserName, "POSTMASTER", -1))
         {
            /* shouldn't/mistake map all users to be POSTMASTERs! */
            if (AuthenticatedNameWildcard)
            {
               if (ConfigCheck)
                  fprintf (stdout, "reject: wildcard POSTMASTER\n");
               continue;
            }

            if (PostMasterAuthenticated = CliPostMasterEnabled)
            {
               strzcpy (MailUserName, CgiRemoteUserPtr, sizeof(MailUserName));
               for (sptr = MailUserName; *sptr; sptr++) *sptr = toupper(*sptr);
               if (Debug) fprintf (stdout, "|%s|\n", MailUserName);
               PostMasterWarningPtr = lang_PostMasterWarning;
            }
            else
            {
               if (ConfigCheck)
                  fprintf (stdout, "reject: POSTMASTER not enabled\n");
               continue;
            }
         }
#endif /* YAHMAIL_POSTMASTER */

         /* access is OK */
         ConfigPrivate = true;
         if (ConfigCheck)
         {
            fprintf (stdout, "allow: %s%s\n-->\n",
                MailUserName, PostMasterAuthenticated ? " as POSTMASTER" : "");
            exit (SS$_NORMAL);
         }

         return;
      }
   }
   else
   {
      /**********/
      /* public */
      /**********/

      /* find the appropriate section */
      while (!strsame (cptr, "[public]", 8))
      {
         /** if (Debug) fprintf (stdout, "cptr |%s|\n", cptr); **/
         while (*cptr && *(unsigned short*)cptr != '\n[') cptr++;
         if (!*cptr) break;
         cptr++;
      }
      /** if (Debug) fprintf (stdout, "cptr |%s|\n", cptr); **/

      if (ConfigCheck) fprintf (stdout, "[public] %s\n", PathPart[0]);

      /* first-element-of-path\VMS-file-path\optional-folder-name */

      if (Debug) fprintf (stdout, "PathPart[0] |%s|\n", PathPart[0]);

      while (*cptr)
      {
         /* find start of next line */
         while (*cptr && *cptr != '\n') cptr++;
         if (!*cptr) break;
         cptr++;
         /** if (Debug) fprintf (stdout, "|%s|\n", cptr); **/

         if (ConfigCheck)
         {
            for (sptr = cptr; *sptr && *sptr != '\n'; sptr++);
            if (*cptr != '\n')
               fprintf (stdout, "|%*.*s|\n", sptr-cptr, sptr-cptr, cptr);
         }

         while (*cptr && isspace(*cptr)) cptr++;
         /* if comment line */
         if (*cptr == '#' || *cptr == '!') continue;
         /* if end of section break */
         if (*cptr == '[') break;

         /* first element of path */
         sptr = cptr;
         while (*cptr && !isspace(*cptr) && *cptr != '\\') cptr++;
         if (*cptr) *cptr++ = '\0';
         if (*sptr != '*' && !strsame (sptr, PathPart[0], -1))
         {
            if (ConfigCheck) fprintf (stdout, "reject: path\n");
            continue;
         }
         strzcpy (PublicUserName, PathPart[0], sizeof(PublicUserName));

         /* mail file name */
         sptr = PublicFileName;
         while (*cptr && !isspace(*cptr) && *cptr != '\\') *sptr++ = *cptr++;
         *sptr = '\0';
         if (*cptr) *cptr++ = '\0';
         if (Debug) fprintf (stdout, "PublicFileName |%s|\n", PublicFileName);
         /* a configuration file name is mandatory */
         if (!PublicFileName[0])
         {
            if (ConfigCheck) fprintf (stdout, "reject: MAIL file name\n");
            continue;
         }

         /* restrict folder to this specified name */
         sptr = PublicFolderName;
         while (*cptr && !isspace(*cptr) && *cptr != '\\') *sptr++ = *cptr++;
         *sptr = '\0';
         if (*cptr) *cptr++ = '\0';
         if (Debug)
            fprintf (stdout, "PublicFolderName |%s|\n", PublicFolderName);

         /* if no specific folder is configured then allowed to browse */
         if (PublicFolderName[0] == '*') PublicFolderName[0] = '\0';
         if (!PublicFolderName[0]) ConfigPublicBrowseFolders = true;

         /* access is OK */
         ConfigPublic = true;

         if (ConfigCheck)
         {
            fprintf (stdout, "allow: %s %s\n-->\n",
                     PublicFileName, PublicFolderName);
            exit (SS$_NORMAL);
         }

         return;
      }
   }

   if (ConfigCheck)
   {
      fprintf (stdout, "end of configuration!\n-->\n");
      exit (SS$_NORMAL);
   }
}

/*****************************************************************************/
/*
*/

ProcessRequest ()
       
{
   register char  *cptr, *sptr, *zptr;

   int  idx,
        status,
        PostBufferCount;
   char  *PostBufferPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProcessRequest()\n");

   DoAddrListAccess = DoAddrListRemove = DoAddrListSelect = DoAddrListUpdate =
      DoMessageEmptyWasteBasket = DoFolderBrowse = DoMessageDelete =
      DoMessageMove = DoMessageSend = DoMessageRead = DoMessageReadOctetStream =
      DoMessageReadPlainText = DoMessageForward = DoMessageReply =
      DoMessageCopy = DoMessageCreate = DoSigFileAccess = DoSigFileUpdate =
      DoUserCreate = DoUserList = DoUserProfile = DoUserSet = 
      PrivateSetupHide = PrivateSetupShow = false;

   BrowsePage = CheckNewMailMinutes = MessageIncrement =
      VmsMailSigFileTextLength = 0;

   if (VmsMailTextPtr != NULL) free (VmsMailTextPtr);
   if (VmsMailSigFileTextPtr != NULL) free (VmsMailSigFileTextPtr);
   VmsMailTextPtr = VmsMailSigFileTextPtr = NULL;

   CurrentMessageCount = -1;

   cptr = CgiPathInfoPtr = CgiLibVar ("WWW_PATH_INFO");

   /* parse the first (up to) four slash-delimited sections of the path */
   for (idx = 0; idx < 4; idx++)
   {
      if (*cptr) cptr++;
      zptr = (sptr = PathPart[idx]) + sizeof(PathPart[idx]);
      while (*cptr && *cptr != '/' && sptr < zptr) *sptr++ = *cptr++;
      if (sptr >= zptr) exit (SS$_RESULTOVF);
      *sptr = '\0';
   }
   if (Debug)
      fprintf (stdout, "PathPart |%s|%s|%s|%s|\n",
               PathPart[0], PathPart[1], PathPart[2], PathPart[3]);

   /* accomodate local mapping requirements for the tilde character */
   if (PathPart[0][0] && !isalnum(PathPart[0][0]))
      TildeChar = PathPart[0][0];
   else
      TildeChar = '~';

   if (PathPart[0][1] == TildeChar)
   {
      /*
         break current authentication
         (allows new username/password to be supplied)
      */
      CgiLibResponseHeader (401, "text/html",
                            "Basic realm=\"yahMAIL cancel\"");
      fputs ("Current authentication cancelled!\n", stdout);
      return;
   }

   CgiRemoteUserPtr = CgiLibVar("WWW_REMOTE_USER");
   CgiAuthRealmPtr = CgiLibVar("WWW_AUTH_REALM");

   ProcessConfig ();

   if (!(ConfigPrivate || ConfigPublic))
   {
      /***********************/
      /* access is forbidden */
      /***********************/

      CgiLibResponseError (FI_LI, 0, lang_ErrConfigRestriction);
      return;
   }

   /*********************/
   /* access is allowed */
   /*********************/

   if (ConfigCharsetPtr == NULL) ConfigCharsetPtr = CliCharsetPtr;

   CgiHttpUserAgentPtr = CgiLibVar ("WWW_HTTP_USER_AGENT");
   CgiScriptNamePtr = CgiLibVar ("WWW_SCRIPT_NAME");
   CgiRequestMethodPtr = CgiLibVar ("WWW_REQUEST_METHOD");

   /********************/
   /* path information */
   /********************/

   /*
       Legitimate path information elements:
       /~/                          (open NEWMAIL folder)
       /~/folder/                   (open folder)
       /~/file/folder/              (open folder found in file)
       /~/999                       (get item from NEWMAIL folder)
       /~/folder/999                (get item from folder)
       /~/file/folder/999           (get item from folder found in file)
       /name/                       (public mail file)
       /name/999                    (get item from NEWMAIL in public file)
       /name/folder/                (open folder found in public file)
       /name/folder/999             (get item from folder in public file)
   */

   if (PathPart[0][0])
   {
      if (ConfigPrivate)
      {
         if (PathPart[0][0] != TildeChar)
         {
            CgiLibResponseError (FI_LI, 0, lang_ErrConfigRestriction);
            return;
         }
         /* only the postmaster can specify who they want to be! */
         if (!PostMasterAuthenticated &&
             PathPart[0][1] &&
             !strsame (MailUserName, PathPart[0]+1, -1))
         {
            CgiLibResponseError (FI_LI, 0, lang_ErrConfigRestriction);
            return;
         }
         if (PostMasterAuthenticated && PathPart[0][1])
         {
            sptr = MailUserName;
            for (cptr = PathPart[0]+1; *cptr; *sptr++ = toupper(*cptr++));
            *sptr = '\0';
         }
      }
      else
      if (PathPart[0][0] == TildeChar)
      {
         CgiLibResponseError (FI_LI, 0, lang_ErrConfigRestriction);
         return;
      }
   }

   if (isinteger(PathPart[1]))
   {
      /* message ID follows username/archive (implies it's from NEWMAIL) */
      MailMessageId[0] = atoi(PathPart[1]);
   }
   else
   if (PathPart[1][0] && (!PathPart[2][0] || isinteger(PathPart[2])))
   {
      /* folder name follows username/archive */
      strzcpy (MailFolderName, PathPart[1], sizeof(MailFolderName));
      /* if a message ID follows the folder name */
      if (isinteger(PathPart[2])) MailMessageId[0] = atoi(PathPart[2]);
   }
   else
   if (PathPart[1][0] && PathPart[2][0] && !isinteger(PathPart[2]))
   {
      strzcpy (MailFileName, PathPart[1], sizeof(MailFileName));
      strzcpy (MailFolderName, PathPart[2], sizeof(MailFolderName));
      if (isinteger(PathPart[3])) MailMessageId[0] = atoi(PathPart[3]);
   }

   if (!strcmp (CgiRequestMethodPtr, "POST"))
   {
      /***************/
      /* POST method */
      /***************/

      /* read and create the CGI variable equivalents of the body fields */
      CgiLibReadRequestBody (&PostBufferPtr, &PostBufferCount);
      if (PostBufferPtr == NULL ||
          PostBufferCount == 0)
      {
         CgiLibResponseError (FI_LI, 0, lang_ErrSanityCheck);
         return;
      }
      CgiLibFormRequestBody (PostBufferPtr, PostBufferCount);
      free (PostBufferPtr);
      if (Debug)
         while ((cptr = CgiLibVarNull ("*")) != NULL)
            fprintf (stdout, "|%s|\n", cptr);
   }
   else
   {
      /**************/
      /* GET method */
      /**************/

      /* if not GET method */
      if (strcmp (CgiRequestMethodPtr, "GET"))
      {
         CgiLibResponseHeader (501, "text/html");
         return;
      }
   }

   /****************/
   /* query string */
   /****************/

   /* need this here in case it's a [next] or [previous] */
   cptr = CgiLibVar ("WWW_FORM_CMC");
   if (isdigit(*cptr)) CurrentMessageCount = atoi(cptr);
   if (Debug) fprintf (stdout, "CMC: %d\n", CurrentMessageCount);
   CgiFormActionPtr = CgiLibVar ("WWW_KEY_1");
   if (!CgiFormActionPtr[0])
      CgiFormActionPtr = CgiLibVar ("WWW_FORM_ACT");

   CgiFormNnnPtr = CgiLibVar ("WWW_FORM_NNN");

   cptr = CgiLibVar ("WWW_FORM_WINDOW");
   if (!cptr[0])
      cptr = CgiLibVar ("WWW_FORM_WIN");
   if (isdigit(*cptr)) BrowseWindow = atoi(cptr);

   cptr = CgiLibVar ("WWW_FORM_PAG");
   if (isdigit(*cptr)) BrowsePage = atoi(cptr);
   if (!BrowsePage) BrowsePage = 1;

   cptr = CgiLibVar ("WWW_FORM_NEWMAIL");
   if (!cptr[0])
      cptr = CgiLibVar ("WWW_FORM_CNM");
   if (isdigit(*cptr)) CheckNewMailMinutes = atoi(cptr);

   cptr = CgiLibVar ("WWW_FORM_ANCHOR");
   if (!cptr[0])
      cptr = CgiLibVar ("WWW_FORM_ANC");
   if (isdigit(*cptr)) MakeAnchors = atoi(cptr);

   cptr = CgiLibVar ("WWW_FORM_WRAP");
   if (!cptr[0])
      cptr = CgiLibVar ("WWW_FORM_WRA");
   if (!cptr[0])
      WrapAt = 72;
   else
      WrapAt = atoi(cptr);
   /* let's keep it within the bounds of reason! */
   if (WrapAt && (WrapAt < 48 || WrapAt > 255)) WrapAt = 72;

   cptr = CgiLibVar ("WWW_FORM_PCC");
   MailSubstringCcLength =
      strzcpy (MailSubstringCc, cptr, sizeof(MailSubstringCc));

   cptr = CgiLibVar ("WWW_FORM_PFR");
   MailSubstringFromLength =
      strzcpy (MailSubstringFrom, cptr, sizeof(MailSubstringFrom));

   cptr = CgiLibVar ("WWW_FORM_PSU");
   MailSubstringSubjLength =
      strzcpy (MailSubstringSubj, cptr, sizeof(MailSubstringSubj));

   cptr = CgiLibVar ("WWW_FORM_PTO");
   MailSubstringToLength =
      strzcpy (MailSubstringTo, cptr, sizeof(MailSubstringTo));

#ifdef YAHMAIL_MUNPACK

   CgiFormMunpackPtr = CgiLibVar ("WWW_FORM_MUN");
   if (CgiFormMunpackPtr[0])
   {
      /* unpack a MIME attachment */
      DoMessageRead = DoMessageReadMunpack = true;
   }
   else

#endif /* YAHMAIL_MUNPACK */

   if (CgiFormActionPtr[0])
   {
      /*****************/
      /* action button */
      /*****************/

      /* first check the text entry field, then if none the selection box */
      CgiFormFolderNamePtr = CgiLibVar ("WWW_FORM_FTE");
      if (!CgiFormFolderNamePtr[0])
         CgiFormFolderNamePtr = CgiLibVar ("WWW_FORM_FSE");

      if (strsame (CgiFormActionPtr, lang_BtnOpen, -1) ||
          (PrivateSetupShow =
           strsame (CgiFormActionPtr, lang_BtnSetupShow, -1)) ||
          (PrivateSetupHide =
           strsame (CgiFormActionPtr, lang_BtnSetupHide, -1)))
      {
         DoFolderBrowse = true;
         /* only authenticated request are allowed to specify the folder */
         if (ConfigPrivate)
         {
            if (!CgiFormFolderNamePtr[0])
            {
               CgiLibResponseError (FI_LI, 0, lang_ErrSanityCheck);
               return;
            }
            MailFolderNameLength = strzcpy (MailFolderName,
                                            CgiFormFolderNamePtr,
                                            sizeof(MailFolderName));
         }
      }
      else
      if (strsame (CgiFormActionPtr, lang_BtnToNewMail, -1))
      {
         ProcessRequestIdNumbers ();
         if (MailMessageIdCount)
         {
            CopyFolderNameLength = strzcpy (CopyFolderName, "NEWMAIL",
                                            sizeof(CopyFolderName));
            DoMessageMove = true;
         }
         else
         {
            MailFolderNameLength = strzcpy (MailFolderName, "NEWMAIL",
                                            sizeof(MailFolderName));
            DoFolderBrowse = true;
         }
      }
      else
      if (strsame (CgiFormActionPtr, lang_BtnToMail, -1))
      {
         ProcessRequestIdNumbers ();
         if (MailMessageIdCount)
         {
            CopyFolderNameLength = strzcpy (CopyFolderName, "MAIL",
                                            sizeof(CopyFolderName));
            DoMessageMove = true;
         }
         else
         {
            MailFolderNameLength = strzcpy (MailFolderName, "MAIL",
                                            sizeof(MailFolderName));
            DoFolderBrowse = true;
         }
      }
      else
      if (strsame (CgiFormActionPtr, lang_BtnToSent, -1))
      {
         ProcessRequestIdNumbers ();
         if (MailMessageIdCount)
         {
            CopyFolderNameLength = strzcpy (CopyFolderName, "SENT",
                                            sizeof(CopyFolderName));
            DoMessageMove = true;
         }
         else
         {
            MailFolderNameLength = strzcpy (MailFolderName, "SENT",
                                            sizeof(MailFolderName));
            DoFolderBrowse = true;
         }
      }
      else
      if (strsame (CgiFormActionPtr, lang_BtnToWasteBasket, -1))
      {
         ProcessRequestIdNumbers ();
         if (MailMessageIdCount)
         {
            CopyFolderNameLength = strzcpy (CopyFolderName, "WASTEBASKET",
                                            sizeof(CopyFolderName));
            DoMessageMove = true;
         }
         else
         {
            MailFolderNameLength = strzcpy (MailFolderName, "WASTEBASKET",
                                            sizeof(MailFolderName));
            DoFolderBrowse = true;
         }
      }
      else
      if (strsame (CgiFormActionPtr, lang_BtnEmptyWasteBasket, -1))
         DoMessageEmptyWasteBasket = true;
      else
      if (CgiFormActionPtr[0] == '\"')
      {
         ProcessRequestIdNumbers ();
         if (MailMessageIdCount)
         {
            zptr = (sptr = CopyFolderName) + sizeof(CopyFolderName)-1;
            for (cptr = CgiFormActionPtr+1;
                 *cptr && *cptr != '\"' && sptr < zptr;
                 *sptr++ = *cptr++);
            *sptr = '\0';
            CopyFolderNameLength = sptr - CopyFolderName;
            DoMessageMove = true;
         }
         else
         {
            zptr = (sptr = MailFolderName) + sizeof(MailFolderName)-1;
            for (cptr = CgiFormActionPtr+1;
                 *cptr && *cptr != '\"' && sptr < zptr;
                 *sptr++ = *cptr++);
            *sptr = '\0';
            MailFolderNameLength = sptr - MailFolderName;
            DoFolderBrowse = true;
         }
      }
      else
      if ((DoMessageCopy = strsame (CgiFormActionPtr, lang_BtnCopy, -1)) ||
          (DoMessageMove = strsame (CgiFormActionPtr, lang_BtnMove, -1)))
      {
         ProcessRequestIdNumbers ();
         if (!MailMessageIdCount)
         {
            CgiLibResponseError (FI_LI, 0, lang_ErrMessageNotSelected);
            return;
         }
         if (!CgiFormFolderNamePtr[0])
         {
            CgiLibResponseError (FI_LI, 0, lang_ErrSanityCheck);
            return;
         }
         CopyFolderNameLength = strzcpy (CopyFolderName,
                                         CgiFormFolderNamePtr,
                                         sizeof(CopyFolderName));
      }
      else
      if (strsame (CgiFormActionPtr, lang_BtnDelete, -1))
      {
         ProcessRequestIdNumbers ();
         if (!MailMessageIdCount)
         {
            CgiLibResponseError (FI_LI, 0, lang_ErrMessageNotSelected);
            return;
         }
         DoMessageDelete = true;
      }
      else
      if (strsame (CgiFormActionPtr, lang_BtnNext, -1))
      {
         if (MailMessageId[0] >= CurrentMessageCount)
         {
            CgiLibResponseError (FI_LI, 0, lang_ErrSanityCheck);
            return;
         }
         MailMessageId[0] += MessageIncrement = 1;
         DoMessageRead = true;
      }
      else
      if (strsame (CgiFormActionPtr, lang_BtnPrevious, -1))
      {
         if (MailMessageId[0] <= 1)
         {
            CgiLibResponseError (FI_LI, 0, lang_ErrSanityCheck);
            return;
         }
         MailMessageId[0] += MessageIncrement = -1;
         DoMessageRead = true;
      }
      else
      if (strsame (CgiFormActionPtr, lang_BtnText, -1))
         DoMessageRead = DoMessageReadPlainText = true;
      else
      if (strsame (CgiFormActionPtr, lang_BtnBinary, -1) ||
          strsame (CgiFormActionPtr, lang_BtnForeign, -1))
         DoMessageRead = DoMessageReadOctetStream = true;
      else
      if (strsame (CgiFormActionPtr, lang_BtnCreateSend, -1))
         DoMessageCreate = true;
      else
      if (strsame (CgiFormActionPtr, lang_BtnReply, -1))
         DoMessageReply = true;
      else
      if (strsame (CgiFormActionPtr, lang_BtnForward, -1))
         DoMessageForward = true;
      else
      if (strsame (CgiFormActionPtr, lang_BtnSend, -1))
         DoMessageSend = true;
      else
      if (strsame (CgiFormActionPtr, lang_BtnSet, -1))
         DoUserSet = true;
      else
      if (strsame (CgiFormActionPtr, lang_ProfileUser, -1))
         DoUserProfile = true;
      else
      if (strsame (CgiFormActionPtr, lang_BtnChangeProfile, -1))
         DoUserProfile = true;
      else
      if (strsame (CgiFormActionPtr, lang_BtnDeleteUserProfile, -1))
         DoUserDelete = true;
      else
      if (strsame (CgiFormActionPtr, lang_BtnCreateUserProfile, -1))
         DoUserCreate = true;
      else
      if (strsame (CgiFormActionPtr, lang_BtnAddressLists, -1))
         DoAddrListSelect = true;
      else
      if (strsame (CgiFormActionPtr, lang_BtnAccess, -1))
      {
         cptr = CgiLibVar ("WWW_FORM_WHA");
         if (!strcmp (cptr, "lis"))
            DoAddrListAccess = true;
         else
         if (!strcmp (cptr, "sig"))
            DoSigFileAccess = true;
         else
         {
            CgiLibResponseError (FI_LI, 0, lang_ErrSanityCheck);
            return;
         }
      }
      else
      if (strsame (CgiFormActionPtr, lang_BtnUpdate, -1))
      {
         cptr = CgiLibVar ("WWW_FORM_WHA");
         if (!strcmp (cptr, "lis"))
            DoAddrListUpdate = true;
         else
         if (!strcmp (cptr, "sig"))
            DoSigFileUpdate = true;
         else
         {
            CgiLibResponseError (FI_LI, 0, lang_ErrSanityCheck);
            return;
         }
      }
      else
      if (strsame (CgiFormActionPtr, lang_BtnRemove, -1))
         DoAddrListRemove = true;
      else
      {
         CgiLibResponseError (FI_LI, 0, lang_ErrSanityCheck);
         return;
      }
   }
   else
   {
      if (Debug) fprintf (stdout, "MailMessageId[0]: %d\n", MailMessageId[0]);
      if (MailMessageId[0])
         DoMessageRead = true;
      else
         DoFolderBrowse = true;
   }

   /*********/
   /* do it */
   /*********/

   EnableAccess (1);

#if YAHMAIL_PRIVATE
   if (ConfigPrivate)
      ProcessPrivateRequest ();
   else
#endif
      ProcessPublicRequest ();

   EnableAccess (0);
}

/*****************************************************************************/
/*
For selected requests scan through possible ID numbers to see whether they
exist as form fields. Possible numbers are 1 .. size of browse window.  Each
form element would contain a message ID.
*/

ProcessRequestIdNumbers ()
       
{
   boolean  DebugBuffer;
   int  IdNumber;
   char  *cptr;
   char  Scratch [64];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProcessRequestIdNumbers()\n");

   /* check for message ID numbers in the query string as form fields */
   if (DebugBuffer = Debug) Debug = false;
   MailMessageIdCount = 0;
   for (IdNumber = 1; IdNumber <= BrowseWindow; IdNumber++)
   {
      sprintf (Scratch, "WWW_FORM_ID%d", IdNumber);
      if ((cptr = CgiLibVarNull (Scratch)) != NULL)
      {
         if (MailMessageIdCount > MAX_MESSAGE_ID)
         {
            CgiLibResponseError (FI_LI, 0, lang_ErrInternalLimit);
            return;
         }
         MailMessageId[MailMessageIdCount++] = atoi(cptr);
         if (DebugBuffer)
            fprintf (stdout, "id: %d=%d\n", MailMessageIdCount, IdNumber);
      }
   }
   /* if checkboxes selected then zero any number supplied with the path */
   if (MailMessageIdCount) MailMessageId[MailMessageIdCount] = 0;
   Debug = DebugBuffer;
   if (Debug) fprintf (stdout, "MailMessageIdCount: %d\n", MailMessageIdCount);
}

/*****************************************************************************/
/*
The username has been authenticated for this request, and is assumed to be a
VMS account username.
*/

ProcessPrivateRequest ()
       
{
   register char  *cptr, *sptr;

   int  idx,
        len,
        status,
        IdNumber;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProcessPrivateRequest()\n");

   if (!MailUserName[0])
   {
      CgiLibResponseError (FI_LI, 0, lang_ErrSanityCheck);
      return;
   }

   if (!MailFileName[0]) strcpy (MailFileName, "MAIL");
   if (!MailFolderName[0]) strcpy (MailFolderName, "NEWMAIL");
   MailUserNameLength = strlen(MailUserName);
   MailFileNameLength = strlen(MailFileName);
   MailFolderNameLength = strlen(MailFolderName);

   if (Debug)
      fprintf (stdout, "|%s|%s|%s|%d|\n",
               MailUserName, MailFileName, MailFolderName, MailMessageId[0]);

   CgiLibHtmlEscape (MailFolderName, -1,
                     HtmlEscapedMailFolderName,
                     sizeof(HtmlEscapedMailFolderName));

   /* if username was specified an asterisk then generate a list of users */
   if (strchr (MailUserName, '*') != NULL ||
       strchr (MailUserName, '%') != NULL) DoUserList = true;

   BuildActionPathInfo ();

   if (DoMessageSend)
   {
      /******************************/
      /* send message (POST method) */
      /******************************/

      SendMailMessage ();
      return;
   }

   if (DoUserList)
   {
      /*********************/
      /* user profile list */
      /*********************/

      if (!PostMasterAuthenticated)
      {
         CgiLibResponseError (FI_LI, 0, lang_ErrConfigRestriction);
         return;
      }
      status = GenerateUserList();
      if (Debug) fprintf (stdout, "GenerateUserList() %%X%08.08X\n", status);
      if (VMSnok (status))
      {
         ResponseErrorMail (FI_LI, status, lang_ErrGenerateUserList);
         return;
      }
      BrowseUserList();
      return;
   }

   if (DoUserCreate)
   {
      /***********************/
      /* user profile create */
      /***********************/

      if (!PostMasterAuthenticated)
      {
         CgiLibResponseError (FI_LI, 0, lang_ErrConfigRestriction);
         return;
      }
      status = MailCreateUserEntry ();
      if (Debug) fprintf (stdout, "MailCreateUserEntry() %%X%08.08X\n", status);
      if (VMSnok (status)) return;
      CgiLibResponseSuccess (
"%s\n\
<P><FORM>\n\
<INPUT TYPE=button VALUE=\"%s\" onClick=\"history.go(-1)\">\n\
</FORM>",
         lang_ProfileCreated, lang_BtnGoBack);
      return;
   }

   if (DoUserDelete)
   {
      /***********************/
      /* user profile delete */
      /***********************/

      if (!PostMasterAuthenticated)
      {
         CgiLibResponseError (FI_LI, 0, lang_ErrConfigRestriction);
         return;
      }
      status = MailDeleteUserEntry();
      if (Debug) fprintf (stdout, "MailDeleteUserEntry() %%X%08.08X\n", status);
      if (VMSnok (status)) return;
      CgiLibResponseSuccess (
"%s\n\
<P><FORM>\n\
<INPUT TYPE=button VALUE=\"%s\" onClick=\"history.go(-1)\">\n\
</FORM>",
         lang_ProfileDeleted, lang_BtnGoBack);
      return;
   }

   if (DoUserSet)
   {
      /********************/
      /* user profile set */
      /********************/

      status = MailSetUserInfo();
      if (Debug) fprintf (stdout, "MailSetUserInfo() %%X%08.08X\n", status);
      if (VMSnok (status)) return;
      CgiLibResponseSuccess (
"%s\n\
<P><FORM>\n\
<INPUT TYPE=button VALUE=\"%s\" onClick=\"history.go(-1)\">\n\
</FORM>",
         lang_ProfileUpdated, lang_BtnGoBack);
      return;
   }

   /*****************************************************************/
   /* most other actions now require at least some of the user info */
   /*****************************************************************/

   status = MailGetUserInfo ();
   if (Debug) fprintf (stdout, "MailGetUserInfo() %%X%08.08X\n", status);
   if (VMSnok (status))
   {
      ResponseErrorMail (FI_LI, status, lang_ErrGetUserInfo);
      return;
   }

   if (DoUserProfile)
   {
      /************************/
      /* user profile display */
      /************************/

      DisplayUserProfile ();
      return;
   }

   if (DoMessageCreate)
   {
      /**********************/
      /* create new message */
      /**********************/

      CreateMailMessage ();
      return;
   }

   if (DoFolderBrowse)
   {
      /*****************/
      /* folder browse */
      /*****************/

      status = MailFileOpen ();
      if (Debug) fprintf (stdout, "MailFileOpen() %%X%08.08X\n", status);
      if (VMSnok (status))
      {
         ResponseErrorMail (FI_LI, status, lang_ErrOpenMail);
         return;
      }

      status = GenerateFolderList ();
      if (Debug) fprintf (stdout, "GenerateFolderList() %%X%08.08X\n", status);
      if (VMSnok (status))
      {
         ResponseErrorMail (FI_LI, status, lang_ErrGenerateFolderList);
         return;
      }

      status = GenerateMessageList ();
      if (VMSok (status))
         BrowseFolderPrivate ();
      else
         ResponseErrorMail (FI_LI, status, lang_ErrGenerateMessageList);

      MailFileClose ();

      return;
   }

   if (DoMessageRead || DoMessageReply || DoMessageForward)
   {
      /******************************/
      /* message read/reply/forward */
      /******************************/

      status = MailFileOpen ();
      if (Debug) fprintf (stdout, "MailFileOpen() %%X%08.08X\n", status);
      if (VMSnok (status))
      {
         ResponseErrorMail (FI_LI, status, lang_ErrOpenMail);
         return;
      }

      status = MailReadMessage ();
      if (Debug) fprintf (stdout, "MailReadMessage() %%X%08.08X\n", status);
      if (VMSok (status))
      {
         if (DoMessageRead)
            DisplayMailMessage ();
         else
         if (DoMessageReply || DoMessageForward)
            CreateMailMessage ();
      }
      else
         ResponseErrorMail (FI_LI, status, lang_ErrMessageRead);

      MailFileClose ();

      return;
   }

   if (DoMessageDelete)
   {
      /******************/
      /* message delete */
      /******************/

      status = MailFileOpen ();
      if (Debug) fprintf (stdout, "MailFileOpen() %%X%08.08X\n", status);
      if (VMSnok (status))
      {
         ResponseErrorMail (FI_LI, status, lang_ErrOpenMail);
         return;
      }

      status = DeleteMessage ();
      if (Debug) fprintf (stdout, "DeleteMessage() %%X%08.08X\n", status);
      if (VMSok (status))
      {
         CgiLibResponseSuccess (
"%s\n\
<P><FORM>\n\
<INPUT TYPE=button VALUE=\"%s\" onClick=\"history.go(-1)\">\n\
</FORM>",
            lang_MessagesDeleted, lang_BtnGoBack);
      }
      else
         ResponseErrorMail (FI_LI, status, lang_ErrMessageDelete);

      MailFileClose ();

      return;
   }

   if (DoMessageCopy || DoMessageMove)
   {
      /*********************/
      /* message copy/move */
      /*********************/

      if (!strcmp (MailFolderName, CopyFolderName))
      {
         CgiLibResponseError (FI_LI, 0, lang_ErrFoldersIdentical);
         return;
      }

      status = MailFileOpen ();
      if (Debug) fprintf (stdout, "MailFileOpen() %%X%08.08X\n", status);
      if (VMSnok (status))
      {
         ResponseErrorMail (FI_LI, status, lang_ErrOpenMail);
         return;
      }

      status = CopyMoveMessage ();
      if (Debug) fprintf (stdout, "CopyMoveMessage() %%X%08.08X\n", status);
      if (VMSok (status))
      {
         if (DoMessageCopy)
            cptr = lang_MessagesCopied;
         else
            cptr = lang_MessagesMoved;
         CgiLibResponseSuccess (
"%s\n\
<P><FORM>\n\
<INPUT TYPE=button VALUE=\"%s\" onClick=\"history.go(-1)\">\n\
</FORM>",
            cptr, lang_BtnGoBack);
      }
      else
         ResponseErrorMail (FI_LI, status, lang_ErrMessageMove);

      MailFileClose ();

      return;
   }

   if (DoMessageEmptyWasteBasket)
   {
      /*********************/
      /* empty wastebasket */
      /*********************/

      status = MailFileOpen ();
      if (Debug) fprintf (stdout, "MailFileOpen() %%X%08.08X\n", status);
      if (VMSnok (status))
      {
         ResponseErrorMail (FI_LI, status, lang_ErrOpenMail);
         return;
      }

      status = EmptyWasteBasket ();
      if (Debug) fprintf (stdout, "EmptyWasteBasket() %%X%08.08X\n", status);
      if (VMSok (status))
      {
         CgiLibResponseSuccess (
"%s\n\
<P><FORM>\n\
<INPUT TYPE=button VALUE=\"%s\" onClick=\"history.go(-1)\">\n\
</FORM>",
            MailFileMessagesDeleted ? lang_WasteBasketPurged :
                                      lang_WasteBasketEmpty,
            lang_BtnGoBack);
      }
      else
         ResponseErrorMail (FI_LI, status, lang_ErrMessageDelete);

      MailFileClose ();

      return;
   }

   if (DoAddrListSelect)
   {
      /**************************/
      /* list all address lists */
      /**************************/

      AddrListSelect ();
      return;
   }

   if (DoAddrListAccess)
   {
      /****************************/
      /* address list access/edit */
      /****************************/

      AddrListAccess ();
      return;
   }

   if (DoAddrListUpdate)
   {
      /*************************************/
      /* update address list (POST method) */
      /*************************************/

      AddrListUpdate ();
      return;
   }

   if (DoAddrListRemove)
   {
      /***********************/
      /* delete address list */
      /***********************/

      AddrListRemove ();
      return;
   }

   if (DoSigFileAccess)
   {
      /***********************/
      /* signature file edit */
      /***********************/

      SigFileAccess ();
      return;
   }

   if (DoSigFileUpdate)
   {
      /***************************************/
      /* update signature file (POST method) */
      /***************************************/

      SigFileUpdate ();
      return;
   }

   CgiLibResponseError (FI_LI, 0, lang_ErrSanityCheck);
}

/*****************************************************************************/
/*
An un-authenticated request is allowed limited read access to specifically
configured public area/archive.
*/

ProcessPublicRequest ()
       
{
   register char  *cptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProcessPublicRequest()\n");

   if (MailFileName[0])
   {
      CgiLibResponseError (FI_LI, 0, lang_ErrConfigRestriction);
      return;
   }
   if (MailFolderName[0] && PublicFolderName[0] &&
      strcmp (MailFolderName, PublicFolderName))
   {
      CgiLibResponseError (FI_LI, 0, lang_ErrConfigRestriction);
      return;
   }

   /* ensure these come from the configuration file, nowhere else!! */
   strzcpy (MailUserName, PublicUserName, sizeof(MailUserName));
   strzcpy (MailFileName, PublicFileName, sizeof(MailFileName));
   /* if the configuration specifies a folder then force that, otherwise */
   if (PublicFolderName[0])
      strzcpy (MailFolderName, PublicFolderName, sizeof(MailFolderName));
   else
   if (CgiFormFolderNamePtr != NULL && CgiFormFolderNamePtr[0])
      strzcpy (MailFolderName, CgiFormFolderNamePtr, sizeof(MailFolderName));
   else
   if (!MailFolderName[0])
      strcpy (MailFolderName, "NEWMAIL");

   MailUserNameLength = strlen(MailUserName);
   MailFileNameLength = strlen(MailFileName);
   MailFolderNameLength = strlen(MailFolderName);

   if (Debug)
      fprintf (stdout, "|%s|%s|%s|%d|\n",
               MailUserName, MailFileName, MailFolderName, MailMessageId[0]);

   CgiLibHtmlEscape (MailFolderName, -1,
                     HtmlEscapedMailFolderName,
                     sizeof(HtmlEscapedMailFolderName));

   DoFolderBrowse = DoMessageRead = false;
   if (MailMessageId[0])
      DoMessageRead = true;
   else
      DoFolderBrowse = true;

   BuildActionPathInfo ();

   if (DoFolderBrowse)
   {
      /************************/
      /* public folder browse */
      /************************/

      status = MailFileOpen ();
      if (Debug) fprintf (stdout, "MailFileOpen() %%X%08.08X\n", status);
      if (VMSnok (status))
      {
         ResponseErrorMail (FI_LI, status, lang_ErrOpenMail);
         return;
      }

      if (ConfigPublicBrowseFolders)
      {
         status = GenerateFolderList ();
         if (Debug)
            fprintf (stdout, "GenerateFolderList() %%X%08.08X\n", status);
         if (VMSnok (status))
         {
            ResponseErrorMail (FI_LI, status, lang_ErrGenerateFolderList);
            return;
         }
         if (Debug) fprintf (stdout, "|%s|\n", MailFolderName);
      }

      status = MailReadPublicConfigMessage ();
      if (Debug)
         fprintf (stdout, "MailReadPublicConfigMessage() %%X%08.08X\n", status);

      status = GenerateMessageList ();
      if (Debug) fprintf (stdout, "GenerateMessageList() %%X%08.08X\n", status);

      if (VMSok (status))
         BrowseFolderPublic ();
      else
         ResponseErrorMail (FI_LI, status, lang_ErrGenerateMessageList);

      MailFileClose ();

      return;
   }

   if (DoMessageRead)
   {
      /************************/
      /* public message read */
      /************************/

      status = MailFileOpen ();
      if (Debug) fprintf (stdout, "MailFileOpen() %%X%08.08X\n", status);
      if (VMSnok (status))
      {
         ResponseErrorMail (FI_LI, status, lang_ErrOpenMail);
         return;
      }

      for (;;)
      {
         status = MailReadMessage ();
         if (Debug) fprintf (stdout, "MailReadMessage() %%X%08.08X\n", status);
         if (VMSok (status))
         {
            if (strstr (VmsMailSubject, PublicConfigSubjPtr) != NULL)
            {
               /* woops, bumped into a public folder configuration message */
               if (MessageIncrement)
               {
                  /* next/previous requested, so onto next message */
                  MailMessageId[0] += MessageIncrement;
                  continue;
               }
               else
                  ResponseErrorMail (FI_LI, MAIL$_NOMOREMSG,
                                     lang_ErrMessageRead);
            }
            else
               DisplayMailMessage ();
         }
         else
            ResponseErrorMail (FI_LI, status, lang_ErrMessageRead);

         break;
      }

      MailFileClose ();

      return;
   }

   CgiLibResponseError (FI_LI, 0, lang_ErrConfigRestriction);
}

/*****************************************************************************/
/*
Build the request path used in consequent YAHMAIL requests.  Two differing
paths are used.  The authenticated path comprises the script name, a tilde and
the user name (e.g. "/~DANIEL"), the mail file name (just the name, e.g.
"MAIL"), and the folder name.  Non-authenticated (public) paths comprise the
script name, the public name, and the folder name.
*/

BuildActionPathInfo ()
       
{
   register char  *cptr, *sptr, *tptr;

   char  Scratch [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "BuildActionPathInfo()\n");

   /* user-based access uses username, mail file name, and folder */
   sptr = ActionPathInfo;
   for (cptr = CgiScriptNamePtr; *cptr; *sptr++ = *cptr++);
   *sptr++ = '/';
   if (ConfigPrivate) *sptr++ = TildeChar;
   tptr = sptr;
   sptr = Scratch;
   for (cptr = MailUserName; *cptr; *sptr++ = *cptr++);
   *sptr++ = '/';
   if (ConfigPrivate)
   {
      for (cptr = MailFileName; *cptr; *sptr++ = *cptr++);
      *sptr++ = '/';
   }
   for (cptr = MailFolderName; *cptr; *sptr++ = *cptr++);
   *sptr++ = '/';
   *sptr = '\0';
   CgiLibUrlEncode (Scratch, -1, tptr, -1);
   if (Debug) fprintf (stdout, "ActionPathInfo |%s|\n", ActionPathInfo);
}

/*****************************************************************************/
/*
Build a query string with sufficient form elements suitable to carry the
context of the current request into the next one.
*/

BuildActionQueryString ()
       
{
   char  *PickCcPtr,
         *PickFromPtr,
         *PickSubjPtr,
         *PickToPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "BuildActionQueryString()\n");

   PickFromPtr = CgiLibUrlEncode (MailSubstringFrom, -1, NULL, 0);
   PickSubjPtr = CgiLibUrlEncode (MailSubstringSubj, -1, NULL, 0);

   if (ConfigPrivate)
   {
      PickCcPtr = CgiLibUrlEncode (MailSubstringCc, -1, NULL, 0);
      PickToPtr = CgiLibUrlEncode (MailSubstringTo, -1, NULL, 0);

      sprintf (ActionQueryString,
"?cmc=%d&anc=%d&pfr=%s&pto=%s&pcc=%s&psu=%s&cnm=%d&win=%d&wra=%d",
               VmsMailMessageSelectedCount, MakeAnchors,
               PickFromPtr, PickToPtr, PickCcPtr, PickSubjPtr,
               CheckNewMailMinutes, BrowseWindow, WrapAt);

      free (PickCcPtr);
      free (PickToPtr);
   }
   else
   {
      sprintf (ActionQueryString,
               "?cmc=%d&anc=%d&pfr=%s&psu=%s&win=%d",
               VmsMailMessageSelectedCount, MakeAnchors,
               PickFromPtr, PickSubjPtr, BrowseWindow);
   }

   free (PickFromPtr);
   free (PickSubjPtr);

   if (Debug) fprintf (stdout, "ActionQueryString |%s|\n", ActionQueryString);
}

/*****************************************************************************/
/*
Display an HTML page containing a single mail message.  Include a form for
requesting the message be returned as a plain-text page or as an octet-stream
download (i.e. binary).  Page is different for public and private access!
*/


DisplayMailMessage ()

{
   int  status;
   char  NextMessage [96],
         PrevMessage [96],
         SizeString [64];
   char  *CcPtr,
         *CharsetPtr,
         *FromPtr,
         *IncludeRfc822HeaderPtr,
         *MimePtr,
         *NextMessagePtr,
         *PrevMessagePtr,
         *PickCcPtr,
         *PickFromPtr,
         *PickSubjPtr,
         *PickToPtr,
         *Rfc822Ptr,
         *SubjectPtr,
         *ToPtr,
         *TextPtr;
   char  RfcHeaderString [128],
         NoneString [64];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DisplayMailMessage()\n");

   if (DoMessageReadPlainText)
   {
      DisplayMailMessagePlainText ();
      return;
   }
   if (DoMessageReadOctetStream)
   {
      DisplayMailMessageOctetStream ();
      return;
   }
#ifdef YAHMAIL_MUNPACK
   if (DoMessageReadMunpack)
   {
      CgiFormMunpackContentTypePtr = CgiLibVar ("WWW_FORM_MCT");
      MunpackYahmailFile (CgiFormMunpackPtr, CgiFormMunpackContentTypePtr);
      return;
   }
#endif /* YAHMAIL_MUNPACK */

   /*************/
   /* HTML page */
   /*************/

   FromPtr = CgiLibAnchorHtmlEscape (VmsMailFrom, -1, NULL, 0, MakeAnchors);
   ToPtr = CgiLibAnchorHtmlEscape (VmsMailTo, -1, NULL, 0, MakeAnchors);
 
   if (VmsMailCc[0])
      CcPtr = CgiLibAnchorHtmlEscape (VmsMailCc, -1, NULL, 0, MakeAnchors);
   else
      sprintf (CcPtr = NoneString, "<I>(%s)</I>", lang_None);

   if (VmsMailSubject[0])
      SubjectPtr = CgiLibHtmlEscape (VmsMailSubject, -1, NULL, 0);
   else
      sprintf (SubjectPtr = NoneString, "<I>(%s)</I>", lang_None);

   if (VmsMailMessageFlags & MAIL$M_EXTNSTD)
      strcpy (SizeString, "[<FONT SIZE=-1>FOREIGN</FONT>]");
   else
   if (VmsMailTextLength)
      sprintf (SizeString, "%d %s", VmsMailTextLength, lang_SizeBytes);
   else
      sprintf (SizeString, "%d %s", VmsMailMessageSize, lang_SizeRecords);

   if (VmsMailMessageId > 1)
   {
      sprintf (PrevMessage,
"<INPUT TYPE=submit NAME=act VALUE=\"%s\" \
onClick=\"buildFormURL(\'%s\')\">",
               lang_BtnPrevious, lang_BtnPrevious);
      PrevMessagePtr = PrevMessage;
   }
   else
      PrevMessagePtr = "";

   if (VmsMailMessageId < VmsMailMessageSelectedCount)
   {
      sprintf (NextMessage,
"<INPUT TYPE=submit NAME=act VALUE=\"%s\" \
onClick=\"buildFormURL(\'%s\')\">",
               lang_BtnNext, lang_BtnNext);
      NextMessagePtr = NextMessage;
   }
   else
      NextMessagePtr = "";

   PickFromPtr = CgiLibHtmlEscape (MailSubstringFrom, -1, NULL, 0);
   PickSubjPtr = CgiLibHtmlEscape (MailSubstringSubj, -1, NULL, 0);
   if (ConfigPrivate)
   {
      PickToPtr = CgiLibHtmlEscape (MailSubstringTo, -1, NULL, 0);
      PickCcPtr = CgiLibHtmlEscape (MailSubstringCc, -1, NULL, 0);
   }

   BuildActionQueryString ();

   if (VmsMailRfc822HeaderLength)
   {
      Rfc822Ptr = "&nbsp;&nbsp;[<FONT SIZE=-1>RFC822</FONT>]";
      sprintf (IncludeRfc822HeaderPtr = RfcHeaderString,
"&nbsp;&nbsp;<INPUT TYPE=checkbox NAME=rfc VALUE=1>\
&nbsp;<I>RFC822&nbsp;&nbsp;%s</I>",
               lang_Header);

      MimePtr = "";
      CharsetPtr = VmsMailRfc822MimeCharset;

#ifdef YAHMAIL_MUNPACK

      if (VmsMailRfc822MimeVersion &&
          (!strsame (VmsMailRfc822MimeContentType, "text/plain", -1) ||
           !(VmsMailRfc822MimeContTransEnc[0] == '7' ||
             VmsMailRfc822MimeContTransEnc[0] == '8' ||
             strsame (VmsMailRfc822MimeContTransEnc, "binary", -1))))
      {
         /* not a simple plain-text MIME message */
         static char  MimeString [128];
         extern int  MunpackAttachmentCount;
         extern char  MunpackCharset[];

         MunpackYahmailProcess (false);

         if (Debug) fprintf (stdout, "MunpackCharset |%s|\n", MunpackCharset);
         if (!*(CharsetPtr = MunpackCharset))
            CharsetPtr = VmsMailRfc822MimeCharset;

         if (MunpackAttachmentCount)
            sprintf (MimePtr = MimeString,
"&nbsp;&nbsp;[<FONT SIZE=-1>MIME</FONT>]&nbsp;&nbsp;%d&nbsp;%s",
               MunpackAttachmentCount,
               MunpackAttachmentCount == 1 ? lang_AttachmentSingular :
                                             lang_AttachmentPlural);
         else
            MimePtr = "&nbsp;&nbsp;[<FONT SIZE=-1>MIME</FONT>]";
      }

#endif /* YAHMAIL_MUNPACK */

   }
   else
   {
      MimePtr = Rfc822Ptr = "";
      IncludeRfc822HeaderPtr =
"<INPUT TYPE=hidden NAME=rfc VALUE=\"\">&nbsp;&nbsp;<I>header</I>";
      CharsetPtr = VmsMailRfc822MimeCharset;
   }

   if (CharsetPtr[0])
      CgiLibResponseHeader (200, "text/html", "; charset=%s\n", CharsetPtr);
   else
      CgiLibResponseHeader (200, "text/html");

   fprintf (stdout,
"<HTML>\n\
<HEAD>\n\
<META NAME=\"generator\" CONTENT=\"%s\">\n\
<META NAME=\"environment\" CONTENT=\"%s\">\n\
<META NAME=\"language\" CONTENT=\"%s %s\">\n\
<TITLE>yahMAIL ... %s</TITLE>\n\
</HEAD>\n\
<SCRIPT LANGUAGE=\"JavaScript\">\n\
<!--\n\
\
var formURL = \"\";\n\
\
function doSubmit () {\n\
   if (formURL == \"\") return true;\n\
   location.replace(formURL);\n\
   return false;\n\
}\n\
\
function buildFormURL (actionButton) {\n\
   var elementName, elementValue;\n\
   formURL = document.formPage.action + \"?act=\" + escape(actionButton);\n\
   for (idx = 0; idx < document.formPage.elements.length; idx++) {\n\
      if (document.formPage.elements[idx].type == \"hidden\") {\n\
         elementName = document.formPage.elements[idx].name;\n\
         elementValue = document.formPage.elements[idx].value;\n\
         formURL += \"&\" + elementName + \"=\" + escape(elementValue);\n\
      }\n\
      else\n\
      if (document.formPage.elements[idx].type == \"checkbox\") {\n\
         if (document.formPage.elements[idx].checked) {\n\
            elementName = document.formPage.elements[idx].name;\n\
            elementValue = document.formPage.elements[idx].value;\n\
            formURL += \"&\" + elementName + \"=\" + escape(elementValue);\n\
         }\n\
      }\n\
   }\n\
}\n\
\
// -->\n\
</SCRIPT>\n\
<BODY %s>\n\
%s\
<FORM ACTION=\"%s%d\" NAME=formPage onSubmit=\"return doSubmit()\">\n\
<INPUT TYPE=hidden NAME=cmc VALUE=\"%d\">\n\
<INPUT TYPE=hidden NAME=wra VALUE=\"%d\">\n\
<TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>\n\
<TR><TH ALIGN=right VALIGN=top>%s:&nbsp;&nbsp;</TH><TD>%20.20s</TD></TR>\n\
<TR><TH ALIGN=right VALIGN=top>%s:&nbsp;&nbsp;</TH><TD>%s</TD></TR>\n\
<TR><TH ALIGN=right VALIGN=top>%s:&nbsp;&nbsp;</TH><TD>%s</TD></TR>\n\
<TR><TH ALIGN=right VALIGN=top>%s:&nbsp;&nbsp;</TH><TD>%s</TD></TR>\n\
<TR><TH ALIGN=right VALIGN=top>%s:&nbsp;&nbsp;</TH><TD>%s</TD></TR>\n\
<TR><TH ALIGN=right VALIGN=top>%s:&nbsp;&nbsp;</TH>\
<TD>%s%s%s<!-- FLAGS: %04.04X --></TD></TR>\n\
<TR><TH ALIGN=right VALIGN=top>%s:&nbsp;&nbsp;</TH>\
<TD>%s / %s%s%s / %d</TD></TR>\n\
<TR><TD HEIGHT=5></TD></TR>\n\
<TR><TD COLSPAN=2>\n\
&nbsp;%s%s%s%s<INPUT TYPE=button VALUE=\"%s\" onClick=\"history.go(-1)\">\n\
</TD></TR>\n\
</TABLE>\n\
<PRE><HR ALIGN=left WIDTH=70%% HEIGHT=1 NOSHADE>",
      SoftwareID, CgiEnvironmentPtr, lang_Language, lang_LanguageAuthor,
      VmsMailSubject[0] ? SubjectPtr : NoneString,
      ConfigBodyTagPtr, PostMasterWarningPtr,
      ActionPathInfo, VmsMailMessageId,
      CurrentMessageCount, WrapAt,
      lang_MsgDate, VmsMailDate,
      lang_MsgFrom, FromPtr,
      lang_MsgSubj, SubjectPtr,
      lang_MsgTo, ToPtr,
      lang_MsgCc, CcPtr, 
      lang_MsgStatus, SizeString, Rfc822Ptr, MimePtr, VmsMailMessageFlags,
      lang_MsgMessage, ConfigPrivate ? MailUserName : PathPart[0],
      ConfigPrivate ? MailFileName : "",
      ConfigPrivate ? " / " : "",
      HtmlEscapedMailFolderName, VmsMailMessageId,
      PrevMessagePtr,
      (PrevMessagePtr[0] && NextMessagePtr[0]) ? "&nbsp;\n" : "",
      NextMessagePtr,
      (PrevMessagePtr[0] || NextMessagePtr[0]) ? "&nbsp;\n" : "",
      lang_BtnGoBack);

   if (VmsMailMessageFlags & MAIL$M_EXTNSTD)
      fprintf (stdout, " <INPUT TYPE=submit NAME=act VALUE=\"%s\">",
               lang_BtnForeign);
#ifdef YAHMAIL_MUNPACK
   else
   if (VmsMailRfc822MimeVersion &&
       (!strsame (VmsMailRfc822MimeContentType, "text/plain", -1) ||
        !(VmsMailRfc822MimeContTransEnc[0] == '7' ||
          VmsMailRfc822MimeContTransEnc[0] == '8' ||
          strsame (VmsMailRfc822MimeContTransEnc, "binary", -1))))
   {
      /* not a simple plain-text MIME message */
      extern int  MunpackAttachmentCount;
      MunpackYahmailText (stdout, true);
      if (MunpackAttachmentCount) MunpackYahmailParts (stdout);
   }
#endif /* YAHMAIL_MUNPACK */
   else
   {
      /* text file */
      TextPtr = CgiLibAnchorHtmlEscape (VmsMailTextPtr +
                                        VmsMailHeaderLength +
                                        VmsMailRfc822HeaderLength,
                                        -1, NULL, 0, MakeAnchors);
      fputs (TextPtr, stdout);
      free (TextPtr);
   }

   fputs (
"<HR ALIGN=left WIDTH=70%% HEIGHT=1 NOSHADE></PRE>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n",
      stdout);

   if (ConfigPrivate)
   {
      fprintf (stdout,
"<TR><TD>\n\
<INPUT TYPE=submit NAME=act VALUE=\"%s\" \
onClick=\"buildFormURL(\'%s\')\">&nbsp;\n\
<INPUT TYPE=submit NAME=act VALUE=\"%s\" \
onClick=\"buildFormURL(\'%s\')\">&nbsp;&nbsp;\n\
</TD><TD>\n\
<INPUT TYPE=checkbox NAME=taw VALUE=%d%s>\
&nbsp;<I>%s</I>&nbsp;\n\
<INPUT TYPE=checkbox NAME=inc VALUE=1 CHECKED>\
&nbsp;<I>%s</I>&nbsp;\n\
<INPUT TYPE=checkbox NAME=quo VALUE=1 CHECKED>\
&nbsp;<I>%s&nbsp;&nbsp;%s</I><BR>\n\
</TD></TR>\n",
         lang_BtnReply, lang_BtnReply,
         lang_BtnForward, lang_BtnForward,
         WrapAt ? WrapAt : 72, WrapAt ? " CHECKED" : "", lang_AutoWrap, 
         lang_ChkInclude, lang_ChkQuote, lang_ChkMessage);
   }

   fprintf (stdout,
"<TR><TD>\n\
<INPUT TYPE=submit NAME=act VALUE=\"%s\">&nbsp;\n\
<INPUT TYPE=submit NAME=act VALUE=\"%s\">&nbsp;&nbsp;\n\
</TD><TD>\n\
<INPUT TYPE=checkbox NAME=hea VALUE=1>&nbsp;<I>VMS</I>%s<BR>\n\
</TD></TR>\n\
<INPUT TYPE=hidden NAME=pfr VALUE=\"%s\">\n\
<INPUT TYPE=hidden NAME=psu VALUE=\"%s\">\n",
      lang_BtnText, lang_BtnBinary,
      IncludeRfc822HeaderPtr, PickFromPtr, PickSubjPtr);

   if (ConfigPrivate)
   {
      fprintf (stdout,
"<INPUT TYPE=hidden NAME=pcc VALUE=\"%s\">\n\
<INPUT TYPE=hidden NAME=pto VALUE=\"%s\">\n\
<INPUT TYPE=hidden NAME=mcs VALUE=\"%s\">\n",
      PickCcPtr, PickToPtr, VmsMailRfc822MimeCharset);
   }

   fprintf (stdout,
"<INPUT TYPE=hidden NAME=nnn VALUE=\"%s\">\n\
</FORM>\n\
</TABLE>\n\
<P><FORM>\n\
&nbsp;<INPUT TYPE=button VALUE=\"%s\" onClick=\"history.go(-1)\">\n\
</FORM>\n\
</BODY>\n\
</HTML>\n",
      CgiFormNnnPtr, lang_BtnGoBack);

   free (PickFromPtr);
   free (PickSubjPtr);
   if (ConfigPrivate)
   {
      free (PickCcPtr);
      free (PickToPtr);
   }

   free (FromPtr);
   free (SubjectPtr);
   free (CcPtr);
   free (ToPtr);
}

/*****************************************************************************/
/*
Return the mail message body as a plain text page.
*/

DisplayMailMessagePlainText ()

{                                                          
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DisplayMailMessagePlainText()\n");

   if (VmsMailMessageFlags & MAIL$M_EXTNSTD)
   {
      DisplayMailMessageForeign ();
      return;
   }

   cptr = CgiLibVar("WWW_FORM_HEA");
   if (cptr[0])
   {
      /* include VMS mail header lines (the whole box and dice) */
      if (VmsMailRfc822MimeCharset[0])
         CgiLibResponseHeader (200, "text/plain",
                               "; charset=%s\n", VmsMailRfc822MimeCharset);
      else
         CgiLibResponseHeader (200, "text/plain");
      fputs (VmsMailTextPtr, stdout);
   }
   else
   {
      cptr = CgiLibVar("WWW_FORM_RFC");
      if (cptr[0])
      {
         /* include RFC822 header lines */
         if (VmsMailRfc822MimeCharset[0])
            CgiLibResponseHeader (200, "text/plain",
                                  "; charset=%s\n", VmsMailRfc822MimeCharset);
         else
            CgiLibResponseHeader (200, "text/plain");
         fputs (VmsMailTextPtr+VmsMailHeaderLength, stdout);
      }
      else
      {
         /* exclude VMS mail and any RFC822 header lines */
         if (VmsMailRfc822MimeCharset[0])
            CgiLibResponseHeader (200, "text/plain",
                                  "; charset=%s\n", VmsMailRfc822MimeCharset);
         else
            CgiLibResponseHeader (200, "text/plain");
         fputs (VmsMailTextPtr +
                VmsMailHeaderLength +
                VmsMailRfc822HeaderLength, stdout);
      }
   }
}

/*****************************************************************************/
/*
Return the mail message body as an octet-stream (i.e. binary).
*/

DisplayMailMessageOctetStream ()

{
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DisplayMailMessageOctetStream()\n");

   if (VmsMailMessageFlags & MAIL$M_EXTNSTD)
   {
      DisplayMailMessageForeign ();
      return;
   }

   CgiLibResponseHeader (200, "application/octet-stream");

   CgiLibEnvironmentBinaryOut();

   cptr = CgiLibVar("WWW_FORM_HEA");
   if (cptr[0])
   {
      /* include VMS mail header lines */
      if (Debug) fprintf (stdout, "%d bytes\n", VmsMailTextLength);
      if (!Debug) fwrite (VmsMailTextPtr, VmsMailTextLength, 1, stdout);
   }
   else
   {
      cptr = CgiLibVar("WWW_FORM_RFC");
      if (cptr[0])
      {
         /* include RFC822 header lines */
         if (Debug)
            fprintf (stdout, "%d bytes\n",
                     VmsMailTextLength - VmsMailHeaderLength);
         if (!Debug)
            fwrite (VmsMailTextPtr + VmsMailHeaderLength,
                    VmsMailTextLength - VmsMailHeaderLength,
                    1,
                    stdout);
      }
      else
      {
         /* exclude VMS mail and any RFC822 header lines */
         if (Debug)
            fprintf (stdout, "%d bytes\n",
                     VmsMailTextLength -
                     VmsMailHeaderLength -
                     VmsMailRfc822HeaderLength);
         if (!Debug)
            fwrite (VmsMailTextPtr +
                    VmsMailHeaderLength +
                    VmsMailRfc822HeaderLength,
                    VmsMailTextLength -
                    VmsMailHeaderLength -
                    VmsMailRfc822HeaderLength,
                    1,
                    stdout);
      }
   }

   CgiLibEnvironmentStandardOut();
}

/*****************************************************************************/
/*
Return the /FOREIGN mail message body as an octet-stream (i.e. binary).
*/

DisplayMailMessageForeign ()

{
   register char  *cptr;

   int  status,
        cnt,
        fd;
   char  Buffer [MESSAGE_TEXT_CHUNK],
         FileName [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DisplayMailMessageForeign()\n");

   if (!((VmsMailMessageFlags & MAIL$M_EXTNSTD) && VmsMailExtId[0]))
   {
      CgiLibResponseError (FI_LI, 0, lang_ErrSanityCheck);
      return;
   }

   if (ConfigPublic)
   {
      /* full specification for locations of mail files */
      strcpy (FileName, MailFileName);
      for (cptr = FileName; *cptr; cptr++);
      while (cptr > FileName && *cptr != ']') cptr--;
      if (*cptr != ']') while (*cptr && *cptr != ':') cptr++;
      if (*cptr == ']' || *cptr == ':') *++cptr = '\0';
   }
   else
   {
      /* user-based location of mail files */
      strcpy (FileName, VmsMailUserFullDirectory);
   }
   strcat (FileName, VmsMailExtId);
   if (Debug) fprintf (stdout, "|%s|\n", FileName);

   if ((fd = open (FileName, O_RDONLY, 0, "ctx=stm")) == -1)
   {
      status = vaxc$errno;
      CgiLibResponseError (FI_LI, status, lang_ErrMessageForeign);
      return;
   }

   if (DoMessageReadPlainText)
      CgiLibResponseHeader (200, "text/plain");
   else
      CgiLibResponseHeader (200, "application/octet-stream");

   CgiLibEnvironmentBinaryOut();

   while ((cnt = read (fd, Buffer, sizeof(Buffer))) > 0)
   {
      if (Debug) fprintf (stdout, "%d bytes\n", cnt);
      if (!Debug) fwrite (Buffer, cnt, 1, stdout);
   }

   CgiLibEnvironmentStandardOut();

   close (fd);
}

/*****************************************************************************/
/*
Display a form containing configurable elements of a VMS mail user profile
entry.  This form can then be submitted changing the actual entry.
*/

DisplayUserProfile ()

{
   static char  ItemChecked [] = " CHECKED";

   int  status;
   char  *UserNamePtr,
         *EditorPtr,
         *FormPtr,
         *ForwardingPtr,
         *FullDirectoryPtr,
         *PersonalNamePtr,
         *QueuePtr,
         *SigFilePtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DisplayUserProfile()\n");

   UserNamePtr = CgiLibHtmlEscape (MailUserName, -1, NULL, 0);
   EditorPtr = CgiLibHtmlEscape (VmsMailUserEditor, -1, NULL, 0);
   ForwardingPtr = CgiLibHtmlEscape (VmsMailUserForwarding, -1, NULL, 0);
   FullDirectoryPtr = CgiLibHtmlEscape (VmsMailUserFullDirectory, -1, NULL, 0);
   PersonalNamePtr = CgiLibHtmlEscape (VmsMailUserPersonalName, -1, NULL, 0);
   QueuePtr = CgiLibHtmlEscape (VmsMailUserQueue, -1, NULL, 0);
   FormPtr = CgiLibHtmlEscape (VmsMailUserForm, -1, NULL, 0);

   CgiLibResponseHeader (200, "text/html");
   fprintf (stdout,
"<HTML>\n\
<HEAD>\n\
<META NAME=\"generator\" CONTENT=\"%s\">\n\
<META NAME=\"environment\" CONTENT=\"%s\">\n\
<META NAME=\"language\" CONTENT=\"%s %s\">\n\
<TITLE>yahMAIL ... %s: %s</TITLE>\n\
</HEAD>\n\
<BODY %s>\n\
%s\
<B><FONT SIZE=+1><U>%s: %s</FONT></U></B>\n\
<BR>&nbsp;<FONT SIZE=-1>yahMAIL<SUP>*</SUP></FONT>\n\
<FORM ACTION=\"%s\">\n\
<TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>\n\
<TR><TH ALIGN=right VALIGN=center>%s:&nbsp;&nbsp;</TH>\
<TD><INPUT TYPE=radio NAME=uap VALUE=1%s>&nbsp;<I>enabled</I>&nbsp;&nbsp;\
<INPUT TYPE=radio NAME=uap VALUE=0%s>&nbsp;<I>disabled</I></TD></TR>\n\
<TR><TH ALIGN=right VALIGN=center>%s:&nbsp;&nbsp;</TH>\
<TD><INPUT TYPE=radio NAME=ucp VALUE=1%s>&nbsp;<I>enabled</I>&nbsp;&nbsp;\
<INPUT TYPE=radio NAME=ucp VALUE=0%s>&nbsp;<I>disabled</I></TD></TR>\n\
<TR><TH ALIGN=right VALIGN=center>%s:&nbsp;&nbsp;</TH>\
<TD><INPUT TYPE=radio NAME=ucf VALUE=1%s>&nbsp;<I>enabled</I>&nbsp;&nbsp;\
<INPUT TYPE=radio NAME=ucf VALUE=0%s>&nbsp;<I>disabled</I></TD></TR>\n\
<TR><TH ALIGN=right VALIGN=center>%s:&nbsp;&nbsp;</TH>\
<TD><INPUT TYPE=radio NAME=ucr VALUE=1%s>&nbsp;<I>enabled</I>&nbsp;&nbsp;\
<INPUT TYPE=radio NAME=ucr VALUE=0%s>&nbsp;<I>disabled</I></TD></TR>\n\
<TR><TH ALIGN=right VALIGN=center>%s:&nbsp;&nbsp;</TH>\
<TD><INPUT TYPE=radio NAME=ucs VALUE=1%s>&nbsp;<I>enabled</I>&nbsp;&nbsp;\
<INPUT TYPE=radio NAME=ucs VALUE=0%s>&nbsp;<I>disabled</I></TD></TR>\n\
<TR><TD HEIGHT=10></TD></TR>\n\
<TR><TH ALIGN=right VALIGN=center>%s:&nbsp;&nbsp;</TH>\
<TD><INPUT TYPE=text SIZE=20 NAME=ued VALUE=\"%s\"></TD></TR>\n\
<TR><TH ALIGN=right VALIGN=center>%s:&nbsp;&nbsp;</TH>\
<TD><INPUT TYPE=text SIZE=60 NAME=ufw VALUE=\"%s\"></TD></TR>\n\
<TR><TH ALIGN=right VALIGN=center>%s:&nbsp;&nbsp;</TH>\
<TD><INPUT TYPE=text SIZE=60 NAME=upn VALUE=\"%s\"></TD></TR>\n",
      SoftwareID, CgiEnvironmentPtr, lang_Language, lang_LanguageAuthor,
      lang_ProUserProfile, UserNamePtr,
      ConfigBodyTagPtr, PostMasterWarningPtr,
      lang_ProUserProfile, UserNamePtr,
      ActionPathInfo,
      lang_ProAutoPurge, VmsMailUserAutoPurge ? ItemChecked : "",
                         VmsMailUserAutoPurge ? "" : ItemChecked,
      lang_ProCcPrompt, VmsMailUserCcPrompt ? ItemChecked : "",
                        VmsMailUserCcPrompt ? "" : ItemChecked,
      lang_ProCopyFwd,  VmsMailUserCopyForward ? ItemChecked : "",
                        VmsMailUserCopyForward ? "" : ItemChecked,
      lang_ProCopyReply, VmsMailUserCopyReply ? ItemChecked : "",
                         VmsMailUserCopyReply ? "" : ItemChecked,
      lang_ProCopySend, VmsMailUserCopySend ? ItemChecked : "",
                        VmsMailUserCopySend ? "" : ItemChecked,
      lang_ProEditor, EditorPtr,
      lang_ProForwarding, ForwardingPtr,
      lang_ProPersonalName, PersonalNamePtr);

   if (VmsVersion >= 70)
   {
      SigFilePtr = CgiLibHtmlEscape (VmsMailUserSigFile, -1, NULL, 0);

      fprintf (stdout,
"<TR><TH ALIGN=right VALIGN=center>%s:&nbsp;&nbsp;</TH>\
<TD><INPUT TYPE=text SIZE=20 NAME=usf VALUE=\"%s\">\
<INPUT TYPE=hidden NAME=wha VALUE=\"sig\">\
&nbsp;&nbsp;<INPUT TYPE=submit NAME=act VALUE=\"%s\"></TD></TR>\n",
         lang_ProSignatureFile, SigFilePtr, lang_BtnAccess);

      free (SigFilePtr);
   }

   fprintf (stdout,
"<TR><TH ALIGN=right VALIGN=center>%s:&nbsp;&nbsp;</TH>\
<TD><INPUT TYPE=text SIZE=20 NAME=upq VALUE=\"%s\"></TD></TR>\n\
<TR><TH ALIGN=right VALIGN=center>%s:&nbsp;&nbsp;</TH>\
<TD><INPUT TYPE=text SIZE=20 NAME=upf VALUE=\"%s\"></TD></TR>\n\
<TR><TH ALIGN=right VALIGN=center>%s:&nbsp;&nbsp;</TH>\
<TD><INPUT TYPE=text SIZE=3 NAME=unm></TD></TR>\n\
</TABLE>\n\
<P><INPUT TYPE=submit NAME=act VALUE=\"%s\">&nbsp;\n\
<INPUT TYPE=reset VALUE=\"%s\">\n",
      lang_ProPrintQueue, QueuePtr,
      lang_ProPrintForm, FormPtr,
      lang_ProNewMessages,
      lang_BtnSet, lang_BtnReset);

   if (PostMasterAuthenticated)
      fprintf (stdout,
"<P><INPUT TYPE=submit NAME=act VALUE=\"%s\"> ... %s\n",
         lang_BtnDeleteUserProfile,
         FullDirectoryPtr[0] ? FullDirectoryPtr : "<I>(no directory set)</I>");

   fprintf (stdout,
"<P><INPUT TYPE=button VALUE=\"%s\" onClick=\"history.go(-1)\">\n\
</FORM>\n\
</BODY>\n\
</HTML>\n",
      lang_BtnGoBack);

   free (UserNamePtr);
   free (EditorPtr);
   free (FormPtr);
   free (ForwardingPtr);
   free (FullDirectoryPtr);
   free (PersonalNamePtr);
   free (QueuePtr);
}

/*****************************************************************************/
/*
Provide a mail message composition page, the contents of which will be form
POSTed to the facility for VMS Mail sending.  Include the contents of the
original message if a reply or forward (if requested).
*/

CreateMailMessage ()

{
   static $DESCRIPTOR (VmsDateFaoDsc, "!17%D\0");

   register char  *cptr, *sptr;

   int  status;
   char  *CopySelfPtr,
         *PersonalPtr,
         *ReFwPtr,
         *SenderPtr,
         *SigFileTextPtr,
         *SubjectPtr,
         *TextPtr;
   char  VmsDate [32];
   $DESCRIPTOR (VmsDateDsc, VmsDate);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CreateMailMessage()\n");

   if (!MailUserName[0])
   {
      CgiLibResponseError (FI_LI, 0, lang_ErrSanityCheck);
      return;
   }

   /* <TEXTAREA> width and automatic wrap at */
   cptr = CgiLibVar ("WWW_FORM_TAW");
   if (!cptr[0])
      WrapAt = 0;
   else
      WrapAt = atoi(cptr);
   /* let's keep it within the bounds of reason! */
   if (WrapAt && (WrapAt < 48 || WrapAt > 255)) WrapAt = 72;

   sys$fao (&VmsDateFaoDsc, 0, &VmsDateDsc, 0);

   if (VmsMailUserPersonalName[0])
      PersonalPtr = CgiLibHtmlEscape (VmsMailUserPersonalName, -1, NULL, 0);
   else
      PersonalPtr = "";

   CopySelfPtr = "";
   if (DoMessageReply)
   {
      SenderPtr = CgiLibHtmlEscape (VmsMailSender, -1, NULL, 0);
      SubjectPtr = CreateMailMessageReFw ("RE", VmsMailSubject);
      if (VmsMailUserCopyReply) CopySelfPtr = " CHECKED";
   }
   else
   if (DoMessageForward)
   {
      SenderPtr = "";
      SubjectPtr = CreateMailMessageReFw ("FW", VmsMailSubject);
      if (VmsMailUserCopyForward) CopySelfPtr = " CHECKED";
   }
   else
   if (DoMessageCreate)
   {
      SenderPtr = "";
      SubjectPtr = "";
      if (VmsMailUserCopySend) CopySelfPtr = " CHECKED";
   }

   SigFileTextPtr = "";
   if (VmsVersion >= 70 &&
       VmsMailUserSigFile[0])
   {
      ReadSigFile ();
      SigFileTextPtr = CgiLibHtmlEscape (VmsMailSigFileTextPtr, -1, NULL, 0);
   }

   if (VmsMailRfc822MimeCharset[0])
      CgiLibResponseHeader (200, "text/html",
                            "; charset=%s\n", VmsMailRfc822MimeCharset);
   else
      CgiLibResponseHeader (200, "text/html");

   fprintf (stdout,
"<HTML>\n\
<HEAD>\n\
<META NAME=\"generator\" CONTENT=\"%s\">\n\
<META NAME=\"environment\" CONTENT=\"%s\">\n\
<META NAME=\"language\" CONTENT=\"%s %s\">\n\
<TITLE>yahMAIL ... %s</TITLE>\n\
<SCRIPT LANGUAGE=\"JavaScript\">\n\
<!--\n\
var listURL = \"\";\n\
\
function setReload () {\n\
   document.cookie = \"YAHMAIL=RELOAD; path=%s/%c\"\n\
}\n\
\
function doSubmit () {\n\
   if (listURL == \"\") return true;\n\
   open (listURL, \"AddrList\", \"%s\");\n\
   listURL = \"\";\n\
   return false;\n\
}\n\
\
function doAddrList () {\n\
   listURL = document.formCreate.action + \"?\" + \"act=address%%20LISTS\";\n\
}\n\
\
// -->\n\
</SCRIPT>\n\
</HEAD>\n\
<BODY %s>\n\
%s\
<FORM METHOD=POST ACTION=\"%s/%c%s\" \
NAME=formCreate onSubmit=\"return doSubmit()\">\n\
<TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>\n\
<TR><TD HEIGHT=10></TD></TR>\n\
<TR><TH ALIGN=right VALIGN=center>%s:&nbsp;&nbsp;</TH><TD>%s</TD></TR>\n\
<TR><TD HEIGHT=5></TD></TR>\n",
      SoftwareID, CgiEnvironmentPtr, lang_Language, lang_LanguageAuthor,
      SubjectPtr[0] ? SubjectPtr : lang_CreateNew,
      CgiScriptNamePtr, TildeChar, ADDRESS_LIST_WINDOW,
      ConfigBodyTagPtr, PostMasterWarningPtr,
      CgiScriptNamePtr, TildeChar, MailUserName,
      lang_MsgDate, VmsDate);

   if (PostMasterAuthenticated)
   {
      fprintf (stdout,
"<TR><TH ALIGN=right VALIGN=center>%s:&nbsp;&nbsp;</TH>\
<TD><INPUT TYPE=text SIZE=60 NAME=sfr VALUE=\"%s %s%s%s\"></TD></TR>\n",
         lang_MsgFrom, MailUserName, PersonalPtr[0] ? "&quot;" : "",
         PersonalPtr, PersonalPtr[0] ? "&quot;" : "");
   }
   else
   {
      fprintf (stdout,
"<TR>\
<TH ALIGN=right VALIGN=center>%s:&nbsp;&nbsp;</TH>\
<TD>%s&nbsp;&nbsp;<INPUT TYPE=text SIZE=45 NAME=spn VALUE=\"%s\"></TD>\
</TR>\n\
<TR><TD HEIGHT=5></TD></TR>\n",
         lang_MsgFrom, MailUserName, PersonalPtr);
   }

   fprintf (stdout,
"<TR><TH></TH>\
<TR><TH ALIGN=right VALIGN=center>%s:&nbsp;&nbsp;</TH>\
<TD><TEXTAREA ROWS=3 COLS=60 NAME=sto>%s</TEXTAREA></TD>\
<TD>&nbsp;<INPUT TYPE=submit NAME=act VALUE=\"address LISTS\" \
onClick=\"return doAddrList()\"></TD></TR>\n\
<TR><TH ALIGN=right VALIGN=center>%s:&nbsp;&nbsp;</TH>\
<TD><TEXTAREA ROWS=2 COLS=60 NAME=scc></TEXTAREA></TD>\
<TD>&nbsp;<INPUT TYPE=checkbox NAME=scs VALUE=1%s>\
<I>&nbsp;%s</I></TD></TR>\n\
<TR><TH ALIGN=right VALIGN=center>%s:&nbsp;&nbsp;</TH>\
<TD><INPUT TYPE=text SIZE=60 NAME=ssu VALUE=\"%s\"></TD></TR>\n\
</TABLE>\n\
<INPUT TYPE=hidden NAME=wra VALUE=%d>\n\
<P><TEXTAREA NAME=sbd ROWS=16 COLS=%d WRAP=%s>\n",
      lang_MsgTo, SenderPtr,
      lang_MsgCc, CopySelfPtr, lang_CcSelf,
      lang_MsgSubj, SubjectPtr,
      WrapAt, WrapAt ? WrapAt : 72, WrapAt ? "on" : "off");

   CgiFormIncludePtr = CgiLibVar ("WWW_FORM_INC");
   if ((DoMessageReply || DoMessageForward) && CgiFormIncludePtr[0])
   {
      CgiFormQuotePtr = CgiLibVar ("WWW_FORM_QUO");

#ifdef YAHMAIL_MUNPACK

      if (VmsMailRfc822MimeVersion &&
          (!strsame (VmsMailRfc822MimeContentType, "text/plain", -1) ||
           !(VmsMailRfc822MimeContTransEnc[0] == '7' ||
             VmsMailRfc822MimeContTransEnc[0] == '8' ||
             strsame (VmsMailRfc822MimeContTransEnc, "binary", -1))))
      {
         /* not a simple plain-text MIME message */
         MunpackYahmailProcess (CgiFormQuotePtr[0]);
         MunpackYahmailText (stdout, false);

         /* cancel any include that might be done immediately below! */
         CgiFormIncludePtr = "";
      }

#endif /* YAHMAIL_MUNPACK */

      if (CgiFormIncludePtr[0])
      {
         cptr = CgiLibVar("WWW_FORM_HEA");
         if (cptr[0])
         {
            /* include the VMS (and consequently any RFC822) header */
            TextPtr = CgiLibHtmlEscape (VmsMailTextPtr, -1, NULL, 0);
         }
         else
         {
            cptr = CgiLibVar("WWW_FORM_RFC");
            if (cptr[0])
            {
               /* include from the RFC822 header */
               TextPtr = CgiLibHtmlEscape (VmsMailTextPtr +
                                           VmsMailHeaderLength,
                                           -1, NULL, 0);
            }
            else
            {
               /* exclude both VMS and RFC822 headers */
               TextPtr = CgiLibHtmlEscape (VmsMailTextPtr +
                                           VmsMailHeaderLength +
                                           VmsMailRfc822HeaderLength,
                                           -1, NULL, 0);
            }
         }

         cptr = TextPtr;
         while (*cptr)
         {
            sptr = cptr;
            while (*cptr && *cptr != '\n') cptr++;
            if (*cptr) cptr++;
            if (CgiFormQuotePtr[0]) fprintf (stdout, "&gt;");
            fwrite (sptr, cptr-sptr, 1, stdout);
         }
         free (TextPtr);
      }
   }

   if (SigFileTextPtr[0]) fputs (SigFileTextPtr, stdout);

   fprintf (stdout,
"</TEXTAREA>\n\
<P><INPUT TYPE=submit NAME=act VALUE=\"%s\" \
onClick=\"setReload()\">&nbsp;\n\
<INPUT TYPE=reset VALUE=\"%s\"><I> ... %s</I>\n\
<P><INPUT TYPE=button VALUE=\"%s\" onClick=\"history.go(-1)\">\n\
</FORM>\n\
</BODY>\n\
</HTML>\n",
      lang_BtnSend, lang_BtnReset, lang_ResetCaution,
      lang_BtnGoBack);

   if (DoMessageReply) free (SenderPtr);
   free (PersonalPtr);
   free (SubjectPtr);
   free (SigFileTextPtr);
}

/*****************************************************************************/
/*
Return a pointer to a "RE:", "RE[n]:", "FW:" or "FW[n]:", subject line as
appropriate.  The pointer should be free()ed after use.
*/

char* CreateMailMessageReFw
(
char *ReFwPtr,
char *SubjectPtr
)
{
   static $DESCRIPTOR (GtOneFaoDsc, "!AZ[!UL]: !AZ");
   static $DESCRIPTOR (OneFaoDsc, "!AZ: !AZ");

   register char  *cptr;

   int  InstanceCount;
   unsigned short  Length;
   char  ReFwSubject [256];
   $DESCRIPTOR (ReFwSubjectDsc, ReFwSubject);

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "CreateMailMessageReFw() |%s|%s|\n",
               ReFwPtr, SubjectPtr);

   InstanceCount = 1;
   for (cptr = SubjectPtr; *cptr && isspace(*SubjectPtr); SubjectPtr++);

   if ((strsame (ReFwPtr, "RE", -1) &&
        strsame (cptr, "RE:", 3) || strsame (cptr, "RE[", 3)) ||
       (strsame (ReFwPtr, "FW", -1) &&
        strsame (cptr, "FW:", 3) || strsame (cptr, "FW[", 3)))
   {
      cptr += 2;
      if (*cptr == ':')
      {
         cptr++;
         InstanceCount++;
      }
      else
      {
         if (*cptr == '[') cptr++;
         if (isdigit(*cptr)) InstanceCount = atoi(cptr) + 1;
         while (*cptr && isdigit(*cptr)) cptr++;
         if (*cptr == ']') cptr++;
         if (*cptr == ':') cptr++;
      }
   }

   while (isspace(*cptr)) cptr++;
   if (Debug) fprintf (stdout, "|%s| %d\n", cptr, InstanceCount);

   if (InstanceCount > 1)
      sys$fao (&GtOneFaoDsc, &Length, &ReFwSubjectDsc,
               ReFwPtr, InstanceCount, cptr);
   else
      sys$fao (&OneFaoDsc, &Length, &ReFwSubjectDsc,
               ReFwPtr, cptr);
   ReFwSubject[Length] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", ReFwSubject);

   cptr = CgiLibHtmlEscape (ReFwSubject, -1, NULL, 0);

   return (cptr);
}

/*****************************************************************************/
/*
Generate an HTML page listing details all the selected messages in the
specified folder.  Private folders have a unique number attached to them (the
"nnn=9999999" seen in query strings).  This exists to make each folder-related
request unique, forcing browsers to bypass any local caching when accessing
folders and messages.  Remember, messages to not have a unique identifier, it's
all relative to the folder and it current retinue of messages.

Use JavaScript to enhance behaviour and detect when folders are modified (via
deletes or moves) then they get reloaded.  This is fairly important as VMS mail
messages do not have a unique identifier, just a current number.  Redo an
action on that number and it may operate on an entirely different message.
*/

BrowseFolderPrivate ()

{
   static char  DateTime [32],
                NnnString [32];
   static $DESCRIPTOR (DateTimeDsc, DateTime);
   static $DESCRIPTOR (DateTimeFaoDsc, "!17%D\0");
   static $DESCRIPTOR (NnnStringDsc, NnnString);
   static $DESCRIPTOR (NnnStringFaoDsc, "!UW!UW!UW!UW!UW!UW\0");

   register struct MessageListStruct  *mlptr;
   register char  *cptr;

   boolean  NewMailBiff;
   int  IdCount,
        IdPage,
        IdFrom,
        IdTo,
        NewMailSeconds;
   unsigned long  BinTime [2];
   unsigned short  NumTime [7];
   char  *DatePtr,
         *FromPtr,
         *SubjectPtr;
   char  NewMailCheck [64],
         NoneString [64];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "BrowseFolderPrivate()\n");

   /* a "number" that makes each folder request unique */
   sys$gettim (&BinTime);
   sys$numtim (&NumTime, &BinTime);
   sys$fao (&NnnStringFaoDsc, 0, &NnnStringDsc,
            NumTime[1], NumTime[2], NumTime[3],
            NumTime[4], NumTime[5], NumTime[6]);

   if (!(PrivateSetupShow || PrivateSetupHide))
      PrivateSetupShow = atoi(CgiLibVar ("WWW_FORM_SPS"));

   BuildActionQueryString ();

   CgiHttpCookiePtr = CgiLibVar ("WWW_HTTP_COOKIE");
   NewMailSeconds = atol(CgiLibVar ("WWW_FORM_NMS"));

   NewMailBiff = false;
   if (CheckNewMailMinutes &&
       !strcmp (MailFolderName, "NEWMAIL"))
   {
      if ((mlptr = MessageListTailPtr) == NULL)
         NewMailSeconds = 0;
      else
      {
         if (decc$fix_time(&mlptr->BinaryDate) > NewMailSeconds &&
             strstr (CgiHttpCookiePtr, "YAHMAIL=NEWMAIL") != NULL)
            NewMailBiff = true;
         else
            NewMailSeconds = decc$fix_time(&mlptr->BinaryDate);
      }

      sprintf (NewMailCheck, " &nbsp;(%d %s)",
               CheckNewMailMinutes, lang_BrowseNewMailMinutes);
   }
   else
   {
      NewMailSeconds = 0;
      NewMailCheck[0] = '\0';
   }

   CgiLibResponseHeader (200, "text/html");

   fprintf (stdout,
"<HTML>\n\
<HEAD>\n\
<META NAME=\"generator\" CONTENT=\"%s\">\n\
<META NAME=\"environment\" CONTENT=\"%s\">\n\
<META NAME=\"language\" CONTENT=\"%s %s\">\n",
      SoftwareID, CgiEnvironmentPtr, lang_Language, lang_LanguageAuthor);

   if (NewMailBiff)
      fprintf (stdout, "<TITLE>%s</TITLE>\n", lang_NewMailBiffTitle);
   else
      fprintf (stdout, "<TITLE>yahMAIL ... %s / %s / %s%s</TITLE>\n",
               MailUserName, MailFileName, HtmlEscapedMailFolderName,
               NewMailCheck);

   fprintf (stdout,
"<SCRIPT LANGUAGE=\"JavaScript\">\n\
<!--\n\
\
var formURL = \"\",\n\
    listURL = \"\",\n\
    refreshNewMailMinutes = 0;\n\
\
function doOnLoad (folderName, refreshMinutes) {\n\
   if (document.cookie.indexOf(\"YAHMAIL=RELOAD\") != -1) {\n\
      resetCookie();\n\
      location.reload(true);\n\
   }\n\
   resetCookie();\n\
   if (folderName == \"NEWMAIL\" && refreshMinutes != 0) {\n\
      refreshNewMailMinutes = refreshMinutes;\n\
      refreshNewMailMinutes++;\n\
      refreshNewMail();\n\
   }\n\
}\n\
\
function refreshNewMail () {\n\
   if (--refreshNewMailMinutes > 0) {\n\
      window.defaultStatus = \"%s / %s / %s (\" + \
refreshNewMailMinutes + \" %s)\"\n\
      setTimeout(\"refreshNewMail()\", 60000);\n\
      return;\n\
   }\n\
   buildFormURL(\'NEWMAIL\');\n\
   document.cookie = \"YAHMAIL=NEWMAIL; path=%s/%c\";\n\
   location.replace(formURL);\n\
}\n\
\
function setReload (actionButton) {\n\
   var reload = false;\n\
   if (actionButton == \"%s\" ||\n\
       actionButton == \"%s\")\n\
      reload = true;\n\
   else {\n\
      for (idx = 0; idx < document.formPage.elements.length; idx++) {\n\
         if (document.formPage.elements[idx].type == \"checkbox\" &&\n\
             document.formPage.elements[idx].checked) {\n\
            elementName = document.formPage.elements[idx].name;\n\
            if (elementName.substring(0,2) == \"id\") reload = true;\n\
         }\n\
      }\n\
   }\n\
   if (reload)\n\
      document.cookie = \"YAHMAIL=RELOAD; path=%s/%c\";\n\
   else\n\
      buildFormURL(actionButton);\n\
}\n\
\
function resetCookie () {\n\
   document.cookie = \"YAHMAIL=; path=%s/%c\";\n\
}\n\
\
function doClearPick () {\n\
   document.formPage.pfr.value =\n\
   document.formPage.pto.value =\n\
   document.formPage.pcc.value =\n\
   document.formPage.psu.value = \"\";\n\
}\n\
\
function doSubmit () {\n\
   if (formURL != \"\") {\n\
      location.replace(formURL);\n\
      return false;\n\
   }\n\
   else\n\
   if (listURL != \"\") {\n\
      open (listURL, \"AddrList\", \"%s\");\n\
      listURL = \"\";\n\
      return false;\n\
   }\n\
   else {\n\
      return true;\n\
   }\n\
}\n\
\
function doRange (rangeURL) {\n\
   location.replace(rangeURL);\n\
   return false;\n\
}\n\
\
function doAddrList () {\n\
   listURL = document.formPage.action + \"?\" + \"act=address%%20LISTS\";\n\
}\n\
\
function buildFormURL (actionButton) {\n\
   var elementName, elementValue;\n\
   formURL = document.formPage.action + \"?act=\" + escape(actionButton);\n\
   for (idx = 0; idx < document.formPage.elements.length; idx++) {\n\
      if (document.formPage.elements[idx].type == \"hidden\" ||\n\
          document.formPage.elements[idx].type == \"text\") {\n\
         elementName = document.formPage.elements[idx].name;\n\
         elementValue = document.formPage.elements[idx].value;\n\
         formURL += \"&\" + elementName + \"=\" + escape(elementValue);\n\
      }\n\
      else\n\
      if (document.formPage.elements[idx].type == \"checkbox\") {\n\
         if (document.formPage.elements[idx].checked) {\n\
            elementName = document.formPage.elements[idx].name;\n\
            elementValue = document.formPage.elements[idx].value;\n\
            formURL += \"&\" + elementName + \"=\" + escape(elementValue);\n\
         }\n\
      }\n\
      else\n\
      if (document.formPage.elements[idx].type == \"select-one\") {\n\
         elementName = document.formPage.elements[idx].name;\n\
         elementValue = document.formPage.elements[idx].options\
[document.formPage.elements[idx].selectedIndex].value;\n\
         formURL += \"&\" + elementName + \"=\" + escape(elementValue);\n\
      }\n\
   }\n\
   resetCookie();\n\
}\n\
\
// -->\n\
</SCRIPT>\n\
</HEAD>\n\
<BODY onLoad=\"doOnLoad(\'%s\',\'%d\')\" %s>\n\
%s\
%s\n\
<FONT SIZE=+1><B><U>%s / %s / %s</U></B></FONT>%s",
      MailUserName, MailFileName, HtmlEscapedMailFolderName,
      lang_BrowseNewMailMinutes,
      CgiScriptNamePtr, TildeChar,
      lang_BtnEmptyWasteBasket, lang_BtnDelete,
      CgiScriptNamePtr, TildeChar,
      CgiScriptNamePtr, TildeChar,
      ADDRESS_LIST_WINDOW,
      HtmlEscapedMailFolderName, CheckNewMailMinutes, ConfigBodyTagPtr,
      PostMasterWarningPtr, ConfigHeaderPtr,
      MailUserName, MailFileName, HtmlEscapedMailFolderName, NewMailCheck);

   if (NewMailBiff)
      fputs (ConfigNewMailPtr == NULL ? lang_NewMailBiff :
                                        ConfigNewMailPtr, stdout);

   fprintf (stdout,
"\n<BR>&nbsp;<FONT SIZE=-1>yahMAIL<SUP>*</SUP></FONT>\n\
<FORM NAME=formPage ACTION=\"%s\" onSubmit=\"return doSubmit()\">\n\
<INPUT TYPE=hidden NAME=cmc VALUE=\"%d\">\n\
<INPUT TYPE=hidden NAME=anc VALUE=\"%d\">\n\
<INPUT TYPE=hidden NAME=wra VALUE=\"%d\">\n\
<INPUT TYPE=hidden NAME=nms VALUE=\"%d\">\n\
<P>",
      ActionPathInfo, VmsMailMessageSelectedCount,
      MakeAnchors, WrapAt, NewMailSeconds);
                                                                    
   BrowseFolderPrivateButtons ('1');

   IdPage = 1;
   IdFrom = VmsMailMessageSelectedCount;
   while (IdFrom > 0)
   {
      if (IdFrom > BrowseWindow)
         IdTo = IdFrom - BrowseWindow + 1;
      else
         IdTo = 1;

      if (IdPage == BrowsePage)
         fprintf (stdout, "[%04d-%04d]\n", IdFrom, IdTo);
      else
         fprintf (stdout,
"[<A HREF=\"%s%s&pag=%d&nnn=%s\" \
onClick=\"return doRange(\'%s%s&pag=%d&nnn=%s\')\">%04d-%04d</A>]\n",
            ActionPathInfo, ActionQueryString, IdPage, NnnString,
            ActionPathInfo, ActionQueryString, IdPage, NnnString,
            IdFrom, IdTo);

      IdFrom -= BrowseWindow;
      IdPage++;
   }

   sys$fao (&DateTimeFaoDsc, 0, &DateTimeDsc, 0);
   fprintf (stdout, "&nbsp;<NOBR>%s %s</NOBR>",
            lang_AsAtDateTime, DateTime);

   if (VmsMailUserNewMessages &&
       (!strcmp (MailFileName, "MAIL") ||
        !strncmp (MailFileName, "MAIL.", 5)) &&
        strcmp (MailFolderName, "NEWMAIL"))
      fprintf (stdout,
"&nbsp;&nbsp;<NOBR>%d %s %s <FONT SIZE=-1>NEWMAIL</FONT></NOBR>\n",
               VmsMailUserNewMessages,
               VmsMailUserNewMessages == 1 ? lang_MessagesInNewMailSingular :
                                             lang_MessagesInNewMailPlural,
               lang_MessagesInNewMail);

   if (VmsMailMessageSelectedCount) fprintf (stdout, "<BR>\n");

   fprintf (stdout, "<HR ALIGN=left WIDTH=70%% HEIGHT=1 NOSHADE>\n");

   IdCount = 1;
   for (mlptr = MessageListTailPtr; mlptr != NULL; mlptr = mlptr->PrevPtr)
   {
      if (Debug) fprintf (stdout, "id: %d\n", mlptr->MessageId);

      FromPtr = CgiLibHtmlEscape (mlptr->FromPtr, -1, NULL, 0);

      if (mlptr->SubjectPtr[0])
         SubjectPtr = CgiLibHtmlEscape (mlptr->SubjectPtr, -1, NULL, 0);
      else
         sprintf (SubjectPtr = NoneString, "<I>(%s)</I>", lang_None);

      /* ROWSPAN=3 here meshes quite nicely with RelatedList() COLSPAN=2 */
      fprintf (stdout,
"<TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>\n\
<TR>\
<TD ROWSPAN=4 VALIGN=top>\
<INPUT TYPE=checkbox NAME=id%d VALUE=%d>\
&nbsp;&nbsp;&nbsp;\
</TD><TD VALIGN=top>\
[<A HREF=\"%s%d%s&nnn=%s\">%04d</A>]\
&nbsp;&nbsp;&nbsp;\
</TD>\
<TH COLSPAN=2 ALIGN=left VALIGN=top><U>%s</U></TH>\
</TR>\n\
<TR><TD COLSPAN=2></TD><TD ALIGN=left VALIGN=top>%s</TD></TR>\n\
<TR><TD COLSPAN=2></TD><TD ALIGN=left VALIGN=top>\
<FONT SIZE=-1>%17.17s</FONT></TD></TR>\n",
         IdCount++, mlptr->MessageId,
         ActionPathInfo, mlptr->MessageId, ActionQueryString, NnnString,
         mlptr->MessageId,
         SubjectPtr, FromPtr, mlptr->DatePtr);

      RelatedList (mlptr->SubjectPtr);

      fprintf (stdout, "<TR><TD HEIGHT=5></TD></TR>\n</TABLE>\n");

      free (FromPtr);
      free (SubjectPtr);
   }

   if (!MessageListCount)
   {
      if (MailSubstringCc[0] || MailSubstringFrom[0] ||
          MailSubstringSubj[0] || MailSubstringTo[0])
         fprintf (stdout, "&nbsp;<I>(%s)</I>\n", lang_NoMessagesSelected);
      else
         fprintf (stdout, "&nbsp;<I>(%s)</I>\n", lang_NoMessages);
   }

   fprintf (stdout, "<HR ALIGN=left WIDTH=70%% HEIGHT=1 NOSHADE>\n");

   BrowseFolderPrivateButtons ('2');

   BrowseFolderPrivateSetup ();

   fprintf (stdout,
"<INPUT TYPE=hidden NAME=nnn VALUE=%s>\n\
</FORM>\n\
</NOBR>\n\
<P>%s\n\
</BODY>\n\
</HTML>\n",
      NnnString,
      ConfigFooterPtr);
}

/*****************************************************************************/
/*
Provide the private folder browse setup fields and buttons.
*/

BrowseFolderPrivateSetup ()

{
   static int  CheckValues[] = { 0, 1, 5, 10, 15, 30, 60, -1 };

   int  idx;
   char  *cptr,
         *CgiFormBwcPtr,
         *CgiFormBwfPtr,
         *CgiFormBwmPtr,
         *CgiFormBwpPtr,
         *CgiFormBwuPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "BrowseFolderPrivateSetup()\n");

   /* standard "mail" folder buttons (default to bottom) */
   cptr = CgiLibVar ("WWW_FORM_BWM");
   if (cptr[0] != '0' && cptr[0] != '1' && cptr[0] != '2') cptr = "2";
   CgiFormBwmPtr = cptr;

   /* "user" (all) folder buttons (default to none) */
   cptr = CgiLibVar ("WWW_FORM_BWU");
   if (cptr[0] != '0' && cptr[0] != '1' && cptr[0] != '2') cptr = "0";
   CgiFormBwuPtr = cptr;

   /* folder selector and open/move/copy buttons (default to bottom) */
   cptr = CgiLibVar ("WWW_FORM_BWF");
   if (cptr[0] != '0' && cptr[0] != '1' && cptr[0] != '2') cptr = "2";
   CgiFormBwfPtr = cptr;

   /* create-send button (default to bottom) */
   cptr = CgiLibVar ("WWW_FORM_BWC");
   if (cptr[0] != '0' && cptr[0] != '1' && cptr[0] != '2') cptr = "2";
   CgiFormBwcPtr = cptr;

   /* "pick" fields (default to none) */
   cptr = CgiLibVar ("WWW_FORM_BWP");
   if (cptr[0] != '0' && cptr[0] != '1' && cptr[0] != '2') cptr = "0";
   CgiFormBwpPtr = cptr;

   if (PrivateSetupShow)
   {
      /**********************/
      /* setup is displayed */
      /**********************/

      fprintf (stdout,
"<P><TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>\n\
<TR WIDTH=90%%><TD>&nbsp;</TD><TD><NOBR>\n");

      HtmlSelectBrowseWindow ();

      fprintf (stdout, "<BR><SELECT NAME=cnm>\n");

      for (idx = 0; CheckValues[idx] != -1; idx++)
      {
         if (CheckValues[idx] == 0)
            fprintf (stdout, "<OPTION VALUE=\"0\"%s>0 %s\n",
                     CheckNewMailMinutes ? "" : " SELECTED",
                     lang_BrowseNewMailDisabled);
         else
         if (CheckNewMailMinutes == CheckValues[idx])
            fprintf (stdout, "<OPTION VALUE=\"%d\" SELECTED>%d %s\n",
                     CheckValues[idx], CheckValues[idx],
                     lang_BrowseNewMailMinutes);
         else
            fprintf (stdout, "<OPTION VALUE=\"%d\">%d %s\n",
                     CheckValues[idx], CheckValues[idx],
                     lang_BrowseNewMailMinutes);

         if (CheckNewMailMinutes > CheckValues[idx] &&
             (CheckNewMailMinutes < CheckValues[idx+1] ||
              CheckValues[idx+1] == -1))
            fprintf (stdout, "<OPTION VALUE=\"%d\" SELECTED>%d %s\n",
                     CheckNewMailMinutes, CheckNewMailMinutes,
                     lang_BrowseNewMailMinutes);
      }

      fprintf (stdout, "</SELECT>&nbsp;<I>%s</I>\n", lang_BrowseNewMailCheck);

      fprintf (stdout,
"<BR><SELECT NAME=bwm>\n\
<OPTION VALUE=0%s>none\n\
<OPTION VALUE=1%s>top\n\
<OPTION VALUE=2%s>bottom\n\
</SELECT>&nbsp;<I>%s</I>\n",
         CgiFormBwmPtr[0] == '0' ? " SELECTED" : "",
         CgiFormBwmPtr[0] == '1' ? " SELECTED" : "",
         CgiFormBwmPtr[0] == '2' ? " SELECTED" : "",
         lang_BrowseMailFolderButtons);

      fprintf (stdout,
"<BR><SELECT NAME=bwu>\n\
<OPTION VALUE=0%s>none\n\
<OPTION VALUE=1%s>top\n\
<OPTION VALUE=2%s>bottom\n\
</SELECT>&nbsp;<I>%s</I>\n",
         CgiFormBwuPtr[0] == '0' ? " SELECTED" : "",
         CgiFormBwuPtr[0] == '1' ? " SELECTED" : "",
         CgiFormBwuPtr[0] == '2' ? " SELECTED" : "",
         lang_BrowseUserFolderButtons);

      fprintf (stdout,
"<BR><SELECT NAME=bwf>\n\
<OPTION VALUE=0%s>none\n\
<OPTION VALUE=1%s>top\n\
<OPTION VALUE=2%s>bottom\n\
</SELECT>&nbsp;<I>%s</I>\n",
         CgiFormBwfPtr[0] == '0' ? " SELECTED" : "",
         CgiFormBwfPtr[0] == '1' ? " SELECTED" : "",
         CgiFormBwfPtr[0] == '2' ? " SELECTED" : "",
         lang_BrowseFolderButtons);

      fprintf (stdout,
"<BR><SELECT NAME=bwc>\n\
<OPTION VALUE=1%s>top\n\
<OPTION VALUE=2%s>bottom\n\
</SELECT>&nbsp;<I>%s</I>\n",
         CgiFormBwcPtr[0] == '1' ? " SELECTED" : "",
         CgiFormBwcPtr[0] == '2' ? " SELECTED" : "",
         lang_BrowseCreateSendButton);

      fprintf (stdout,
"<BR><SELECT NAME=bwp>\n\
<OPTION VALUE=0%s>none\n\
<OPTION VALUE=1%s>top\n\
<OPTION VALUE=2%s>bottom\n\
</SELECT>&nbsp;<I>%s</I>\n",
         CgiFormBwpPtr[0] == '0' ? " SELECTED" : "",
         CgiFormBwpPtr[0] == '1' ? " SELECTED" : "",
         CgiFormBwpPtr[0] == '2' ? " SELECTED" : "",
         lang_BrowsePickFields);

      fprintf (stdout,
"</NOBR></TD></TR>\n\
</TABLE>\n\
<P><INPUT TYPE=submit NAME=act VALUE=\"%s\" \
onClick=\"return buildFormURL(\'%s\')\">\n\
<INPUT TYPE=hidden NAME=sps VALUE=1>\n",
               lang_BtnSetupHide, lang_BtnSetupHide);
   }
   else
   {
      /*******************/
      /* setup is hidden */
      /*******************/

      fprintf (stdout,
"<P><INPUT TYPE=submit NAME=act VALUE=\"%s\" \
onClick=\"return buildFormURL(\'%s\')\">\n\
<INPUT TYPE=hidden NAME=win VALUE=%d>\n\
<INPUT TYPE=hidden NAME=cnm VALUE=%d>\n\
<INPUT TYPE=hidden NAME=bwm VALUE=\"%s\">\n\
<INPUT TYPE=hidden NAME=bwu VALUE=\"%s\">\n\
<INPUT TYPE=hidden NAME=bwf VALUE=\"%s\">\n\
<INPUT TYPE=hidden NAME=bwc VALUE=\"%s\">\n\
<INPUT TYPE=hidden NAME=bwp VALUE=\"%s\">\n\
<INPUT TYPE=hidden NAME=sps VALUE=0>\n",
          lang_BtnSetupShow, lang_BtnSetupShow,
          BrowseWindow,
          CheckNewMailMinutes,
          CgiFormBwmPtr,
          CgiFormBwuPtr,
          CgiFormBwfPtr,
          CgiFormBwcPtr,
          CgiFormBwpPtr);
   }

   fprintf (stdout,
"<BR><INPUT TYPE=submit NAME=act VALUE=\"%s\" \
onClick=\"return doAddrList()\">\n\
<BR><INPUT TYPE=submit NAME=act VALUE=\"%s\">\n",
      lang_BtnAddressLists, lang_BtnChangeProfile);
}

/*****************************************************************************/
/*
Provide the private browse page appropriate group(s) of buttons, etc.,
depending on "setup" and  whether the function is being called from the top of
bottom of the message listing.
*/

BrowseFolderPrivateButtons (int TopOrBottom)

{
   register struct FolderListStruct  *flptr;

   boolean  DeleteButton;
   int  SectionCount;
   char  *cptr,
         *PickCcPtr,
         *PickFromPtr,
         *PickSubjPtr,
         *PickToPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "BrowseFolderPrivateButtons() %d\n", TopOrBottom);

   SectionCount = 0;
   DeleteButton = false;

   /* standard "mail" folder buttons (default to bottom) */
   cptr = CgiLibVar ("WWW_FORM_BWM");
   if (cptr[0] != '0' && cptr[0] != '1' && cptr[0] != '2') cptr = "2";
   if (cptr[0] == TopOrBottom)
   {
      if (!SectionCount++)
         fprintf (stdout,
"<P><TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>\n\
<TR WIDTH=90%%><TD>&nbsp;</TD><TD>\n");

      fprintf (stdout,
"<INPUT TYPE=submit NAME=act VALUE=\"%s\" \
onClick=\"setReload(\'%s\')\">&nbsp;\n\
<INPUT TYPE=submit NAME=act VALUE=\"%s\" \
onClick=\"setReload(\'%s\')\">&nbsp;\n\
<INPUT TYPE=submit NAME=act VALUE=\"%s\" \
onClick=\"setReload(\'%s\')\">&nbsp;\n\
<INPUT TYPE=submit NAME=act VALUE=\"%s\" \
onClick=\"setReload(\'%s\')\">\
<INPUT TYPE=submit NAME=act VALUE=\"%s\" \
onClick=\"setReload(\'%s\')\">&nbsp;\n\
<INPUT TYPE=submit NAME=act VALUE=\"%s\" \
onClick=\"setReload(\'%s\')\">\n",
         lang_BtnToNewMail, lang_BtnToNewMail,
         lang_BtnToMail, lang_BtnToMail,
         lang_BtnToSent, lang_BtnToSent,
         lang_BtnToWasteBasket, lang_BtnToWasteBasket,
         lang_BtnEmptyWasteBasket, lang_BtnEmptyWasteBasket,
         lang_BtnDelete, lang_BtnDelete);
      DeleteButton = true;
   }

   /* "user" (all) folder buttons (default to none) */
   cptr = CgiLibVar ("WWW_FORM_BWU");
   if (cptr[0] != '0' && cptr[0] != '1' && cptr[0] != '2') cptr = "0";
   if (cptr[0] == TopOrBottom &&
       FolderListHeadPtr != NULL)
   {
      if (!SectionCount++)
         fprintf (stdout,
"<P><TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>\n\
<TR WIDTH=90%%><TD>&nbsp;</TD><TD>\n");
      else
         fprintf (stdout, "<P>");

      for (flptr = FolderListHeadPtr; flptr != NULL; flptr = flptr->NextPtr)
      {
         cptr = CgiLibHtmlEscape (flptr->FolderNamePtr, -1, NULL, 0);
         fprintf (stdout,
"<INPUT TYPE=submit NAME=act VALUE=\"&quot;%s&quot;\" \
onClick=\"setReload(\'&quot;%s&quot;\')\">&nbsp;\n",
                  cptr, cptr);
         free (cptr);
      }
   }

   /* folder selector and open/move/copy buttons (default to bottom) */
   cptr = CgiLibVar ("WWW_FORM_BWF");
   if (cptr[0] != '0' && cptr[0] != '1' && cptr[0] != '2') cptr = "2";

   if (TopOrBottom == '1' && cptr[0] != '1' && cptr[0] != '2')
      fprintf (stdout, "<INPUT TYPE=hidden NAME=fse VALUE=\"%s\">\n",
               HtmlEscapedMailFolderName);
   else
   if (cptr[0] == TopOrBottom)
   {
      if (!SectionCount++)
         fprintf (stdout,
"<P><TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>\n\
<TR WIDTH=90%%><TD>&nbsp;</TD><TD>\n\
<NOBR>\n");
      else
         fprintf (stdout,
"<P>\
<NOBR>\n");

      fprintf (stdout, "<INPUT TYPE=text SIZE=15 NAME=fte>&nbsp;");

      FolderListHtmlSelect ();

      fprintf (stdout,
"&nbsp;<INPUT TYPE=submit NAME=act VALUE=\"%s\" \
onClick=\"buildFormURL(\'%s\')\">&nbsp;\n\
<INPUT TYPE=submit NAME=act VALUE=\"%s\" \
onClick=\"setReload(\'%s\')\">&nbsp;\n\
<INPUT TYPE=submit NAME=act VALUE=\"%s\" \
onClick=\"setReload(\'%s\')\">&nbsp;\n",
         lang_BtnOpen, lang_BtnOpen,
         lang_BtnMove, lang_BtnMove,
         lang_BtnCopy, lang_BtnCopy);

      if (!DeleteButton)
         fprintf (stdout,
"<INPUT TYPE=submit NAME=act VALUE=\"%s\" \
onClick=\"setReload(\'%s\')\">\n",
            lang_BtnDelete, lang_BtnDelete);

         fprintf (stdout, "</NOBR>\n");
   }

   /* create-send button (default to bottom) */
   cptr = CgiLibVar ("WWW_FORM_BWC");
   if (cptr[0] != '0' && cptr[0] != '1' && cptr[0] != '2') cptr = "2";
   if (cptr[0] == TopOrBottom)
   {
      if (!SectionCount++)
         fprintf (stdout,
"<P><TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>\n\
<TR WIDTH=90%%><TD>&nbsp;</TD><TD>\n");
      else
         fprintf (stdout, "<P>");

      fprintf (stdout,
"<INPUT TYPE=submit NAME=act VALUE=\"%s\">&nbsp;&nbsp;\
<INPUT TYPE=checkbox NAME=taw VALUE=%d%s>&nbsp;<I>%s</I>\n",
         lang_BtnCreateSend, 
         WrapAt ? WrapAt : 72, WrapAt ? " CHECKED" : "", lang_AutoWrap);
   }

   /* "pick" fields (default to none) */
   cptr = CgiLibVar ("WWW_FORM_BWP");
   if (cptr[0] != '0' && cptr[0] != '1' && cptr[0] != '2') cptr = "0";
   if (cptr[0] == TopOrBottom)
   {
      if (!SectionCount++)
         fprintf (stdout,
"<P><TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>\n\
<TR WIDTH=90%%><TD>&nbsp;</TD><TD>\n");
      else
         fprintf (stdout, "<P>");

      PickCcPtr = CgiLibHtmlEscape (MailSubstringCc, -1, NULL, 0);
      PickFromPtr = CgiLibHtmlEscape (MailSubstringFrom, -1, NULL, 0);
      PickSubjPtr = CgiLibHtmlEscape (MailSubstringSubj, -1, NULL, 0);
      PickToPtr = CgiLibHtmlEscape (MailSubstringTo, -1, NULL, 0);

      fprintf (stdout,
"<INPUT TYPE=text SIZE=20 NAME=psu VALUE=\"%s\">\
&nbsp;<I>%s &quot;%s:&quot;</I>&nbsp;&nbsp;&nbsp;\n\
<INPUT TYPE=button VALUE=\"%s\" onClick=\"doClearPick()\">\n\
<BR><INPUT TYPE=text SIZE=20 NAME=pfr VALUE=\"%s\">\
&nbsp;<I>%s &quot;%s:&quot;</I>\n\
<BR><INPUT TYPE=text SIZE=20 NAME=pto VALUE=\"%s\">\
&nbsp;<I>%s &quot;%s:&quot;</I>\n\
<BR><INPUT TYPE=text SIZE=20 NAME=pcc VALUE=\"%s\">\
&nbsp;<I>%s &quot;%s:&quot;</I>\n",
         PickSubjPtr, lang_BrowsePick, lang_MsgSubj, lang_BtnClear,
         PickFromPtr, lang_BrowsePick, lang_MsgFrom,
         PickToPtr, lang_BrowsePick, lang_MsgTo,
         PickCcPtr, lang_BrowsePick, lang_MsgCc);

      free (PickCcPtr);
      free (PickFromPtr);
      free (PickSubjPtr);
      free (PickToPtr);
   }
   else
   if (TopOrBottom == '2' && (cptr[0] == '0' || !cptr[0]))
   {
      /* provide empty "pick" fields for JavaScript munging */
      fprintf (stdout,
"<INPUT TYPE=hidden NAME=psu VALUE=\"\">\n\
<INPUT TYPE=hidden NAME=pfr VALUE=\"\">\n\
<INPUT TYPE=hidden NAME=pto VALUE=\"\">\n\
<INPUT TYPE=hidden NAME=pcc VALUE=\"\">\n");
   }

   if (SectionCount)
      fprintf (stdout,
"</TD></TR>\n\
</TABLE>\n\
<P>");
}

/*****************************************************************************/
/*
Generate an HTML page listing details all the selected messages in the
specified folder.
*/

BrowseFolderPublic ()

{
   register char  *cptr;
   register struct MessageListStruct  *mlptr;

   int  IdPage,
        IdFrom,
        IdTo,
        FolderCount;
   char  *DatePtr,
         *FromPtr,
         *PickFromPtr,
         *PickSubjPtr,
         *SubjectPtr;
   char  HtmlFolderName [256],
         NoneString [64];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "BrowseFolderPublic()\n");

   if (PublicTitlePtr == NULL)
      PublicTitlePtr = PublicUserName;

   /* in-folder message header overrides public, which overrides general */
   if (PublicHeaderPtr == NULL)
      PublicHeaderPtr = ConfigPublicHeaderPtr;
   if (PublicHeaderPtr == NULL)
      PublicHeaderPtr = ConfigHeaderPtr;

   /* in-folder message footer overrides public, which overrides general */
   if (PublicFooterPtr == NULL)
      PublicFooterPtr = ConfigPublicFooterPtr;
   if (PublicFooterPtr == NULL)
      PublicFooterPtr = ConfigFooterPtr;

   PickFromPtr = CgiLibHtmlEscape (MailSubstringFrom, -1, NULL, 0);
   PickSubjPtr = CgiLibHtmlEscape (MailSubstringSubj, -1, NULL, 0);

   BuildActionQueryString ();

   CgiLibResponseHeader (200, "text/html");
   fprintf (stdout,
"<HTML>\n\
<HEAD>\n\
<META NAME=\"generator\" CONTENT=\"%s\">\n\
<META NAME=\"environment\" CONTENT=\"%s\">\n\
<META NAME=\"language\" CONTENT=\"%s %s\">\n\
<TITLE>yahMAIL ... %s</TITLE>\n\
<SCRIPT LANGUAGE=\"JavaScript\">\n\
<!--\n\
\
var formURL = \"\";\n\
\
function doRange (rangeURL) {\n\
   location.replace(rangeURL);\n\
   return false;\n\
}\n\
\
function doClearPick () {\n\
   document.formPage.pfr.value =\n\
   document.formPage.psu.value = \"\";\n\
}\n\
\
function doSubmit () {\n\
   if (formURL == \"\") return true;\n\
   location.replace(formURL);\n\
   return false;\n\
}\n\
\
function buildFormURL (actionButton) {\n\
   var elementName, elementValue;\n\
   formURL = document.formPage.action + \"?act=\" + escape(actionButton);\n\
   for (idx = 0; idx < document.formPage.elements.length; idx++) {\n\
      if (document.formPage.elements[idx].type == \"hidden\" ||\n\
          document.formPage.elements[idx].type == \"text\") {\n\
         elementName = document.formPage.elements[idx].name;\n\
         elementValue = document.formPage.elements[idx].value;\n\
         formURL += \"&\" + elementName + \"=\" + escape(elementValue);\n\
      }\n\
      else\n\
      if (document.formPage.elements[idx].type == \"checkbox\") {\n\
         if (document.formPage.elements[idx].checked) {\n\
            elementName = document.formPage.elements[idx].name;\n\
            elementValue = document.formPage.elements[idx].value;\n\
            formURL += \"&\" + elementName + \"=\" + escape(elementValue);\n\
         }\n\
      }\n\
      else\n\
      if (document.formPage.elements[idx].type == \"select-one\") {\n\
         elementName = document.formPage.elements[idx].name;\n\
         elementValue = document.formPage.elements[idx].options\
[document.formPage.elements[idx].selectedIndex].value;\n\
         formURL += \"&\" + elementName + \"=\" + escape(elementValue);\n\
      }\n\
   }\n\
   resetCookie();\n\
}\n\
\
// -->\n\
</SCRIPT>\n\
</HEAD>\n\
<BODY %s>\n\
%s\
<FONT SIZE=+1><B><U>%s</U></B></FONT>\n\
<BR>&nbsp;<FONT SIZE=-1>yahMAIL<SUP>*</SUP></FONT>\n\
<FORM NAME=formPage ACTION=\"%s\" onSubmit=\"return doSubmit()\">\n\
<INPUT TYPE=hidden NAME=cmc VALUE=\"%d\">\n\
<INPUT TYPE=hidden NAME=anc VALUE=\"%d\">\n",
      SoftwareID, CgiEnvironmentPtr, lang_Language, lang_LanguageAuthor,
      PublicTitlePtr, ConfigBodyTagPtr, PublicHeaderPtr, PublicTitlePtr,
      ActionPathInfo, VmsMailMessageSelectedCount, MakeAnchors);

   IdPage = 1;
   IdFrom = VmsMailMessageSelectedCount - PublicConfigSubjCount;
   while (IdFrom > 0)
   {
      if (IdFrom > BrowseWindow)
         IdTo = IdFrom - BrowseWindow + 1;
      else
         IdTo = 1;

      if (IdPage == BrowsePage)
         fprintf (stdout, "[%04d-%04d]\n", IdFrom, IdTo);
      else
         fprintf (stdout,
"[<A HREF=\"%s%s&pag=%d\" \
onClick=\"return doRange(\'%s%s&pag=%d\')\">%04d-%04d</A>]\n",
            ActionPathInfo, ActionQueryString, IdPage,
            ActionPathInfo, ActionQueryString, IdPage,
            IdFrom, IdTo);

      IdFrom -= BrowseWindow;
      IdPage++;
   }
   if (VmsMailMessageSelectedCount - PublicConfigSubjCount)
      fprintf (stdout, "<BR>\n");

   fprintf (stdout, "<HR ALIGN=left WIDTH=70%% HEIGHT=1 NOSHADE>\n");

   for (mlptr = MessageListTailPtr; mlptr != NULL; mlptr = mlptr->PrevPtr)
   {
      if (Debug) fprintf (stdout, "id: %d\n", mlptr->MessageId);

      FromPtr = CgiLibHtmlEscape (mlptr->FromPtr, -1, NULL, 0);

      if (mlptr->SubjectPtr[0])
         SubjectPtr = CgiLibHtmlEscape (mlptr->SubjectPtr, -1, NULL, 0);
      else
         sprintf (SubjectPtr = NoneString, "<I>(%s)</I>", lang_None);

      /* ROWSPAN=3 here meshes quite nicely with RelatedList() COLSPAN=2 */
      fprintf (stdout,
"<TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>\n\
<TR>\
<TD ROWSPAN=3 VALIGN=top>[<A HREF=\"%s%d%s\">%04d</A>]&nbsp;&nbsp;&nbsp;</TD>\
<TH ALIGN=left VALIGN=top><U>%s</U></TH>\
</TR>\n\
<TR><TD ALIGN=left VALIGN=top>%s</TD></TR>\n\
<TR><TD ALIGN=left VALIGN=top>\
<FONT SIZE=-1>%17.17s</FONT></TD></TR>\n",
         ActionPathInfo, mlptr->MessageId, ActionQueryString, mlptr->MessageId,
         SubjectPtr, FromPtr, mlptr->DatePtr);

      RelatedList (mlptr->SubjectPtr);

      fprintf (stdout, "<TR><TD HEIGHT=5></TD></TR>\n</TABLE>\n");

      free (FromPtr);
      free (SubjectPtr);
   }

   if (!MessageListCount)
   {
      if (MailSubstringFrom[0] || MailSubstringSubj[0])
         fprintf (stdout, "&nbsp;<I>(%s)</I>\n", lang_NoMessagesSelected);
      else
         fprintf (stdout, "&nbsp;<I>(%s)</I>\n", lang_NoMessages);
   }

   fprintf (stdout, "<HR ALIGN=left WIDTH=70%% HEIGHT=1 NOSHADE>\n");

   if (FolderListCount)
   {
      fprintf (stdout, "<P>");

      FolderListHtmlSelect ();

      fprintf (stdout,
"<INPUT TYPE=submit NAME=act VALUE=\"%s\" \
onClick=\"buildFormURL(\'%s\')\">\n",
         lang_BtnOpen, lang_BtnOpen);
   }

   PickFromPtr = CgiLibHtmlEscape (MailSubstringFrom, -1, NULL, 0);
   PickSubjPtr = CgiLibHtmlEscape (MailSubstringSubj, -1, NULL, 0);

   fprintf (stdout,
"<P>&nbsp;<INPUT TYPE=text SIZE=20 NAME=psu VALUE=\"%s\">\
&nbsp;<I>%s &quot;%s:&quot;</I>&nbsp;&nbsp;&nbsp;\n\
<INPUT TYPE=button VALUE=\"%s\" onClick=\"doClearPick()\">\n\
<BR>&nbsp;<INPUT TYPE=text SIZE=20 NAME=pfr VALUE=\"%s\">\
&nbsp;<I>%s &quot;%s:&quot;</I>\n\
<BR>&nbsp;",
      PickSubjPtr, lang_BrowsePick, lang_MsgSubj, lang_BtnClear,
      PickFromPtr, lang_BrowsePick, lang_MsgFrom);

   HtmlSelectBrowseWindow ();

   if (!FolderListCount)
      fprintf (stdout,
"<P><INPUT TYPE=submit NAME=act VALUE=\"%s\" \
onClick=\"buildFormURL(\'%s\')\">\n",
         lang_BtnOpen, lang_BtnOpen);

   fprintf (stdout,
"</FORM>\n\
<P>%s\n\
</BODY>\n\
</HTML>\n",
      PublicFooterPtr);

   free (PickFromPtr);
   free (PickSubjPtr);
}

/*****************************************************************************/
/*
Display a list of (possible) replies and replied-to based on the content of the
original's subject field.   Replies are identified by a leading "RE:" or
"RE[n]:" in the reply's subject field followed by the same string.
*/

RelatedList (char *SubjectPtr)

{
   register char  *cptr;
   register struct MessageListStruct  *mlptr;

   int  ReplyCount;
   char  *FromPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "RelatedList() |%s|\n", SubjectPtr);

   ReplyCount = 0;
   while isspace((*SubjectPtr)) SubjectPtr++;

   for (mlptr = MessageListTailPtr; mlptr != NULL; mlptr = mlptr->PrevPtr)
   {
      /** if (Debug) fprintf (stdout, "|%s|\n", mlptr->SubjectPtr); **/
      for (cptr = mlptr->SubjectPtr; *cptr && isspace(*cptr); cptr++);
      if (!*cptr) continue;

      if (toupper(mlptr->SubjectPtr[0]) == 'R' &&
          toupper(mlptr->SubjectPtr[1]) == 'E' &&
          (mlptr->SubjectPtr[2] == ':' ||
           mlptr->SubjectPtr[2] == '[') ||
          strsame(SubjectPtr, mlptr->SubjectPtr, -1))
      {
         cptr = mlptr->SubjectPtr + 3;
         if (*cptr == ':')
            cptr++;
         else
         {
            if (*cptr == '[') cptr++;
            while (*cptr && isdigit(*cptr)) cptr++;
            if (*cptr == ']') cptr++;
         }
         while (isspace(*cptr)) cptr++;
         if (Debug) fprintf (stdout, "|%s|%s|\n", cptr, SubjectPtr);
         if (!SubjectPtr[0] || !strsame (cptr, SubjectPtr, -1)) continue;

         if (ReplyCount++)
            fprintf (stdout, "<BR>");
         else
         {
            fprintf (stdout,
"<TR><TD%s></TD><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>\n\
<TH ALIGN=left VALIGN=top>%s:&nbsp;</TH><TD ALIGN=left>\n",
               ConfigPrivate ? " COLSPAN=2" : "",
               lang_BrowseRelated);
         }

         FromPtr = CgiLibHtmlEscape (mlptr->FromPtr, -1, NULL, 0);

         fprintf (stdout, "[<A HREF=\"%s%d?%s&nnn=%s\">%04d</A>]&nbsp;%s\n",
                  ActionPathInfo, mlptr->MessageId, ActionQueryString,
                  CgiFormNnnPtr, mlptr->MessageId, FromPtr);

         free (FromPtr);
      }
   }

   if (toupper(SubjectPtr[0]) == 'R' &&
       toupper(SubjectPtr[1]) == 'E' &&
       (SubjectPtr[2] == ':' ||
        SubjectPtr[2] == '['))
   {
      cptr = SubjectPtr + 3;
      if (*cptr == ':')
         cptr++;
      else
      {
         if (*cptr == '[') cptr++;
         while (*cptr && isdigit(*cptr)) cptr++;
         if (*cptr == ']') cptr++;
      }
      while (isspace(*cptr)) cptr++;
      SubjectPtr = cptr;
      if (Debug) fprintf (stdout, "|%s|\n", SubjectPtr);

      for (mlptr = MessageListTailPtr; mlptr != NULL; mlptr = mlptr->PrevPtr)
      {
         /** if (Debug) fprintf (stdout, "|%s|\n", mlptr->SubjectPtr); **/
         for (cptr = mlptr->SubjectPtr; *cptr && isspace(*cptr); cptr++);
         if (!*cptr) continue;
   
         if (!SubjectPtr[0] || !strsame (cptr, SubjectPtr, -1)) continue;

         if (ReplyCount++)
            fprintf (stdout, "<BR>");
         else
         {
            fprintf (stdout,
"<TR><TD%s></TD><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>\n\
<TH ALIGN=left VALIGN=top>%s:&nbsp;</TH><TD ALIGN=left>\n",
                  ConfigPrivate ? " COLSPAN=2" : "",
                  lang_BrowseRelated);
         }

         FromPtr = CgiLibHtmlEscape (mlptr->FromPtr, -1, NULL, 0);

         fprintf (stdout, "[<A HREF=\"%s%d?%s&nnn=%s\">%04d</A>]&nbsp;%s\n",
                  ActionPathInfo, mlptr->MessageId, ActionQueryString,
                  CgiFormNnnPtr, mlptr->MessageId, FromPtr);

         free (FromPtr);
      }
   }

   if (ReplyCount)
      fprintf (stdout,
"</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n");
}

/*****************************************************************************/
/*
Provide an HTML selection list containing number of items per browse page.
*/

HtmlSelectBrowseWindow ()

{
   static int  WindowValues[] = { 5, 10, 15, 20, 25, 50, 100, 9999, -1 };

   int  idx;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HtmlSelectBrowseWindow()\n");

   fprintf (stdout, "<SELECT NAME=win>\n");

   for (idx = 0; WindowValues[idx] != -1; idx++)
   {
      if (WindowValues[idx] == 9999 && BrowseWindow == 9999)
         fprintf (stdout, "<OPTION VALUE=\"9999\" SELECTED>%s\n",
                  lang_BrowseAllItems);
      else
      if (WindowValues[idx] == 9999)
         fprintf (stdout, "<OPTION VALUE=\"9999\">%s\n",
                  lang_BrowseAllItems);
      else
      if (BrowseWindow >= WindowValues[idx] &&
          (BrowseWindow < WindowValues[idx+1] || WindowValues[idx+1] == -1))
         fprintf (stdout, "<OPTION VALUE=\"%d\" SELECTED>%d %s\n",
                          WindowValues[idx], WindowValues[idx],
                          lang_BrowseItem);
      else
         fprintf (stdout, "<OPTION VALUE=\"%d\">%d %s\n",
                          WindowValues[idx], WindowValues[idx],
                          lang_BrowseItem);
   }

   fprintf (stdout, "</SELECT>&nbsp;<I>%s</I>\n", lang_BrowseWindow);
}

/*****************************************************************************/
/*
Write a <SELECT><OPTION>...</SELECT> HTML structure to <stdout>, listing the
folders available in the current mailfile.  Make sure the current folder
('MailFolderName') occurs as the first entry in the list of options.  If folder
MAIL does not exist add a bogus, last entry for it.
*/

FolderListHtmlSelect ()

{
   register char  *cptr;
   register struct FolderListStruct  *flptr;

   boolean  FolderFound,
            MailFolderFound,
            NewmailFolderFound;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "FolderListHtmlSelect()\n");

   fprintf (stdout, "<SELECT NAME=fse>\n");

   FolderFound = MailFolderFound = NewmailFolderFound = false;
   for (flptr = FolderListHeadPtr; flptr != NULL; flptr = flptr->NextPtr)
   {
      if (Debug)
         fprintf (stdout, "|%s|%s|\n", MailFolderName, flptr->FolderNamePtr);

      if (!strcmp (flptr->FolderNamePtr, "MAIL"))
         MailFolderFound = true;
      else
      if (!strcmp (flptr->FolderNamePtr, "NEWMAIL"))
         NewmailFolderFound = true;

      cptr = CgiLibHtmlEscape (flptr->FolderNamePtr, -1, NULL, 0);

      if (!strcmp (flptr->FolderNamePtr, MailFolderName))
      {
         fprintf (stdout, "<OPTION VALUE=\"%s\" SELECTED>%s: %s\n",
                  cptr, lang_BrowseFolder, cptr);
         FolderFound = true;
      }
      else
         fprintf (stdout, "<OPTION VALUE=\"%s\">%s: %s\n",
                  cptr, lang_BrowseFolder, cptr);

      free (cptr);
   }

   if (ConfigPrivate)
   {
      if (!MailFolderFound)
         fprintf (stdout, "<OPTION VALUE=\"MAIL\"%s>%s: MAIL\n",
                 !strcmp (MailFolderName, "MAIL") ? " SELECTED" : "",
                 lang_BrowseFolder);

      if (!NewmailFolderFound)
         fprintf (stdout, "<OPTION VALUE=\"NEWMAIL\"%s>%s: NEWMAIL\n",
                  !strcmp (MailFolderName, "NEWMAIL") ? " SELECTED" : "",
                  lang_BrowseFolder);
   }

   fprintf (stdout, "</SELECT>\n");
}

/*****************************************************************************/
/*
Generate an HTML page listing all those in the user profile file.
*/

BrowseUserList ()

{
   register struct UserListStruct  *ulptr;

   char  LastAlphabetic;
   char  HtmlUserName [256],
         UrlUserName [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "BrowseUserList()\n");

   LastAlphabetic = '\0';

   CgiLibResponseHeader (200, "text/html");
   fprintf (stdout,
"<HTML>\n\
<HEAD>\n\
<META NAME=\"generator\" CONTENT=\"%s\">\n\
<META NAME=\"environment\" CONTENT=\"%s\">\n\
<META NAME=\"language\" CONTENT=\"%s %s\">\n\
<TITLE>yahMAIL ... %s</TITLE>\n\
</HEAD>\n\
<BODY %s>\n\
%s\
<B><FONT SIZE=+1><U>%s: &nbsp;%d %s</FONT></U></B>\n\
<BR>&nbsp;<FONT SIZE=-1>yahMAIL<SUP>*</SUP></FONT>\n\
<BLOCKQUOTE>",
      SoftwareID, CgiEnvironmentPtr, lang_Language, lang_LanguageAuthor,
      lang_ProUserProfileList, ConfigBodyTagPtr, PostMasterWarningPtr,
      lang_ProUserProfileList, UserListCount, lang_Entries);

   for (ulptr = UserListHeadPtr; ulptr != NULL; ulptr = ulptr->NextPtr)
   {
      CgiLibHtmlEscape (ulptr->UserNamePtr, -1,
                        HtmlUserName, sizeof(HtmlUserName));
      CgiLibUrlEncode (ulptr->UserNamePtr, -1,
                       UrlUserName, sizeof(UrlUserName));

      if (LastAlphabetic && (ulptr->UserNamePtr[0] == LastAlphabetic))
         fprintf (stdout, ",\n<A HREF=\"%s/%c%s?act=%s\">%s</A>",
                  CgiScriptNamePtr, TildeChar, UrlUserName,
                  lang_ProfileUser, HtmlUserName);
      else
         fprintf (stdout, "\n<P><A HREF=\"%s/%c%s?act=%s\">%s</A>",
                  CgiScriptNamePtr, TildeChar, UrlUserName,
                  lang_ProfileUser, HtmlUserName);
      LastAlphabetic = ulptr->UserNamePtr[0];
   }

   fprintf (stdout,
"\n\
</BLOCKQUOTE>\n\
</BODY>\n\
</HTML>\n");
}

/*****************************************************************************/
/*
Read the header information and body of the specified message.
*/

int MailReadMessage ()

{
   static unsigned long  MessageContext,
                         MessageSelected,
                         MessageRecordLength;

   static char  MessageRecord [255+2];

   static struct VmsItemStruct   MessageBeginItem [] = {
      { sizeof(MailFileContext), MAIL$_MESSAGE_FILE_CTX, &MailFileContext, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   static struct VmsItemStruct   MessageContinueItem [] = {
      { 0, MAIL$_MESSAGE_CONTINUE, 0, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   static struct VmsItemStruct   MessageIdItem [] = {
      { sizeof(MailMessageId[0]), MAIL$_MESSAGE_ID, &MailMessageId[0], 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   static struct VmsItemStruct   MessageInfoItem [] = {
      { sizeof(VmsMailExtId)-1, MAIL$_MESSAGE_EXTID,
        &VmsMailExtId, &VmsMailExtIdLength },
      { sizeof(VmsMailMessageId), MAIL$_MESSAGE_CURRENT_ID,
        &VmsMailMessageId, 0 },
      { sizeof(VmsMailMessageSize), MAIL$_MESSAGE_SIZE,
        &VmsMailMessageSize, 0 },
      { sizeof(VmsMailMessageFlags), MAIL$_MESSAGE_RETURN_FLAGS,
        &VmsMailMessageFlags, 0 },
      { sizeof(VmsMailCc)-1, MAIL$_MESSAGE_CC,
        VmsMailCc, &VmsMailCcLength },
      { sizeof(VmsMailDate)-1, MAIL$_MESSAGE_DATE,
        VmsMailDate, &VmsMailDateLength },
      { sizeof(VmsMailFrom)-1, MAIL$_MESSAGE_FROM,
        VmsMailFrom, &VmsMailFromLength },
      { sizeof(VmsMailSender)-1, MAIL$_MESSAGE_SENDER,
        VmsMailSender, &VmsMailSenderLength },
      { sizeof(VmsMailSubject)-1, MAIL$_MESSAGE_SUBJECT,
        VmsMailSubject, &VmsMailSubjectLength },
      { sizeof(VmsMailTo)-1, MAIL$_MESSAGE_TO,
        VmsMailTo, &VmsMailToLength },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   static struct VmsItemStruct   MessageNextItem [] = {
      { 0, MAIL$_MESSAGE_NEXT, 0, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   static struct VmsItemStruct   MessageRecordItem [] = {
      { sizeof(MessageRecord)-1, MAIL$_MESSAGE_RECORD,
        MessageRecord, &MessageRecordLength },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   static struct VmsItemStruct   MessagesSelectedItem [] = {
      { sizeof(VmsMailMessageSelectedCount), MAIL$_MESSAGE_SELECTED,
        &VmsMailMessageSelectedCount, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   register char  *cptr, *sptr;

   boolean  Rfc822From,
            Rfc822Subject,
            Rfc822To,
            Rfc822Received,
            Rfc822ReturnPath;
   int  status,
        LineCount;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MailReadMessage()\n");

   Rfc822From = Rfc822To = Rfc822Subject =
      Rfc822Received = Rfc822ReturnPath = false;
   VmsMailRfc822MimeCharset[0] = 
      VmsMailRfc822MimeContTransEnc[0] = 
      VmsMailRfc822MimeContentType[0] = '\0';
   MessageContext = VmsMailRfc822MimeVersion = 0;

   if (VmsMailTextPtr != NULL)
   {
      free (VmsMailTextPtr);
      VmsMailTextPtr = NULL;
   }
   VmsMailTextLength = VmsMailHeaderLength = VmsMailRfc822HeaderLength = 0;
   VmsMailTextPtr = calloc (1, MESSAGE_TEXT_CHUNK);
   VmsMailTextSize = MESSAGE_TEXT_CHUNK;

   status = mail$message_begin (&MessageContext,
                                &MessageBeginItem,
                                &NullItem);
   if (Debug) fprintf (stdout, "mail$message_begin() %%X%08.08X\n", status);
   if (VMSnok (status)) return (status);

   MailMessageSelect (&MessageContext);

   if (CurrentMessageCount != -1 &&
       CurrentMessageCount != VmsMailMessageSelectedCount)
   {
      /* just use this bogus status to indicate a message count disparity */
      status = SS$_POWERFAIL;
   }

   while (VMSok (status))
   {
      status = mail$message_get (&MessageContext,
                                 &MessageIdItem,
                                 &MessageInfoItem);
      if (VMSnok (status)) break;

      if (Debug)
         fprintf (stdout, "mail$message_get() %%X%08.08X %d\n",
                  status, VmsMailMessageId);
                  
      if (VMSnok (status)) break;

      VmsMailTo[VmsMailToLength] = '\0';
      VmsMailCc[VmsMailCcLength] = '\0';
      VmsMailDate[VmsMailDateLength] = '\0';
      if (isspace(VmsMailDate[0])) VmsMailDate[0] = '0';
      VmsMailExtId[VmsMailExtIdLength] = '\0';
      VmsMailSender[VmsMailSenderLength] = '\0';
      VmsMailSubject[VmsMailSubjectLength] = '\0';

      if (Debug)
         fprintf (stdout, "|%s|\n|%s|\n|%s|\n|%s|\n|%s|\n|%s|\n|%s|\n",
                  VmsMailDate, VmsMailTo, VmsMailCc, VmsMailFrom,
                  VmsMailSender, VmsMailSubject, VmsMailExtId);

      LineCount = 0;

      for (;;)
      {
         status = mail$message_get (&MessageContext,
                                    &MessageContinueItem,
                                    &MessageRecordItem);
/**
         if (Debug)
            fprintf (stdout, "mail$message_get() %%X%08.08X %d bytes\n",
                     status, MessageRecordLength);
**/
         if (VMSnok (status)) break;

         MessageRecord[MessageRecordLength++] = '\n';
         MessageRecord[MessageRecordLength] = '\0';
         /** if (Debug) fprintf (stdout, "|%s|\n", MessageRecord); **/

         if (VmsMailTextLength + MessageRecordLength > VmsMailTextSize)
         {
             VmsMailTextSize += MESSAGE_TEXT_CHUNK;
             VmsMailTextPtr = realloc (VmsMailTextPtr, VmsMailTextSize);
         }

         /* add in the date as if it's part of the VMS mail heading */
         if (!LineCount++)
            VmsMailTextLength +=
               sprintf (VmsMailTextPtr, "Date:\t%s\n", VmsMailDate);

         if (MessageRecordLength == 1 &&
             !VmsMailHeaderLength)
            VmsMailHeaderLength = VmsMailTextLength + 1;
         else
         if (MessageRecordLength == 1 &&
             VmsMailHeaderLength &&
             !VmsMailRfc822HeaderLength)
         {
            VmsMailRfc822HeaderLength = VmsMailTextLength -
                                        VmsMailHeaderLength + 1;
            if (Debug)
               fprintf (stdout, "rfc822: %d\n", VmsMailRfc822HeaderLength);
         }

         /* if haven't encountered the first empty line in the message body */
         if (VmsMailHeaderLength &&
             !VmsMailRfc822HeaderLength)
         {
            /* anything look a little RFC822? */
            if (strsame (MessageRecord, "From:", 5))
               Rfc822From = true;
            else
            if (strsame (MessageRecord, "Received:", 9))
               Rfc822Received = true;
            else
            if (strsame (MessageRecord, "Return-Path:", 12))
               Rfc822ReturnPath = true;
            else
            if (strsame (MessageRecord, "Subject:", 8))
               Rfc822Subject= true;
            else
            if (strsame (MessageRecord, "To:", 3))
               Rfc822To = true;
            else
            if (strsame (MessageRecord, "MIME-Version:", 13))
               VmsMailRfc822MimeVersion = 1;
            else
            if (strsame (MessageRecord, "Content-Transfer-Encoding:", 26))
            {
               /* MIME content-transfer-encoding field */
               cptr = MessageRecord + 26;
               while (*cptr && isspace(*cptr)) cptr++;
               sptr = VmsMailRfc822MimeContTransEnc;
               while (*cptr && *cptr != ';' && !isspace(*cptr))
                  *sptr++ = *cptr++;
               *sptr = '\0';
               if (Debug)
                  fprintf (stdout, "|%s|\n", VmsMailRfc822MimeContTransEnc);
            }
            else
            if (strsame (MessageRecord, "Content-Type:", 13))
            {
               /* MIME content-type field */

               cptr = MessageRecord + 13;
               while (*cptr && isspace(*cptr)) cptr++;
               sptr = VmsMailRfc822MimeContentType;
               while (*cptr && *cptr != ';' && !isspace(*cptr))
                  *sptr++ = *cptr++;
               *sptr = '\0';
               if (Debug)
                  fprintf (stdout, "|%s|\n", VmsMailRfc822MimeContentType);

               if ((cptr = strstr (MessageRecord, "charset=")) != NULL ||
                   (cptr = strstr (MessageRecord, "CHARSET=")) != NULL)
               {
                  cptr += 8;
                  sptr = VmsMailRfc822MimeCharset;
                  while (*cptr && !isspace(*cptr)) *sptr++ = *cptr++;
                  *sptr = '\0';
                  if (Debug)
                     fprintf (stdout, "|%s|\n", VmsMailRfc822MimeCharset);
               }
            }
         }

         /* copy the terminating null as well! */
         memcpy (VmsMailTextPtr+VmsMailTextLength,
                 MessageRecord, MessageRecordLength+1);

         VmsMailTextLength += MessageRecordLength;
      }

      if (Debug)
         fprintf (stdout, "%d %d %d %d |%s|\n",
                  VmsMailTextSize, VmsMailTextLength, VmsMailHeaderLength,
                  VmsMailRfc822HeaderLength, "" /** VmsMailTextPtr **/);

      if (!VmsMailHeaderLength) VmsMailHeaderLength = VmsMailTextLength;
   }

   /* no more records is just fine! */
   if (status == MAIL$_NOMOREREC) status = SS$_NORMAL;

   /* if it doesn't look like an RFC822 headered body then forget it! */
   if (VmsMailRfc822HeaderLength &&
       !(Rfc822From && Rfc822To && Rfc822Subject &&
         Rfc822Received && Rfc822ReturnPath))
   {
      VmsMailRfc822HeaderLength = VmsMailRfc822MimeVersion = 0;
      VmsMailRfc822MimeCharset[0] =
         VmsMailRfc822MimeContTransEnc[0] =
         VmsMailRfc822MimeContentType[0] = '\0';
   }

   mail$message_end (&MessageContext, &NullItem, &NullItem);

   return (status);
}

/*****************************************************************************/
/*
Generate a linked list containing details of all selected messages in the
selected folder.  Begin with the message ID calculated using 'BrowsePage' and
'BrowseWindow' into 'IdTo' (the least recent) and get up until 'IdFrom' (the
most recent).
*/

int GenerateMessageList ()

{
   static int  IdTo;
   static unsigned long  MessageContext,
                         MessageSelected;

   static struct VmsItemStruct   MessageBeginItem [] = {
      { sizeof(MailFileContext), MAIL$_MESSAGE_FILE_CTX, &MailFileContext, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   static struct VmsItemStruct   MessageInfoItem [] = {
      { sizeof(VmsMailMessageId), MAIL$_MESSAGE_CURRENT_ID,
        &VmsMailMessageId, 0 },
      { sizeof(VmsMailBinaryDate), MAIL$_MESSAGE_BINARY_DATE,
        &VmsMailBinaryDate, 0 },
      { sizeof(VmsMailDate)-1, MAIL$_MESSAGE_DATE,
        VmsMailDate, &VmsMailDateLength },
      { sizeof(VmsMailFrom)-1, MAIL$_MESSAGE_FROM,
        VmsMailFrom, &VmsMailFromLength },
      { sizeof(VmsMailSubject)-1, MAIL$_MESSAGE_SUBJECT,
        VmsMailSubject, &VmsMailSubjectLength },
      { sizeof(VmsMailTo)-1, MAIL$_MESSAGE_TO,
        VmsMailTo, &VmsMailToLength },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   static struct VmsItemStruct   MessageNextItem [] = {
      { 0, MAIL$_MESSAGE_NEXT, 0, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   /* least recent item of window */
   static struct VmsItemStruct   MessageFirstItem [] = {
      { sizeof(IdTo), MAIL$_MESSAGE_ID, &IdTo, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   static struct VmsItemStruct   MessagesSelectedItem [] = {
      { sizeof(VmsMailMessageSelectedCount), MAIL$_MESSAGE_SELECTED,
        &VmsMailMessageSelectedCount, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   int  status,
        CallocSize,
        IdFrom,
        IdPage,
        LineCount;
   struct MessageListStruct  *mlptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "GenerateMessageList()\n");

   if (MessageListHeadPtr != NULL)
   {
      /* empty any current list of messages */
      struct MessageListStruct  *meptr;
      mlptr = MessageListHeadPtr;
      while (mlptr != NULL)
      {
         meptr = mlptr;
         mlptr = mlptr->NextPtr;
         free (meptr);
      }
      MessageListHeadPtr = MessageListTailPtr = NULL;
      MessageListCount = 0;
   }

   MessageContext = PublicConfigSubjCount = 0;

   status = mail$message_begin (&MessageContext,
                                &MessageBeginItem,
                                &NullItem);
   if (Debug) fprintf (stdout, "mail$message_begin() %%X%08.08X\n", status);
   if (VMSnok (status)) return (status);

   MailMessageSelect (&MessageContext);

   /* after it is established how many messages we have to work with */
   IdPage = 1;
   IdFrom = VmsMailMessageSelectedCount;
   IdTo = IdFrom - BrowseWindow + 1;
   if (IdTo <= 0) IdTo = 1;
   while (IdPage != BrowsePage)
   {
      IdFrom -= BrowseWindow;
      IdTo = IdFrom - BrowseWindow + 1;
      IdPage++;
      if (IdTo <= 0)
      {
         IdTo = 1;
         break;
      }
   }

   /* start with the least recent of the browse window */
   status = mail$message_get (&MessageContext,
                              &MessageFirstItem,
                              &MessageInfoItem);
   if (Debug)
      fprintf (stdout, "mail$message_get() %%X%08.08X %d\n",
               status, VmsMailMessageId);

   while (VMSok (status))
   {
      if (Debug)
         fprintf (stdout, "mail$message_get() %%X%08.08X %d\n",
                  status, VmsMailMessageId);

      if (VMSnok (status)) break;

      VmsMailTo[VmsMailToLength] = '\0';
      VmsMailDate[VmsMailDateLength] = '\0';
      if (isspace(VmsMailDate[0])) VmsMailDate[0] = '0';
      VmsMailFrom[VmsMailFromLength] = '\0';
      VmsMailSubject[VmsMailSubjectLength] = '\0';

      if (Debug)
         fprintf (stdout, "|%s|\n|%s|\n|%s|\n|%s|\n",
                  VmsMailDate, VmsMailTo, VmsMailSender, VmsMailSubject);

      if (ConfigPublic &&
          strstr (VmsMailSubject, PublicConfigSubjPtr) != NULL)
      {
         /* ignore the YAHMAIL folder configuration message */
         PublicConfigSubjCount++;
      }
      else
      {
         /* add to message list */
         CallocSize = sizeof(struct MessageListStruct) + 
                      VmsMailDateLength + 1 +
                      VmsMailToLength + 1 +
                      VmsMailFromLength + 1 +
                      VmsMailSubjectLength + 1;

         mlptr = calloc (1, CallocSize);
         if (mlptr == NULL) exit (vaxc$errno);
         mlptr->NextPtr = NULL;
         mlptr->MessageId = VmsMailMessageId;
         memcpy (&mlptr->BinaryDate,
                 &VmsMailBinaryDate,
                 sizeof(mlptr->BinaryDate));
         mlptr->DatePtr = &mlptr->Data;
         strcpy (mlptr->DatePtr, VmsMailDate);
         mlptr->ToPtr = mlptr->DatePtr + VmsMailDateLength + 1;
         strcpy (mlptr->ToPtr, VmsMailTo);
         mlptr->FromPtr = mlptr->ToPtr + VmsMailToLength + 1;
         strcpy (mlptr->FromPtr, VmsMailFrom);
         mlptr->SubjectPtr = mlptr->FromPtr + VmsMailFromLength + 1;
         strcpy (mlptr->SubjectPtr, VmsMailSubject);

         if (MessageListHeadPtr == NULL)
         {
            MessageListHeadPtr = MessageListTailPtr = mlptr;
            mlptr->PrevPtr = mlptr->NextPtr = NULL;
         }
         else
         {
            mlptr->PrevPtr = MessageListTailPtr;
            mlptr->NextPtr = NULL;
            MessageListTailPtr->NextPtr = mlptr;
            MessageListTailPtr = mlptr;
         }
         MessageListCount++;
      }

      /* just did the most recent of the browse window */
      if (VmsMailMessageId == IdFrom) break;

      status = mail$message_get (&MessageContext,
                                 &MessageNextItem,
                                 &MessageInfoItem);
   }

   if (VMSnok (status))
   {
      /* cannot get this to work for the life of me!!! */
      /** if (status == MAIL$_NOMOREMSG) status = SS$_NORMAL; */
      if (status - MAIL$_NOMOREMSG == 0) status = SS$_NORMAL;
   }

   mail$message_end (&MessageContext, &NullItem, &NullItem);

   return (status);
}

/*****************************************************************************/
/*
Read the body of the special YAHMAIL configuration message.  Selects messages
that contain the special configuration subject string and then reads the latest
(highest message ID number) into the configuration text.
*/

int MailReadPublicConfigMessage ()

{
   static unsigned long  MessageContext,
                         MessageRecordLength;

   static char  MessageRecord [255+2];

   static char  *ConfigTextPtr;

   static struct VmsItemStruct   MessageBeginItem [] = {
      { sizeof(MailFileContext), MAIL$_MESSAGE_FILE_CTX, &MailFileContext, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   static struct VmsItemStruct   MessageContinueItem [] = {
      { 0, MAIL$_MESSAGE_CONTINUE, 0, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   static struct VmsItemStruct   MessageIdItem [] = {
      { sizeof(VmsMailMessageSelectedCount), MAIL$_MESSAGE_ID,
        &VmsMailMessageSelectedCount, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   static struct VmsItemStruct  MessageSelectItem [] = {
      { 0, MAIL$_MESSAGE_FOLDER, MailFolderName, 0 },
      { 0, MAIL$_MESSAGE_SUBJ_SUBSTRING, 0, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   static struct VmsItemStruct   MessageNextItem [] = {
      { 0, MAIL$_MESSAGE_NEXT, 0, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   static struct VmsItemStruct   MessageRecordItem [] = {
      { sizeof(MessageRecord)-1, MAIL$_MESSAGE_RECORD,
        MessageRecord, &MessageRecordLength },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   static struct VmsItemStruct   MessageSelectedItem [] = {
      { sizeof(VmsMailMessageSelectedCount), MAIL$_MESSAGE_SELECTED,
        &VmsMailMessageSelectedCount, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   register char  *bptr, *cptr, *sptr, *tptr;

   int  cnt,
        status,
        ConfigTextLength,
        ConfigTextSize;

   /*********/
   /* begin */
   /*********/
                                 
   if (Debug) fprintf (stdout, "MailReadPublicConfigMessage()\n");

   PublicHeaderPtr = PublicFooterPtr = PublicTitlePtr = NULL;

   MessageContext = 0;

   status = mail$message_begin (&MessageContext,
                                &MessageBeginItem,
                                &NullItem);
   if (Debug) fprintf (stdout, "mail$message_begin() %%X%08.08X\n", status);
   if (VMSnok (status)) return (status);

   MessageSelectItem[0].buf_len = MailFolderNameLength;
   MessageSelectItem[1].buf_len = PublicConfigSubjLength;
   MessageSelectItem[1].buf_addr = PublicConfigSubjPtr;

   status = mail$message_select (&MessageContext,
                                 &MessageSelectItem,
                                 &MessageSelectedItem);
   if (Debug)
      fprintf (stdout, "mail$message_select() %%X%08.08X %d\n",
               status, VmsMailMessageSelectedCount);

   while (VMSok (status))
   {
      /* always get the latest config message in the folder */
      status = mail$message_get (&MessageContext,
                                 &MessageIdItem,
                                 &NullItem);
      if (VMSnok (status)) break;

      if (Debug)
         fprintf (stdout, "mail$message_get() %%X%08.08X %d\n",
                  status, VmsMailMessageId);
                  
      if (VMSnok (status)) break;

      /************************/
      /* get the message text */
      /************************/

      if (ConfigTextPtr != NULL) free (ConfigTextPtr);
      ConfigTextLength = 0;
      ConfigTextPtr = calloc (1, CONFIG_TEXT_CHUNK);
      ConfigTextSize = CONFIG_TEXT_CHUNK;

      for (;;)
      {
         status = mail$message_get (&MessageContext,
                                    &MessageContinueItem,
                                    &MessageRecordItem);
         if (Debug)
            fprintf (stdout, "mail$message_get() %%X%08.08X %d bytes\n",
                     status, MessageRecordLength);
         if (VMSnok (status)) break;

         MessageRecord[MessageRecordLength++] = '\n';
         MessageRecord[MessageRecordLength] = '\0';
         if (Debug) fprintf (stdout, "|%s|\n", MessageRecord);

         if (ConfigTextLength+MessageRecordLength > ConfigTextSize)
         {
             ConfigTextSize += CONFIG_TEXT_CHUNK;
             ConfigTextPtr = realloc (ConfigTextPtr, ConfigTextSize);
         }

         /* copy the terminating null as well! */
         memcpy (ConfigTextPtr+ConfigTextLength,
                 MessageRecord, MessageRecordLength+1);

         ConfigTextLength += MessageRecordLength;
      }

      if (Debug)
         fprintf (stdout, "%d %d |%s|\n",
                  ConfigTextSize, ConfigTextLength, ConfigTextPtr);

      /*********************************/
      /* process configuration message */
      /*********************************/

      cptr = ConfigTextPtr;
      while (*cptr)
      {
         if (strsame (cptr, bptr = "[body]", cnt = 6) ||
             strsame (cptr, bptr = "[header]", cnt = 8) ||
             strsame (cptr, bptr = "[footer]", cnt = 8) ||
             strsame (cptr, bptr = "[title]", cnt = 7))
         {
            /* find the start of the next directive or end of text */
            sptr = (cptr += cnt);
            while (*cptr && *(unsigned short*)cptr != '\n[') cptr++;
            if (*cptr) *cptr++ = '\0';

            /* trim leading then trailing white space */
            while (*sptr && isspace(*sptr)) sptr++;
            tptr = sptr;
            while (*sptr) sptr++;
            if (sptr > tptr) sptr--;
            while (sptr > tptr && isspace(*sptr)) sptr--;
            if (sptr > tptr) *++sptr = '\0';

            /* set the appropriate pointer */
            if (strsame (bptr, "[body]", 6))
               ConfigBodyTagPtr = tptr;
            else
            if (strsame (bptr, "[header]", 8))
               PublicHeaderPtr = tptr;
            else
            if (strsame (bptr, "[footer]", 8))
               PublicFooterPtr = tptr;
            else
            if (strsame (bptr, "[title]", 7))
               PublicTitlePtr = tptr;
         }
         else
         {
            /* find the start of the next line */
            while (*cptr && *cptr != '\n') cptr++;
            if (*cptr) cptr++;
         }
      }

      if (Debug)
         fprintf (stdout, "|%s|\n|%s|\n|%s|\n|%s|\n",
                  ConfigBodyTagPtr, PublicTitlePtr,
                  PublicHeaderPtr, PublicFooterPtr);
   }

   /* no more records is just fine! */
   if (status == MAIL$_NOMOREREC) status = SS$_NORMAL;

   mail$message_end (&MessageContext, &NullItem, &NullItem);

   return (status);
}

/*****************************************************************************/
/*
Delete one or more mail messages.  Folder name is specified by 'MailFolderName'
and messages IDs in 'MailMessageId[0..n]'.
*/

int DeleteMessage ()

{
   static unsigned long  MessageContext,
                         MessageSelected;

   static struct VmsItemStruct   MessageBeginItem [] = {
      { sizeof(MailFileContext), MAIL$_MESSAGE_FILE_CTX, &MailFileContext, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   static struct VmsItemStruct   MessageDeleteItem [] = {
      { sizeof(unsigned long), MAIL$_MESSAGE_ID, 0, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   int  status,
        MessageIdCount;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DeleteMessage()\n");

   MessageContext = 0;

   status = mail$message_begin (&MessageContext,
                                &MessageBeginItem,
                                &NullItem);
   if (Debug) fprintf (stdout, "mail$message_begin() %%X%08.08X\n", status);
   if (VMSnok (status)) return (status);

   MailMessageSelect (&MessageContext);

   if (CurrentMessageCount != -1 &&
       CurrentMessageCount != VmsMailMessageSelectedCount)
   {
      /* just use this bogus status to indicate a message count disparity */
      status = SS$_POWERFAIL;
   }

   MessageIdCount = 0;

   while (VMSok (status))
   {
      if (!MailMessageId[MessageIdCount]) break;
      if (Debug) fprintf (stdout, "id %d\n", MailMessageId[MessageIdCount]);
      MessageDeleteItem[0].buf_addr = &MailMessageId[MessageIdCount];
      MessageIdCount++;

      status = mail$message_delete (&MessageContext,
                                    &MessageDeleteItem,
                                    &NullItem);
      if (Debug)
         fprintf (stdout, "mail$message_delete() %%X%08.08X\n", status);

      if (VMSnok (status)) break;
   }

   mail$message_end (&MessageContext, &NullItem, &NullItem);

   if (VMSok (status))
   {
      /* if the NEWMAIL folder was involved then adjust new message count */
      if (!strcmp (MailFileName, "MAIL") ||
          !strncmp (MailFileName, "MAIL.", 5))
      {
         if (!strcmp (MailFolderName, "NEWMAIL"))
            MailSetUserNewMessages (-MessageIdCount);
      }
   }

   return (status);
}

/*****************************************************************************/
/*
Purge the WASTEBASKET folder.
*/

int EmptyWasteBasket ()

{
   static struct VmsItemStruct   MailFileMessagesDeletedItem [] = {
      { sizeof(MailFileMessagesDeleted), MAIL$_MAILFILE_MESSAGES_DELETED,
        &MailFileMessagesDeleted, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "EmptyWasteBasket()\n");

   MailFileMessagesDeleted = 0;

   status = mail$mailfile_purge_waste (&MailFileContext,
                                       &NullItem,
                                       &MailFileMessagesDeletedItem);
   if (Debug) fprintf (stdout, "mail$message_select() %%X%08.08X\n", status);

   return (status);
}

/*****************************************************************************/
/*
Copy or move (copy with delete from original) one or more mail messages. 
Source folder name is specified by  'MailFolderName', destination folder name
by 'CopyFolderName' and message IDs in 'MailMessageId[0..n]'.
*/

int CopyMoveMessage ()

{
   static unsigned long  MessageContext,
                         MessageSelected;

   static struct VmsItemStruct   MessageBeginItem [] = {
      { sizeof(MailFileContext), MAIL$_MESSAGE_FILE_CTX, &MailFileContext, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   static struct VmsItemStruct   MessageCopyItem [] = {
      { sizeof(unsigned long), MAIL$_MESSAGE_ID, 0, 0 },
      { 0, MAIL$_MESSAGE_FOLDER, 0, 0 },
      { 0,0,0,0 }
   };

   static struct VmsItemStruct   MessageMoveItem [] = {
      { sizeof(unsigned long), MAIL$_MESSAGE_ID, 0, 0 },
      { 0, MAIL$_MESSAGE_FOLDER, 0, 0 },
      { 0, MAIL$_MESSAGE_DELETE, 0, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   int  status,
        MessageIdCount;
   struct VmsItemStruct   *ItemPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "CopyMoveMessage() |%s|->|%s|\n",
               MailFolderName, CopyFolderName);


   MessageContext = 0;

   status = mail$message_begin (&MessageContext,
                                &MessageBeginItem,
                                &NullItem);
   if (Debug) fprintf (stdout, "mail$message_begin() %%X%08.08X\n", status);
   if (VMSnok (status)) return (status);

   MailMessageSelect (&MessageContext);

   if (CurrentMessageCount != -1 &&
       CurrentMessageCount != VmsMailMessageSelectedCount)
   {
      /* just use this bogus status to indicate a message count disparity */
      status = SS$_POWERFAIL;
   }

   if (DoMessageCopy)
      ItemPtr = &MessageCopyItem;
   else
      ItemPtr = &MessageMoveItem;

   ItemPtr[1].buf_addr = CopyFolderName;
   ItemPtr[1].buf_len = CopyFolderNameLength;

   MessageIdCount = 0;

   while (VMSok (status))
   {
      if (!MailMessageId[MessageIdCount]) break;
      if (Debug) fprintf (stdout, "id %d\n", MailMessageId[MessageIdCount]);
      ItemPtr[0].buf_addr = &MailMessageId[MessageIdCount];
      MessageIdCount++;

      status = mail$message_copy (&MessageContext,
                                  ItemPtr,
                                  &NullItem);
      if (Debug) fprintf (stdout, "mail$message_copy() %%X%08.08X\n", status);

      if (VMSnok (status)) break;
   }

   mail$message_end (&MessageContext, &NullItem, &NullItem);

   if (VMSok (status))
   {
      /* if the NEWMAIL folder was involved then adjust new message count */
      if (!strcmp (MailFileName, "MAIL") ||
          !strncmp (MailFileName, "MAIL.", 5))
      {
         if (!strcmp (MailFolderName, "NEWMAIL"))
            MailSetUserNewMessages (-MessageIdCount);
         else
         if (!strcmp (CopyFolderName, "NEWMAIL"))
            MailSetUserNewMessages (MessageIdCount);
      }
   }

   return (status);
}

/*****************************************************************************/
/*
Generate a linked list of folder names by calling GenerateFolderListEntry() for
each folder in the current mail file.
*/

int GenerateFolderList ()

{
   static unsigned long  MessageContext;

   static struct VmsItemStruct   MessageBeginItem [] = {
      { sizeof(MailFileContext), MAIL$_MESSAGE_FILE_CTX, &MailFileContext, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   static struct VmsItemStruct   MessageFolderRoutineItem [] = {
      { sizeof(void*), MAIL$_MAILFILE_FOLDER_ROUTINE,
        GenerateFolderListEntry, 0 },
      { sizeof(void*), MAIL$_MAILFILE_USER_DATA, &MessageContext, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "GenerateFolderList()\n");

   if (FolderListHeadPtr != NULL)
   {
      /* empty any current list of folders */
      struct FolderListStruct  *flptr, *feptr;
      flptr = FolderListHeadPtr;
      while (flptr != NULL)
      {
         feptr = flptr;
         flptr = flptr->NextPtr;
         free (feptr);
      }
      FolderListHeadPtr = FolderListTailPtr = NULL;
      FolderListCount = 0;
   }

   MessageContext = 0;

   status = mail$mailfile_info_file (&MailFileContext,
                                     &MessageFolderRoutineItem,
                                     &NullItem);
   if (Debug)
      fprintf (stdout, "mail$message_select() %%X%08.08X\n",
               status);

   mail$message_end (&MessageContext, &NullItem, &NullItem);

   return (status);
}

/*****************************************************************************/
/*
Called once from GenerateFolderList() for each folder name in the current mail
file.  Put the name into a linked list.
*/

int GenerateFolderListEntry
(
unsigned long UserData,
struct dsc$descriptor *FolderDscPtr
)
{
   register struct FolderListStruct  *flptr;

   int  status,
        CallocSize;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "GenerateFolderListEntry() %d\n",
               FolderDscPtr->dsc$w_length);

   if (!FolderDscPtr->dsc$w_length) return (SS$_NORMAL);

   strncpy (VmsMailFolder, FolderDscPtr->dsc$a_pointer,
                           FolderDscPtr->dsc$w_length);
   VmsMailFolder[FolderDscPtr->dsc$w_length] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", VmsMailFolder);

   CallocSize = sizeof(struct FolderListStruct) + 
                FolderDscPtr->dsc$w_length + 1;
   flptr = calloc (1, CallocSize);
   if (flptr == NULL) exit (vaxc$errno);
   flptr->NextPtr = NULL;
   flptr->FolderNamePtr = &flptr->Data;
   strcpy (flptr->FolderNamePtr, VmsMailFolder);
   if (FolderListHeadPtr == NULL)
      FolderListHeadPtr = FolderListTailPtr = flptr;
   else
   {
      FolderListTailPtr->NextPtr = flptr;
      FolderListTailPtr = flptr;
   }
   FolderListCount++;

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Open the Mail file in 'MailFileName'.  If this contains a ':' then consider it
a full VMS file specification (e.g. USER_DISK:[ME.MAIL]TEST.MAI), otherwise a
mail file based in the current Mail directory (e.g. MAIL).  The ".MAI"
extension defaults and so is not needed.
*/

int MailFileOpen ()

{
   static struct VmsItemStruct   MailFileNameItem [] = {
      { 0, MAIL$_MAILFILE_NAME, MailFileName, &MailFileNameLength },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   int  status;
   char  Scratch [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MailFileOpen()\n");

   UserContext = 0;

   MailFileContext = 0;

   status = mail$mailfile_begin (&MailFileContext, &NullItem, &NullItem);
   if (Debug) fprintf (stdout, "mail$mailfile_begin() %%X%08.08X\n", status);
   if (VMSnok (status)) return (status);

   if (strchr (MailFileName, ':') == NULL)
   {
      strcpy (Scratch, VmsMailUserFullDirectory);
      strcat (Scratch, MailFileName);
      MailFileNameItem[0].buf_addr = Scratch;
      MailFileNameItem[0].buf_len = strlen(Scratch);
   }
   else
   {
      MailFileNameItem[0].buf_addr = MailFileName;
      MailFileNameItem[0].buf_len = MailFileNameLength;
   }
   if (Debug)
      fprintf (stdout, "MailFileNameItem[0].buf_addr |%s|\n",
               MailFileNameItem[0].buf_addr);

   status = mail$mailfile_open (&MailFileContext,
                                &MailFileNameItem,
                                &NullItem);
   if (Debug) fprintf (stdout, "mail$mailfile_open() %%X%08.08X\n", status);
                            
   return (status);
}

/*****************************************************************************/
/*
Close the currently open mail file.
*/ 

MailFileClose ()

{
   static unsigned long  MailFileTotalReclaimBytes;

   static struct VmsItemStruct   MailFileFullCloseItem [] = {
      { 0, MAIL$_MAILFILE_FULL_CLOSE, 0, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   static struct VmsItemStruct   MailFileTotalReclaimItem [] = {
      { sizeof(MailFileTotalReclaimBytes), MAIL$_MAILFILE_TOTAL_RECLAIM,
        &MailFileTotalReclaimBytes, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MailFileClose()\n");

   MailFileTotalReclaimBytes = 0;

/**
   status = mail$mailfile_close (&MailFileContext,
                                 &MailFileFullCloseItem,
                                 &MailFileTotalReclaimItem);
**/

   status = mail$mailfile_close (&MailFileContext, &NullItem, &NullItem);
   if (Debug) fprintf (stdout, "mail$mailfile_close() %%X%08.08X\n", status);
   if (VMSnok (status)) return (status);

   status = mail$mailfile_end (&MailFileContext, &NullItem, &NullItem);
   if (Debug) fprintf (stdout, "mail$mailfile_end() %%X%08.08X\n", status);
   return (status);
}

/*****************************************************************************/
/*
Select messages from the currently open Mail file, in the 'MailFolderName'
folder, based on the contents of 'MailSubstringCc', 'MailSubstringFrom',
'MailSubstringSubj', and 'MailSubstringTo'.
*/

int MailMessageSelect (unsigned long *MessageContextPtr)

{
   static struct VmsItemStruct   MessageSelectItem [7];

   static struct VmsItemStruct   MessageSelectedItem [] = {
      { sizeof(VmsMailMessageSelectedCount), MAIL$_MESSAGE_SELECTED,
        &VmsMailMessageSelectedCount, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   int  idx,
        status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MailMessageSelect() |%s|\n", MailFolderName);

   memset (&MessageSelectItem[idx=0], 0, sizeof(MessageSelectItem));

   MessageSelectItem[idx].item = MAIL$_MESSAGE_FOLDER;
   MessageSelectItem[idx].buf_addr = MailFolderName;
   MessageSelectItem[idx].buf_len = MailFolderNameLength;

   if (MailSubstringCcLength)
   {
      MessageSelectItem[++idx].item = MAIL$_MESSAGE_CC_SUBSTRING;
      MessageSelectItem[idx].buf_addr = MailSubstringCc;
      MessageSelectItem[idx].buf_len = MailSubstringCcLength;
   }

   if (MailSubstringFromLength)
   {
      MessageSelectItem[++idx].item = MAIL$_MESSAGE_FROM_SUBSTRING;
      MessageSelectItem[idx].buf_addr = MailSubstringFrom;
      MessageSelectItem[idx].buf_len = MailSubstringFromLength;
   }

   if (MailSubstringSubjLength)
   {
      MessageSelectItem[++idx].item = MAIL$_MESSAGE_SUBJ_SUBSTRING;
      MessageSelectItem[idx].buf_addr = MailSubstringSubj;
      MessageSelectItem[idx].buf_len = MailSubstringSubjLength;
   }

   if (MailSubstringToLength)
   {
      MessageSelectItem[++idx].item = MAIL$_MESSAGE_TO_SUBSTRING;
      MessageSelectItem[idx].buf_addr = MailSubstringTo;
      MessageSelectItem[idx].buf_len = MailSubstringToLength;
   }

   MessageSelectItem[++idx].item = MAIL$_NOSIGNAL;

   status = mail$message_select (MessageContextPtr,
                                 &MessageSelectItem,
                                 &MessageSelectedItem);
   if (Debug)
      fprintf (stdout, "mail$message_select() %%X%08.08X selected: %d\n",
               status, VmsMailMessageSelectedCount);

   return (status);
}

/*****************************************************************************/
/*
This function retieves user profile information for the user name specified by
the global storage 'MailUserName'.
*/

int MailGetUserInfo ()

{
   static struct VmsItemStruct   UserDetailsItem [] = {
      { sizeof(VmsMailUserNewMessages), MAIL$_USER_NEW_MESSAGES,
        &VmsMailUserNewMessages, 0 },
      { sizeof(VmsMailUserAutoPurge), MAIL$_USER_AUTO_PURGE,
        &VmsMailUserAutoPurge, 0 },
      { sizeof(VmsMailUserCcPrompt), MAIL$_USER_CC_PROMPT,
        &VmsMailUserCcPrompt, 0 },
      { sizeof(VmsMailUserCopyForward), MAIL$_USER_COPY_FORWARD,
        &VmsMailUserCopyForward, 0 },
      { sizeof(VmsMailUserCopyReply), MAIL$_USER_COPY_REPLY,
        &VmsMailUserCopyReply, 0 },
      { sizeof(VmsMailUserCopySend), MAIL$_USER_COPY_SEND,
        &VmsMailUserCopySend, 0 },
      { sizeof(VmsMailUserForm)-1, MAIL$_USER_FORM,
        VmsMailUserForm, &VmsMailUserFormLength },
      { sizeof(VmsMailUserEditor)-1, MAIL$_USER_EDITOR,
        VmsMailUserEditor, &VmsMailUserEditorLength },
      { sizeof(VmsMailUserForwarding)-1, MAIL$_USER_FORWARDING,
        VmsMailUserForwarding, &VmsMailUserForwardingLength },
      { sizeof(VmsMailUserFullDirectory)-1, MAIL$_USER_FULL_DIRECTORY,
        VmsMailUserFullDirectory, &VmsMailUserFullDirectoryLength },
      { sizeof(VmsMailUserPersonalName)-1, MAIL$_USER_PERSONAL_NAME,
        VmsMailUserPersonalName, &VmsMailUserPersonalNameLength },
      { sizeof(VmsMailUserQueue)-1, MAIL$_USER_QUEUE,
        VmsMailUserForm, &VmsMailUserQueueLength },
      { 0,0,0,0 }
   };

   static struct VmsItemStruct   UserSigFileItem [] = {
      { sizeof(VmsMailUserSigFile)-1, MAIL$_USER_SIGFILE,
        VmsMailUserSigFile, &VmsMailUserSigFileLength },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   static struct VmsItemStruct   UserNameItem [] = {
      { 0, MAIL$_USER_USERNAME, MailUserName, &MailUserNameLength },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MailGetUserInfo()\n");

   UserContext = 0;

   UserNameItem[0].buf_len = MailUserNameLength;

   status = mail$user_begin (&UserContext, &NullItem, &NullItem);
   if (Debug) fprintf (stdout, "mail$user_begin() %%X%08.08X\n", status);

   /* status from mail$user_begin() */
   if (VMSnok (status)) return (status);

   status = mail$user_get_info (&UserContext,
                                &UserNameItem,
                                &UserDetailsItem);
   if (Debug) fprintf (stdout, "mail$user_get_info() %%X%08.08X\n", status);

   if (VMSok (status))
   {
      VmsMailUserEditor[VmsMailUserEditorLength] = '\0';
      VmsMailUserForwarding[VmsMailUserForwardingLength] = '\0';
      VmsMailUserForm[VmsMailUserFormLength] = '\0';
      VmsMailUserFullDirectory[VmsMailUserFullDirectoryLength] = '\0';
      VmsMailUserPersonalName[VmsMailUserPersonalNameLength] = '\0';
      VmsMailUserQueue[VmsMailUserQueueLength] = '\0';
      if (Debug)
         fprintf (stdout,
"%d %d %d %d %d %d\n|%s|\n|%s|\n|%s|\n|%s|\n|%s|\n|%s|\n|%s|\n",
                 VmsMailUserNewMessages,
                 VmsMailUserAutoPurge,
                 VmsMailUserCcPrompt,
                 VmsMailUserCopyForward,
                 VmsMailUserCopyReply, 
                 VmsMailUserCopySend, 
                 VmsMailUserEditor,
                 VmsMailUserForm,
                 VmsMailUserForwarding,
                 VmsMailUserFullDirectory,
                 VmsMailUserPersonalName,
                 VmsMailUserSigFile,
                 VmsMailUserQueue);
   }

   VmsMailUserSigFile[0] = '\0';
   if (VmsVersion >= 70 &&
       VMSok (status))
   {
      status = mail$user_get_info (&UserContext,
                                   &UserNameItem,
                                   &UserSigFileItem);
      if (Debug) fprintf (stdout, "mail$user_get_info() %%X%08.08X\n", status);
      if (VMSok (status))
      {
         VmsMailUserSigFile[VmsMailUserSigFileLength] = '\0';
         if (Debug) fprintf (stdout, "|%s|\n", VmsMailUserSigFile);
      }
      status = SS$_NORMAL;
   }

   mail$user_end (&UserContext, &NullItem, &NullItem);

   return (status);
}

/*****************************************************************************/
/*
This function changes entries in the user profile specified by the global
storage 'MailUserName'.  It checks for the existance of CGI variables
representing each of the possible items to set.  If that CGI variables exists
it uses the information to set the item.  If it does not exist it makes no
change to that item.
*/

MailSetUserInfo ()

{
   static struct VmsItemStruct   UserSetItem [] = {
      { 0, 0, 0, 0 },
      { 0, MAIL$_USER_USERNAME, MailUserName, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   register char  *cptr;

   int  status;
   unsigned short  NewMessages;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MailSetUserInfo()\n");

   UserContext = 0;

   status = mail$user_begin (&UserContext, &NullItem, &NullItem);
   if (Debug) fprintf (stdout, "mail$user_begin() %%X%08.08X\n", status);

   /* status from mail$user_begin() */
   if (VMSnok (status)) return (status);

   UserSetItem[1].buf_len = MailUserNameLength;

   if ((cptr = CgiLibVarNull ("WWW_FORM_UAP")) != NULL)
   {
      memset (&UserSetItem[0], 0, sizeof(UserSetItem[0]));
      if (cptr[0] == '1')
         UserSetItem[0].item = MAIL$_USER_SET_AUTO_PURGE;
      else
         UserSetItem[0].item = MAIL$_USER_SET_NO_AUTO_PURGE;
      status = mail$user_set_info (&UserContext,
                                   &UserSetItem,
                                   &NullItem);
      if (Debug) fprintf (stdout, "mail$user_set_info() %%X%08.08X\n", status);
      if (VMSnok (status))
         ResponseErrorMail (FI_LI, status, lang_ErrSetUserAutoPurge);
   }

   if (VMSok (status) &&
       (cptr = CgiLibVarNull ("WWW_FORM_UCP")) != NULL)
   {
      memset (&UserSetItem[0], 0, sizeof(UserSetItem[0]));
      if (cptr[0] == '1')
         UserSetItem[0].item = MAIL$_USER_SET_CC_PROMPT;
      else
         UserSetItem[0].item = MAIL$_USER_SET_NO_CC_PROMPT;
      status = mail$user_set_info (&UserContext,
                                   &UserSetItem,
                                   &NullItem);
      if (Debug) fprintf (stdout, "mail$user_set_info() %%X%08.08X\n", status);
      if (VMSnok (status))
         ResponseErrorMail (FI_LI, status, lang_ErrSetUserCcPrompt);
   }

   if (VMSok (status) &&
       (cptr = CgiLibVarNull ("WWW_FORM_UCF")) != NULL)
   {
      memset (&UserSetItem[0], 0, sizeof(UserSetItem[0]));
      if (cptr[0] == '1')
         UserSetItem[0].item = MAIL$_USER_SET_COPY_FORWARD;
      else
         UserSetItem[0].item = MAIL$_USER_SET_NO_COPY_FORWARD;
      status = mail$user_set_info (&UserContext,
                                   &UserSetItem,
                                   &NullItem);
      if (Debug) fprintf (stdout, "mail$user_set_info() %%X%08.08X\n", status);
      if (VMSnok (status))
         ResponseErrorMail (FI_LI, status, lang_ErrSetUserCopyForward);
   }

   if (VMSok (status) &&
       (cptr = CgiLibVarNull ("WWW_FORM_UCR")) != NULL)
   {
      memset (&UserSetItem[0], 0, sizeof(UserSetItem[0]));
      if (cptr[0] == '1')
         UserSetItem[0].item = MAIL$_USER_SET_COPY_REPLY;
      else
         UserSetItem[0].item = MAIL$_USER_SET_NO_COPY_REPLY;
      status = mail$user_set_info (&UserContext,
                                   &UserSetItem,
                                   &NullItem);
      if (Debug) fprintf (stdout, "mail$user_set_info() %%X%08.08X\n", status);
      if (VMSnok (status))
         ResponseErrorMail (FI_LI, status, lang_ErrSetUserCopyReply);
   }

   if (VMSok (status) &&
       (cptr = CgiLibVarNull ("WWW_FORM_UCS")) != NULL)
   {
      memset (&UserSetItem[0], 0, sizeof(UserSetItem[0]));
      if (cptr[0] == '1')
         UserSetItem[0].item = MAIL$_USER_SET_COPY_SEND;
      else
         UserSetItem[0].item = MAIL$_USER_SET_NO_COPY_SEND;
      status = mail$user_set_info (&UserContext,
                                   &UserSetItem,
                                   &NullItem);
      if (Debug) fprintf (stdout, "mail$user_set_info() %%X%08.08X\n", status);
      if (VMSnok (status))
         ResponseErrorMail (FI_LI, status, lang_ErrSetUserCopySend);
   }

   if (VMSok (status) &&
       (cptr = CgiLibVarNull ("WWW_FORM_UED")) != NULL)
   {
      memset (&UserSetItem[0], 0, sizeof(UserSetItem[0]));
      if (cptr[0])
      {
         UserSetItem[0].item = MAIL$_USER_SET_EDITOR;
         UserSetItem[0].buf_addr = cptr;
         UserSetItem[0].buf_len = strlen(cptr);
      }
      else
         UserSetItem[0].item = MAIL$_USER_SET_NO_EDITOR;
      status = mail$user_set_info (&UserContext,
                                   &UserSetItem,
                                   &NullItem);
      if (Debug) fprintf (stdout, "mail$user_set_info() %%X%08.08X\n", status);
      if (VMSnok (status))
         ResponseErrorMail (FI_LI, status, lang_ErrSetUserEditor);
   }

   if (VMSok (status) &&
       (cptr = CgiLibVarNull ("WWW_FORM_UFW")) != NULL)
   {
      memset (&UserSetItem[0], 0, sizeof(UserSetItem[0]));
      if (cptr[0])
      {
         UserSetItem[0].item = MAIL$_USER_SET_FORWARDING;
         UserSetItem[0].buf_addr = cptr;
         UserSetItem[0].buf_len = strlen(cptr);
      }
      else
         UserSetItem[0].item = MAIL$_USER_SET_NO_FORWARDING;
      status = mail$user_set_info (&UserContext,
                                   &UserSetItem,
                                   &NullItem);
      if (Debug) fprintf (stdout, "mail$user_set_info() %%X%08.08X\n", status);
      if (VMSnok (status))
         ResponseErrorMail (FI_LI, status, lang_ErrSetUserForwarding);
   }

   if (VMSok (status) &&
       (cptr = CgiLibVarNull ("WWW_FORM_UNM")) != NULL)
   {
      memset (&UserSetItem[0], 0, sizeof(UserSetItem[0]));
      if (isdigit(cptr[0]))
      {
         NewMessages = atoi(cptr);
         if (Debug) fprintf (stdout, "NewMessages: %d\n", NewMessages);

         UserSetItem[0].item = MAIL$_USER_SET_NEW_MESSAGES;
         UserSetItem[0].buf_addr = &NewMessages;
         UserSetItem[0].buf_len = sizeof(NewMessages);

         status = mail$user_set_info (&UserContext,
                                      &UserSetItem,
                                      &NullItem);
         if (Debug)
            fprintf (stdout, "mail$user_set_info() %%X%08.08X\n", status);
         if (VMSnok (status))
            ResponseErrorMail (FI_LI, status, lang_ErrSetUserNewMessages);
      }
   }

   if (VMSok (status) &&
       (cptr = CgiLibVarNull ("WWW_FORM_UPN")) != NULL)
   {
      memset (&UserSetItem[0], 0, sizeof(UserSetItem[0]));
      if (cptr[0])
      {
         UserSetItem[0].item = MAIL$_USER_SET_PERSONAL_NAME;
         UserSetItem[0].buf_addr = cptr;
         UserSetItem[0].buf_len = strlen(cptr);
      }
      else
         UserSetItem[0].item = MAIL$_USER_SET_NO_PERSONAL_NAME;
      status = mail$user_set_info (&UserContext,
                                   &UserSetItem,
                                   &NullItem);
      if (Debug) fprintf (stdout, "mail$user_set_info() %%X%08.08X\n", status);
      if (VMSnok (status))
         ResponseErrorMail (FI_LI, status, lang_ErrSetUserPersonalName);
   }

   if (VmsVersion >= 70 &&
       VMSok (status) &&
       (cptr = CgiLibVarNull ("WWW_FORM_USF")) != NULL)
   {
      memset (&UserSetItem[0], 0, sizeof(UserSetItem[0]));
      if (cptr[0])
      {
         UserSetItem[0].item = MAIL$_USER_SET_SIGFILE;
         UserSetItem[0].buf_addr = cptr;
         UserSetItem[0].buf_len = strlen(cptr);
      }
      else
         UserSetItem[0].item = MAIL$_USER_SET_NO_SIGFILE;
      status = mail$user_set_info (&UserContext,
                                   &UserSetItem,
                                   &NullItem);
      if (Debug) fprintf (stdout, "mail$user_set_info() %%X%08.08X\n", status);
      if (VMSnok (status))
         ResponseErrorMail (FI_LI, status, lang_ErrSetUserSigFile);
   }

   if (VMSok (status) &&
       (cptr = CgiLibVarNull ("WWW_FORM_UPQ")) != NULL)
   {
      memset (&UserSetItem[0], 0, sizeof(UserSetItem[0]));
      if (cptr[0])
      {
         UserSetItem[0].item = MAIL$_USER_SET_QUEUE;
         UserSetItem[0].buf_addr = cptr;      
         UserSetItem[0].buf_len = strlen(cptr);
      }
      else
         UserSetItem[0].item = MAIL$_USER_SET_NO_QUEUE;
      status = mail$user_set_info (&UserContext,
                                   &UserSetItem,
                                   &NullItem);
      if (Debug) fprintf (stdout, "mail$user_set_info() %%X%08.08X\n", status);
      if (VMSnok (status))
         ResponseErrorMail (FI_LI, status, lang_ErrSetUserQueue);
   }

   if (VMSok (status) &&
       (cptr = CgiLibVarNull ("WWW_FORM_UPF")) != NULL)
   {
      memset (&UserSetItem[0], 0, sizeof(UserSetItem[0]));
      if (cptr[0])
      {
         UserSetItem[0].item = MAIL$_USER_SET_FORM;
         UserSetItem[0].buf_addr = cptr;
         UserSetItem[0].buf_len = strlen(cptr);
      }
      else
         UserSetItem[0].item = MAIL$_USER_SET_NO_FORM;
      status = mail$user_set_info (&UserContext,
                                   &UserSetItem,
                                   &NullItem);
      if (Debug) fprintf (stdout, "mail$user_set_info() %%X%08.08X\n", status);
      if (VMSnok (status))
         ResponseErrorMail (FI_LI, status, lang_ErrSetUserForm);
   }

   mail$user_end (&UserContext, &NullItem, &NullItem);

   return (status);
}

/*****************************************************************************/
/*
If messages are moved to or from, or deleted from the NEWMAIL folder update the
count of new messages.
*/

MailSetUserNewMessages (int DeltaValue)

{
   static short  NewMessages;

   static struct VmsItemStruct   UserGetItem [] = {
      { sizeof(NewMessages), MAIL$_USER_NEW_MESSAGES, &NewMessages, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   static struct VmsItemStruct   UserSetItem [] = {
      { sizeof(NewMessages), MAIL$_USER_SET_NEW_MESSAGES, &NewMessages, 0 },
      { 0, MAIL$_USER_USERNAME, MailUserName, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   static struct VmsItemStruct   UserNameItem [] = {
      { 0, MAIL$_USER_USERNAME, MailUserName, &MailUserNameLength },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MailSetUserNewMessages()\n");

   UserContext = 0;

   status = mail$user_begin (&UserContext, &NullItem, &NullItem);
   if (Debug) fprintf (stdout, "mail$user_begin() %%X%08.08X\n", status);

   /* status from mail$user_begin() */
   if (VMSnok (status)) return (status);

   UserNameItem[0].buf_len = MailUserNameLength;

   status = mail$user_get_info (&UserContext,
                                &UserNameItem,
                                &UserGetItem);
   if (Debug) fprintf (stdout, "mail$user_get_info() %%X%08.08X\n", status);

   if (VMSok (status))
   {
      if (Debug)
         fprintf (stdout, "NewMessages: %d (%d) %d\n",
                  NewMessages, DeltaValue, NewMessages+DeltaValue);
      NewMessages += DeltaValue;
      if (NewMessages < 0) NewMessages = 0;

      UserSetItem[1].buf_len = MailUserNameLength;

      status = mail$user_set_info (&UserContext,
                                   &UserSetItem,
                                   &NullItem);
      if (Debug) fprintf (stdout, "mail$user_set_info() %%X%08.08X\n", status);
   }

   mail$user_end (&UserContext, &NullItem, &NullItem);

   return (status);
}

/*****************************************************************************/
/*
This function creates the user profile entry specified by the global storage
'MailUserName'.  As this is probably a "bogus" entry forwarding is set to some
"real" mail username. 
*/

MailCreateUserEntry ()

{
   static struct VmsItemStruct   UserSetItem [] = {
      { 0, MAIL$_USER_SET_FORWARDING, 0, 0 },
      { 0, MAIL$_USER_USERNAME, MailUserName, 0 },
      { 0, MAIL$_USER_CREATE_IF, 0, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   register char  *cptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MailCreateUserEntry()\n");

   UserContext = 0;

   status = mail$user_begin (&UserContext, &NullItem, &NullItem);
   if (Debug) fprintf (stdout, "mail$user_begin() %%X%08.08X\n", status);

   /* status from mail$user_begin() */
   if (VMSnok (status)) return (status);

   UserSetItem[1].buf_len = MailUserNameLength;

   if ((cptr = CgiLibVarNull ("WWW_FORM_UFW")) != NULL)
   {
      if (cptr[0])
      {
         UserSetItem[0].buf_addr = cptr;
         UserSetItem[0].buf_len = strlen(cptr);

         status = mail$user_set_info (&UserContext,
                                      &UserSetItem,
                                      &NullItem);
         if (Debug)
            fprintf (stdout, "mail$user_set_info() %%X%08.08X\n", status);
         if (VMSnok (status))
            ResponseErrorMail (FI_LI, status, lang_ErrUserCreate);
      }
      else
      {
         CgiLibResponseError (FI_LI, 0, lang_ErrUserCreateForward);
         status = STS$K_ERROR;
      }
   }

   mail$user_end (&UserContext, &NullItem, &NullItem);

   return (status);
}

/*****************************************************************************/
/*
This function deletes the user profile entry specified by the global storage
'MailUserName'. 
*/

MailDeleteUserEntry ()

{
   static struct VmsItemStruct   UserDeleteItem [] = {
      { 0, MAIL$_USER_USERNAME, MailUserName, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MailDeleteUserEntry()\n");

   UserContext = 0;

   status = mail$user_begin (&UserContext, &NullItem, &NullItem);
   if (Debug) fprintf (stdout, "mail$user_begin() %%X%08.08X\n", status);

   /* status from mail$user_begin() */
   if (VMSnok (status)) return (status);

   UserDeleteItem[0].buf_len = MailUserNameLength;

   status = mail$user_delete_info (&UserContext,
                                   &UserDeleteItem,
                                   &NullItem);
   if (Debug) fprintf (stdout, "mail$user_delete_info() %%X%08.08X\n", status);
   if (VMSnok (status))
      ResponseErrorMail (FI_LI, status, lang_ErrUserDelete);

   mail$user_end (&UserContext, &NullItem, &NullItem);

   return (status);
}

/*****************************************************************************/
/*
Generate a linked list of user names from the VMS mail profile file.
*/

int GenerateUserList ()

{
   static unsigned short  UserForwardingLength,
                          UserReturnNameLength;
   static char  UserForwarding [256],
                UserReturnName [256];

   static struct VmsItemStruct   UserDetailsItem [] = {
      { sizeof(UserForwarding)-1, MAIL$_USER_FORWARDING,
        UserForwarding, &UserForwardingLength },
      { sizeof(UserReturnName)-1, MAIL$_USER_RETURN_USERNAME,
        UserReturnName, &UserReturnNameLength },
      { 0,0,0,0 }
   };

   static struct VmsItemStruct   UserFirstItem [] = {
      { 0, MAIL$_USER_FIRST, 0, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   static struct VmsItemStruct   UserNextItem [] = {
      { 0, MAIL$_USER_NEXT, 0, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   register struct UserListStruct  *ulptr;

   int  status,
        CallocSize;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "GenerateUserList()\n");

   if (UserListHeadPtr != NULL)
   {
      /* empty any current list of users */
      struct UserListStruct  *ueptr;
      ulptr = UserListHeadPtr;
      while (ulptr != NULL)
      {
         ueptr = ulptr;
         ulptr = ulptr->NextPtr;
         free (ueptr);
      }
      UserListHeadPtr = UserListTailPtr = NULL;
      UserListCount = 0;
   }

   UserContext = 0;

   status = mail$user_begin (&UserContext, &NullItem, &NullItem);
   if (Debug) fprintf (stdout, "mail$user_begin() %%X%08.08X\n", status);

   /* status from mail$user_begin() */
   if (VMSnok (status)) return (status);

   status = mail$user_get_info (&UserContext,
                                &UserFirstItem,
                                &UserDetailsItem);
   if (Debug) fprintf (stdout, "mail$user_get_info() %%X%08.08X\n", status);

   while (VMSok (status))
   {
      UserReturnName[UserReturnNameLength] = '\0';
      if (Debug)
         fprintf (stdout, "%d |%s|\n",
                  UserReturnNameLength, UserReturnName);

      if (isprint(UserReturnName[0]) &&
          decc$match_wild (UserReturnName, MailUserName))
      {
         CallocSize = sizeof(struct UserListStruct) + 
                      UserReturnNameLength + 1;
         ulptr = calloc (1, CallocSize);
         if (ulptr == NULL) exit (vaxc$errno);
         ulptr->NextPtr = NULL;
         ulptr->UserNamePtr = &ulptr->Data;
         strcpy (ulptr->UserNamePtr, UserReturnName);
         if (UserListHeadPtr == NULL)
            UserListHeadPtr = UserListTailPtr = ulptr;
         else
         {
            UserListTailPtr->NextPtr = ulptr;
            UserListTailPtr = ulptr;
         }
         UserListCount++;
      }

      status = mail$user_get_info (&UserContext,
                                   &UserNextItem,
                                   &UserDetailsItem);
      if (Debug) fprintf (stdout, "mail$user_get_info() %%X%08.08X\n", status);
   }

   if (status == MAIL$_NOMOREREC) status = SS$_NORMAL;

   mail$user_end (&UserContext, &NullItem, &NullItem);

   return (status);
}

/*****************************************************************************/
/*
Read the user's signature file into memory pointed to by
'VmsMailSigFileTextPtr'.  To prevent access to just any old file yahMAIL
insists that the signature file reside in the same directory as the mail files
and that the profile signature entry comprises just the file name and type.
*/

ReadSigFile ()

{
   int  status;
   char  FileName [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ReadSigFile() |%s|\n", VmsMailUserSigFile);

   strcpy (FileName, VmsMailUserFullDirectory);
   strcat (FileName, VmsMailUserSigFile);
   if (Debug) fprintf (stdout, "FileName |%s|\n", FileName);

   if (VmsMailSigFileTextPtr != NULL) free (VmsMailSigFileTextPtr);
   VmsMailSigFileTextLength = 0;
   status = ReadFileIntoMemory (FileName, &VmsMailSigFileTextPtr,
                                          &VmsMailSigFileTextLength);
   if (VMSnok (status))
   {
      char  String [256];
      unsigned short  Length;
      $DESCRIPTOR (StringDsc, String);

      if (VMSok (sys$getmsg (status, &Length, &StringDsc, 1, 0))) 
         String[Length] = '\0';
      else
         strcpy (String, "sys$getmsg() error!");

      VmsMailSigFileTextLength = Length + strlen(lang_ErrSigFileOpen) + 32;
      VmsMailSigFileTextPtr = calloc (VmsMailSigFileTextLength, 1);
      sprintf (VmsMailSigFileTextPtr, "%s: %s (%%X%08.08X)",
               lang_ErrSigFileOpen, String, status);
   }

   if (Debug) fprintf (stdout, "|%s|\n", VmsMailSigFileTextPtr);
}

/*****************************************************************************/
/*
Display the contents of the user's signature file in a <TEXTAREA></TEXTAREA>
tag.   This allows changes to be made and updated.
*/

SigFileAccess ()
       
{
   int  status;
   char  *SigFileTextPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SigFileAccess()\n");

   if (!VmsMailUserSigFile[0])
      strzcpy (VmsMailUserSigFile,
               CgiLibVar ("WWW_FORM_USF"),
               sizeof(VmsMailUserSigFile));

   if (!VmsMailUserSigFile[0])
   {
      CgiLibResponseError (FI_LI, 0, lang_ErrSigFileNoName);
      return;
   }

   ReadSigFile ();
   if (VmsMailSigFileTextPtr[0])
      SigFileTextPtr = CgiLibHtmlEscape (VmsMailSigFileTextPtr, -1, NULL, 0);
   else
      SigFileTextPtr = "";

   CgiLibResponseHeader (200, "text/html");
   fprintf (stdout,
"<HTML>\n\
<HEAD>\n\
<META NAME=\"generator\" CONTENT=\"%s\">\n\
<META NAME=\"environment\" CONTENT=\"%s\">\n\
<META NAME=\"language\" CONTENT=\"%s %s\">\n\
<TITLE>yahMAIL ... %s / %s</TITLE>\n\
</HEAD>\n\
<BODY %s>\n\
%s\
<B><FONT SIZE=+1><U>%s / %s</FONT></U></B>\n\
<BR>&nbsp;<FONT SIZE=-1>yahMAIL<SUP>*</SUP></FONT>\n\
<FORM ACTION=\"%s\" METHOD=POST>\n\
<TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>\n\
<TR><TD>\n\
<TEXTAREA NAME=txt ROWS=6 COLS=72 WRAP=off>\n\
%s\
</TEXTAREA>\n\
</TD></TR>\n\
</TABLE>\n\
<P><INPUT TYPE=hidden NAME=wha VALUE=\"sig\">\n\
<INPUT TYPE=submit NAME=act VALUE=\"%s\">&nbsp;\n\
<INPUT TYPE=reset VALUE=\"%s\">\
<I> ... %s</I>\n\
<P><INPUT TYPE=button VALUE=\"%s\" onClick=\"history.go(-1)\">\n\
</FORM>\n\
</BODY>\n\
</HTML>\n",
      SoftwareID, CgiEnvironmentPtr, lang_Language, lang_LanguageAuthor,
      MailUserName, lang_ProSignatureFile,
      ConfigBodyTagPtr, PostMasterWarningPtr,
      MailUserName, lang_ProSignatureFile,
      ActionPathInfo, SigFileTextPtr,
      lang_BtnUpdate, lang_BtnReset, lang_ResetCaution,
      lang_BtnGoBack);

   free (SigFileTextPtr);
}

/*****************************************************************************/
/*
Update/create the signature file contents.
*/

SigFileUpdate ()
       
{
   register char  *cptr, *sptr;

   int  status;
   char  FileName [256];
   char  *FileTextPtr;
   FILE  *FilePtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SigFileUpdate()\n");

   if (!VmsMailUserSigFile[0])
      strzcpy (VmsMailUserSigFile,
               CgiLibVar ("WWW_FORM_USF"),
               sizeof(VmsMailUserSigFile));

   if (!VmsMailUserSigFile[0])
   {
      CgiLibResponseError (FI_LI, 0, lang_ErrSigFileNoName);
      return;
   }

   strcpy (FileName, VmsMailUserFullDirectory);
   strcat (FileName, VmsMailUserSigFile);
   if (Debug) fprintf (stdout, "FileName |%s|\n", FileName);

   FileTextPtr = CgiLibVar ("WWW_FORM_TXT");

   FilePtr = fopen (FileName, "w", "shr=nil");
   if (FilePtr == NULL)
   {
      status = vaxc$errno;
      if (Debug) fprintf (stdout, "fopen() %%X%08.08X\n", status);
      return;
   }

   if (fputs (FileTextPtr, FilePtr) == EOF)
   {
      status = vaxc$errno;
      CgiLibResponseError (FI_LI, status, "%s", FileName);
   }
   else
   {
      /* purge the versions back to a maximum of 3 */
      strcat (FileName, ";-3");
      while (!remove (FileName));

      CgiLibResponseSuccess (
"%s\n\
<P><FORM>\n\
<INPUT TYPE=button VALUE=\"%s\" onClick=\"history.go(-1)\">\n\
</FORM>",
         lang_SigFileUpdated, lang_BtnGoBack);
   }

   fclose (FilePtr);
}

/*****************************************************************************/
/*
Prepare and send a message POSTed from CreateMailMessage() for actual sending
using the generic mailing function VmsEmailMessage().
*/

SendMailMessage ()

{
   register char  *cptr, *sptr;

   int  PostBufferCount;
   char  *BodyPtr,
         *CcPtr,
         *FromPtr,
         *PersonalPtr,
         *PostBufferPtr,
         *SubjPtr,
         *ToPtr;
   char  CcString [512],
         FromString [512];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SendMailMessage()\n");

   if (!MailUserName[0])
   {
      CgiLibResponseError (FI_LI, 0, lang_ErrSanityCheck);
      return;
   }

   ToPtr = CgiLibVar("WWW_FORM_STO");
   CcPtr = CgiLibVar("WWW_FORM_SCC");
   SubjPtr = CgiLibVar("WWW_FORM_SSU");
   BodyPtr = CgiLibVar("WWW_FORM_SBD");

   cptr = CgiLibVar("WWW_FORM_SCS");
   if (cptr[0])
   {
      /* "courtesy copy to self" checkbox */
      strcpy (CcString, CcPtr);
      strcat (CcString, ",");
      strcat (CcString, MailUserName);
      CcPtr = CcString;
   }

   if (PostMasterAuthenticated)
      FromPtr = CgiLibVar("WWW_FORM_SFR");
   else
   {
      PersonalPtr = CgiLibVar("WWW_FORM_SPN");
      if (PersonalPtr[0])
      {
         strcpy (FromString, MailUserName);
         strcat (FromString, " \"");
         strcat (FromString, PersonalPtr);
         strcat (FromString, "\"");
         FromPtr = FromString;
      }
      else
         FromPtr = MailUserName;
   }

   cptr = VmsEmailMessage (ConfigSmtpTransportPtr,
                           FromPtr, ToPtr, CcPtr, SubjPtr,
                           BodyPtr, WrapAt);

   if (cptr == NULL)
   {
      CgiLibResponseSuccess (
"%s\n\
<P><FORM>\n\
<INPUT TYPE=button VALUE=\"%s\" onClick=\"history.go(-1)\">&nbsp;\n\
</FORM>",
         lang_MessageSent, lang_BtnGoBack);
   }
   else
      CgiLibResponseError (FI_LI, 0, cptr);
}

/*****************************************************************************/
/*
This is a self-contained, generic function that could be used else-where!
Use the VMS callable mail interface to create and send a VMS mail message.  
'To' and 'Cc' can be  lists of comma-separated addresses.  'Subject' is a
null-terminated string. 'Body' is a  null-terminated string of '\n'-separated
lines of plain text.  Must be installed with SYSPRV to be able to use the
'From' parameter (without SYSPRV it sends with a MAIL-generated "From:").
*/ 

char* VmsEmailMessage
(
char *SmtpTransport,
char *From,
char *To,
char *Cc,
char *Subject,
char *Body,
int WrapAt
)
{
   static unsigned long  SysPrvMask [2] = { PRV$M_SYSPRV, 0 };
   static unsigned short  UserNameTypeTo = MAIL$_TO,
                          UserNameTypeCc = MAIL$_CC;
   static unsigned short  Length;
   static char  String [256];
   static $DESCRIPTOR (StringDsc, String);

   register int  cnt;
   register char  *cptr, *sptr, *zptr;

   int  status,
        MailAddressLength,
        SmtpTransportLength,
        WrapCount;
   unsigned long  SendContext = 0;
   unsigned long  PrevPrvMask [2];
   char  QuoteChar;
   char  MailAddress [256],
         SmtpAddress [256+64];

   struct {
      short int  buf_len;
      short int  item;
      unsigned char   *buf_addr;
      unsigned short  *ret_len;
   }
   BodyPartItem [] =
   {
      { 0, MAIL$_SEND_RECORD, 0, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      {0,0,0,0}
   },
   FromItem [] =
   {
      { 0, MAIL$_SEND_FROM_LINE, From, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      {0,0,0,0}
   },
   SendBeginSigFileItem [] =
   {
      { 0, MAIL$_SEND_NO_SIGFILE, 0, 0 },
      { 0, MAIL$_SEND_NO_PERS_NAME, 0, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      {0,0,0,0}
   },
   SendBeginNoSigFileItem [] =
   {
      { 0, MAIL$_SEND_NO_PERS_NAME, 0, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      {0,0,0,0}
   },
   SendToItem [] =
   {
      { 0, MAIL$_SEND_USERNAME, 0, 0 },
      { sizeof(UserNameTypeTo), MAIL$_SEND_USERNAME_TYPE, &UserNameTypeTo, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      {0,0,0,0}
   },
   SendCcItem [] =
   {
      { 0, MAIL$_SEND_USERNAME, 0, 0 },
      { sizeof(UserNameTypeCc), MAIL$_SEND_USERNAME_TYPE, &UserNameTypeCc, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      {0,0,0,0}
   },
   SubjectItem [] =
   {
      { 0, MAIL$_SEND_SUBJECT, Subject, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      {0,0,0,0}
   },
   NoSignalItem [] =
   {
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      {0,0,0,0}
   },
   NullItem = {0,0,0,0};

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "VmsEmailMessage()\n");

   if (To == NULL || !To[0]) return (lang_ErrSendRecipient);

   /* let's keep it within the bounds of reason! */
   if (WrapAt && (WrapAt < 48 || WrapAt > 255)) WrapAt = 72;

   if (SmtpTransport == NULL)
      SmtpTransportLength = 0;
   else
      SmtpTransportLength = strlen(SmtpTransport);

   /* basically to support pre-VMS6.2 where SEND_NO_SIGFILE is an error */
   if (VmsVersion >= 70)
   {
      status = mail$send_begin (&SendContext,
                                &SendBeginSigFileItem,
                                &NullItem);
      if (Debug) fprintf (stdout, "mail$send_begin() %%X%08.08X\n", status);
   }
   else
   {
      status = mail$send_begin (&SendContext,
                                &SendBeginNoSigFileItem,
                                &NullItem);
      if (Debug) fprintf (stdout, "mail$send_begin() %%X%08.08X\n", status);
   }
   if (VMSnok (status)) goto VmsEmailMessageError;

   /********/
   /* from */
   /********/

   if (From != NULL && From[0])
   {
      status = sys$setprv (1, &SysPrvMask, 0, &PrevPrvMask);
      if (Debug) fprintf (stdout, "sys$setprv() %%X%08.08X\n", status);
      if (VMSok (status) && status != SS$_NOTALLPRIV)
      {
         if ((FromItem[0].buf_len = strlen(From)) > 255)
            FromItem[0].buf_len = 255;
         status = mail$send_add_attribute (&SendContext, &FromItem, &NullItem);
         if (Debug)
            fprintf (stdout, "mail$send_add_attribute() %%X%08.08X\n", status);
         if (VMSnok (status)) goto VmsEmailMessageError;

         if (!(PrevPrvMask[0] & PRV$M_SYSPRV))
         {
            status = sys$setprv (0, &SysPrvMask, 0, 0);
            if (Debug) fprintf (stdout, "sys$setprv() %%X%08.08X\n", status);
            if (VMSnok (status)) exit (status);
         }
      }
   }

   /******************/
   /* to address(es) */
   /******************/

   /* a single, or multiple comma-separated addresses */
   cptr = To;
   while (*cptr)
   {
      while (*cptr && (*cptr == ',' || isspace(*cptr))) cptr++;
      if (*cptr == '!' || *cptr == '#')
      {
         /* comment "line" */
         while (*cptr && *cptr != '\n') cptr++;
         if (*cptr) cptr++;
         continue;
      }
      zptr = (sptr = MailAddress ) + sizeof(MailAddress)-1;
      while (*cptr && *cptr != ',' && !isspace(*cptr) && sptr < zptr)
      {
         if (*cptr != '\"' && *cptr != '\'')
         {
            if (sptr < zptr) *sptr++ = *cptr++;
            continue;
         }
         /* encountered a quote */
         QuoteChar = *cptr++;
         if (sptr < zptr)
         {
            if (SmtpTransportLength)
               *sptr++ = '\'';
            else
               *sptr++ = QuoteChar;
         }

         while (*cptr && *cptr != QuoteChar && sptr < zptr) *sptr++ = *cptr++;
         if (*cptr)
         {
            cptr++;
            if (sptr < zptr)
            {
               if (SmtpTransportLength)
                  *sptr++ = '\'';
               else
                  *sptr++ = QuoteChar;
            }
         }
      }
      if (sptr >= zptr) return (lang_ErrSendMailAddressOvf);
      *sptr = '\0';
      MailAddressLength = sptr - MailAddress;
      if (Debug) fprintf (stdout, "%d |%s|\n", MailAddressLength, MailAddress);

      SendToItem[0].buf_addr = MailAddress;
      SendToItem[0].buf_len = MailAddressLength;

      if (SmtpTransportLength)
      {
         for (sptr = MailAddress; *sptr && *sptr != '@'; sptr++);
         if (*sptr)
         {
            /* looks like an RFC822 address, check for a Mail transport */
            for (sptr = MailAddress;
                 *sptr && *(unsigned short*)sptr != '%\"';
                 sptr++);
            if (!*sptr)
            {
               /* no Mail transport, add one */
               strcpy (sptr = SmtpAddress, SmtpTransport);
               sptr += SmtpTransportLength;
               *sptr++ = '\"';
               strcpy (sptr, MailAddress);
               sptr += MailAddressLength;
               *sptr++ = '\"';
               *sptr = '\0';

               SendToItem[0].buf_addr = SmtpAddress;
               SendToItem[0].buf_len = sptr - SmtpAddress;;
               if (Debug)
                  fprintf (stdout, "%d |%s|\n", sptr-SmtpAddress, SmtpAddress);
            }
         }
      }

      if (Debug)
         fprintf (stdout, "To: address |%s|\n", SendToItem[0].buf_addr);
      status = mail$send_add_address (&SendContext, &SendToItem, &NullItem);
      if (Debug)
         fprintf (stdout, "mail$send_add_address() %%X%08.08X\n", status);
      if (VMSnok (status)) goto VmsEmailMessageError;
   }

   /******************/
   /* cc address(es) */
   /******************/

   /* a single, or multiple comma-separated addresses */
   if ((cptr = Cc) == NULL) cptr = "";
   while (*cptr)
   {
      while (*cptr && (*cptr == ',' || isspace(*cptr))) cptr++;
      if (*cptr == '!' || *cptr == '#')
      {
         /* comment "line" */
         while (*cptr && *cptr != '\n') cptr++;
         if (*cptr) cptr++;
         continue;
      }
      zptr = (sptr = MailAddress ) + sizeof(MailAddress)-1;
      while (*cptr && *cptr != ',' && !isspace(*cptr) && sptr < zptr)
      {
         if (*cptr != '\"' && *cptr != '\'')
         {
            if (sptr < zptr) *sptr++ = *cptr++;
            continue;
         }
         /* encountered a quote */
         QuoteChar = *cptr++;
         if (sptr < zptr)
         {
            if (SmtpTransportLength)
               *sptr++ = '\'';
            else
               *sptr++ = QuoteChar;
         }
         while (*cptr && *cptr != QuoteChar && sptr < zptr) *sptr++ = *cptr++;
         if (*cptr)
         {
            cptr++;
            if (sptr < zptr)
            {
               if (SmtpTransportLength)
                  *sptr++ = '\'';
               else
                  *sptr++ = QuoteChar;
            }
         }
      }
      if (sptr >= zptr) return (lang_ErrSendMailAddressOvf);
      *sptr = '\0';
      MailAddressLength = sptr - MailAddress;
      if (Debug) fprintf (stdout, "%d |%s|\n", MailAddressLength, MailAddress);

      SendCcItem[0].buf_addr = MailAddress;
      SendCcItem[0].buf_len = sptr - MailAddress;

      if (SmtpTransportLength)
      {
         for (sptr = MailAddress; *sptr && *sptr != '@'; sptr++);
         if (*sptr)
         {
            /* looks like an RFC822 address, check for a Mail transport */
            for (sptr = MailAddress;
                 *sptr && *(unsigned short*)sptr != '%\"';
                 sptr++);
            if (!*sptr)
            {
               /* no Mail transport, add one */
               strcpy (sptr = SmtpAddress, SmtpTransport);
               sptr += SmtpTransportLength;
               *sptr++ = '\"';
               strcpy (sptr, MailAddress);
               sptr += MailAddressLength;
               *sptr++ = '\"';
               *sptr = '\0';

               SendCcItem[0].buf_addr = SmtpAddress;
               SendCcItem[0].buf_len = sptr - SmtpAddress;;
               if (Debug)
                  fprintf (stdout, "%d |%s|\n", sptr-SmtpAddress, SmtpAddress);
            }
         }
      }

      if (Debug)
         fprintf (stdout, "CC: address |%s|\n", SendCcItem[0].buf_addr);
      status = mail$send_add_address (&SendContext, &SendCcItem, &NullItem);
      if (Debug)
         fprintf (stdout, "mail$send_add_address() %%X%08.08X\n", status);
      if (VMSnok (status)) goto VmsEmailMessageError;
   }

   /***********/
   /* subject */
   /***********/

   if ((cptr = Subject) == NULL) cptr = "";
   if ((SubjectItem[0].buf_len = strlen(cptr)) > 255)
      SubjectItem[0].buf_len = 255;
   status = mail$send_add_attribute (&SendContext, &SubjectItem, &NullItem);
   if (Debug)
      fprintf (stdout, "mail$send_add_attribute() %%X%08.08X\n", status);
   if (VMSnok (status)) goto VmsEmailMessageError;

   /********/
   /* body */
   /********/

   cptr = sptr = Body;
   while (*cptr)
   {
      BodyPartItem[0].buf_addr = cptr;

      /* only wrap non-quoted lines (those without leading '>') */
      if (WrapAt && *sptr != '>')
         cnt = WrapAt;
      else
         cnt = 255;

      while (*cptr && *cptr != '\r' && *cptr != '\n' && cnt)
      {
         cptr++;
         cnt--;
      }
      if (!cnt && *cptr != ' ' && *cptr != '\t')
      {
         /* reached the "wrap-at" limit while in a "word", back a little */
         zptr = cptr;
         while (*cptr && *cptr != ' ' && *cptr != '\t' &&
                cptr > BodyPartItem[0].buf_addr) cptr--;
         /* if it's a continuous line of characters then just break it! */
         if (cptr == BodyPartItem[0].buf_addr) cptr = zptr;
      }

      BodyPartItem[0].buf_len = cptr - (char*)BodyPartItem[0].buf_addr;
      status = mail$send_add_bodypart (&SendContext, &BodyPartItem, &NullItem);
      if (Debug)
         fprintf (stdout, "mail$send_add_bodypart() %%X%08.08X\n", status);
      if (VMSnok (status)) goto VmsEmailMessageError;

      /* if it was wrapped then this will absorb leading white-space */
      while (*cptr && (*cptr == ' ' || *cptr == '\t')) cptr++;
      if (*cptr == '\r') sptr = ++cptr;
      if (*cptr == '\n') sptr = ++cptr;
   }

   status = mail$send_message (&SendContext, &NoSignalItem, &NoSignalItem);
   if (Debug) fprintf (stdout, "mail$send_message() %%X%08.08X\n", status);
   if (VMSnok (status)) goto VmsEmailMessageError;

   mail$send_end (&SendContext, &NullItem, &NullItem);

   return (NULL);

VmsEmailMessageError:

   mail$send_end (&SendContext, &NullItem, &NullItem);

   if (status == MAIL$_LOGLINK)
      sprintf (String, "%s \'%s\'.", lang_ErrSendInvalidUser, MailAddress);
   else
   if (status == MAIL$_NOSUCHUSR)
      sprintf (String, "%s \'%s\'.", lang_ErrSendNoSuchUser, MailAddress);
   else
   if (status == MAIL$_PARSEFAIL)
      sprintf (String, "%s \'%s\'.", lang_ErrSendParsing, MailAddress);
   else
   if (status == MAIL$_USERSPEC)
      sprintf (String, "%s \'%s\'.", lang_ErrSendInvalidUser, MailAddress);
   else
   if (status == MAIL$_ERRACTRNS)
      sprintf (String, "%s \'%s\'.", lang_ErrSendTransport, MailAddress);
   else
   if (VMSok (status = sys$getmsg (status, &Length, &StringDsc, 1, 0))) 
   {
      String[Length] = '.';
      String[Length+1] = '\0';
      String[0] = toupper(String[0]);
   }
   else
      strcpy (String, "sys$getmsg() error!");

   if (Debug) fprintf (stdout, "|%s|\n", String);
   return (String);
}

/*****************************************************************************/
/*
Filter some callable MAIL status to generate specific messages.
*/

ResponseErrorMail
(
char *SrcFileName,
int SrcLineNumber,
int StatusValue,
char *Text
)
{
   char  String [256];

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "ResponseErrorMail() %%X%08.08X (%d)\n",
               StatusValue, StatusValue);

   /* cannot get this to work for the life of me!!! */
   /** if (StatusValue == MAIL$_NOMOREMSG) **/
   if (StatusValue - MAIL$_NOMOREMSG == 0)
      strcpy (String, lang_ErrMailMessageNotFound);
   else
   if (StatusValue == MAIL$_NOSUCHUSR)
   {
      sprintf (String, "%s \'%s\'.", lang_ErrMailNoSuchUser, MailUserName);
      CgiLibResponseError (SrcFileName, SrcLineNumber, 0, String); 

      if (!PostMasterAuthenticated) return (SS$_NORMAL);

      fprintf (stdout,
"<P><FORM ACTION=\"%s\">\n\
<FONT SIZE=+1 COLOR=\"#ff0000\"><B><U>%s</U></B></FONT>\n\
<BR><B>%s:</B>&nbsp;&nbsp;<INPUT TYPE=text SIZE=60 NAME=ufw>\n\
<BR><INPUT TYPE=submit NAME=act VALUE=\"%s\">\n\
</FORM>\n",
         ActionPathInfo, lang_PostMaster, lang_ProForwarding,
         lang_BtnCreateUserProfile);

      return;
   }
   else
   if (StatusValue == MAIL$_NOTEXIST)
   {
      sprintf (String, "%s \'%s\'", lang_ErrMailNoSuchFolder, MailFolderName);
      CgiLibResponseError (SrcFileName, SrcLineNumber, 0, String); 

      if (!FolderListCount) return (SS$_NORMAL);

      fprintf (stdout, "<P><FORM ACTION=\"%s\">\n", ActionPathInfo);

      FolderListHtmlSelect ();

      fprintf (stdout,
"<INPUT TYPE=submit NAME=act VALUE=\"%s\">&nbsp;\n\
<INPUT TYPE=reset VALUE=\"%s\">\n\
</FORM>\n",
         lang_BtnOpen, lang_BtnReset);

      return;
   }
   else
   if (StatusValue == SS$_POWERFAIL)
   {
      /* this is a bogus status, to indicate a message disparity */
      CgiLibResponseError (SrcFileName, SrcLineNumber, 0,
                           lang_ErrCurrentMessageCount); 
      return;
   }
   else
   if (StatusValue == MAIL$_ILLPERNAM)
      strcpy (String, lang_ErrMailIllegalPersonal);
   else
   if (StatusValue == MAIL$_OPENIN)
      sprintf (String, "%s \'%s\'.", lang_ErrMailOpening, MailFileName);
   else
      String[0] = '\0';

   if (String[0])
      CgiLibResponseError (SrcFileName, SrcLineNumber, 0, String); 
   else
      CgiLibResponseError (SrcFileName, SrcLineNumber, StatusValue, Text);
   return;
}

/*****************************************************************************/
/*
Address list selection page.  A <SELECT><SELECT> list is generated from the
newline-separated entries text pointed at by 'ConfigAddrListPtr' and by
searching the user's mail directory for files ending in ".DIS".  Address list
files, quite deliberately, must be located in the user's mail directory and may
only end with ".DIS".  These helps prevent wholesale access to any old file.
*/

AddrListSelect ()
       
{
   register char  *cptr, *sptr;

   int  status,
        FileCount;
   unsigned long  FindFileCtx;
   unsigned short  NameLength;
   char  *NamePtr;
   char  FileName [256],
         FileSpec [256],
         ListName [256];
   $DESCRIPTOR (FileSpecDsc, FileSpec);
   $DESCRIPTOR (FileNameDsc, FileName);
   struct
   {
      short int  com_len;
      short int  item;
      unsigned char   *com_addr;
   } FileScanItem [] = {
      { 0, FSCN$_NAME, 0 },
      { 0, 0, 0 }
   };

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AddrListSelect()\n");

   CgiLibResponseHeader (200, "text/html");
   fprintf (stdout,
"<HTML>\n\
<HEAD>\n\
<META NAME=\"generator\" CONTENT=\"%s\">\n\
<META NAME=\"environment\" CONTENT=\"%s\">\n\
<META NAME=\"language\" CONTENT=\"%s %s\">\n\
<TITLE>yahMAIL ... %s: %s</TITLE>\n\
</HEAD>\n\
<BODY onLoad=\"self.focus()\" %s>\n\
%s\
<B><FONT SIZE=+1><U>%s: &nbsp;%s</FONT></U></B>\n\
<BR>&nbsp;<FONT SIZE=-1>yahMAIL<SUP>*</SUP></FONT>\n\
<FORM ACTION=\"%s\">\n\
<TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>\n\
<TR><TD>\n\
&nbsp;<SELECT NAME=lse SIZE=5>\n",
      SoftwareID, CgiEnvironmentPtr, lang_Language, lang_LanguageAuthor,
      lang_ListAddressPlural, MailUserName,
      ConfigBodyTagPtr, PostMasterWarningPtr,
      lang_ListAddressPlural, MailUserName,
      ActionPathInfo);

   /* lists specified as part of the configuration */
   cptr = ConfigAddrListPtr;
   while (*cptr)
   {
      while (*cptr && isspace(*cptr)) cptr++;
      if (!*cptr) break;
      sptr = ListName;
      while (*cptr && !isspace(*cptr)) *sptr++ = tolower(*cptr++);
      *sptr = '\0';
      fprintf (stdout, "<OPTION VALUE=\"!%s\">&nbsp;%s&nbsp;</A>\n",
               ListName, ListName);
      while (*cptr && *cptr != '\n') cptr++;
      if (*cptr) cptr++;
   }

   /* files ending in ".DIS" in the user's mail directory */
   strcpy (FileSpec, VmsMailUserFullDirectory);
   strcat (FileSpec, "*.DIS;0");
   FileSpecDsc.dsc$w_length = strlen(FileSpec);
   if (Debug) fprintf (stdout, "|%s|\n", FileSpec);

   FindFileCtx = FileCount = 0;
   for (;;)
   {
      FileNameDsc.dsc$w_length = sizeof(FileName)-1;
      status = lib$find_file (&FileSpecDsc, &FileNameDsc, &FindFileCtx,
                              0, 0, 0, 0);
      if (Debug) fprintf (stdout, "lib$find_file() %%X%08.08X\n", status);
      if (VMSnok (status)) break;

      FileCount++;

      FileName[FileNameDsc.dsc$w_length] = '\0';
      for (cptr = FileName; *cptr && !isspace(*cptr); cptr++);
      FileNameDsc.dsc$w_length = cptr - FileName;
      FileName[FileNameDsc.dsc$w_length] = '\0';
      if (Debug) fprintf (stdout, "|%s|\n", FileNameDsc.dsc$a_pointer);

      status = sys$filescan (&FileNameDsc, &FileScanItem, 0, 0, 0);
      if (Debug) fprintf (stdout, "sys$filescan() %%X%08.08X\n", status);
      if (VMSnok (status)) break;
      strncpy (ListName, FileScanItem[0].com_addr, FileScanItem[0].com_len);
      ListName[FileScanItem[0].com_len] = '\0';
      if (Debug) fprintf (stdout, "|%s|\n", ListName);

      fprintf (stdout, "<OPTION VALUE=\"%s\">&nbsp;%s&nbsp;</A>\n",
               ListName, ListName);
   }
   lib$find_file_end (&FindFileCtx);

   /* if no user address list files the "suggest" 'YAHMAIL' :^) */
   if (!FileCount)
      fprintf (stdout,
"<OPTION VALUE=\"YAHMAIL\">&nbsp;YAHMAIL&nbsp;</A>\n");

   fprintf (stdout,
"</SELECT>\n\
<BR>&nbsp;<INPUT TYPE=text SIZE=20 NAME=lte>\
&nbsp;<I>list name</I>\n\
</TD></TR>\n\
</TABLE>\n\
<P><INPUT TYPE=hidden NAME=wha VALUE=\"lis\">\n\
<INPUT TYPE=submit NAME=act VALUE=\"%s\">&nbsp;&nbsp;\n\
<INPUT TYPE=button VALUE=\"refresh\" onClick=\"location.reload(true)\">&nbsp;\n\
<INPUT TYPE=button VALUE=\"%s\" onClick=\"window.close()\">\n\
</FORM>\n\
</BODY>\n\
</HTML>\n",
      lang_BtnAccess, lang_BtnClose);
}

/*****************************************************************************/
/*
Display the contents of an address list file in a <TEXTAREA></TEXTAREA> tag. 
This allows changes to be made and updated as well as all or selected addresses
to be edit/copied, then edit/pasted elsewhere (e.g. into the "To:" field when
sending).
*/

AddrListAccess ()
       
{
   register char  *cptr, *sptr;

   boolean  UserList;
   int  status,
        ListLength;
   char  ConfigListName [256],
         FileName [256],
         ListFileName [256],
         ListName [256];
   char  *EscapedListPtr,
         *NewListPtr,
         *ListNamePtr,
         *ListPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AddrListAccess()\n");

   ListNamePtr = CgiLibVar ("WWW_FORM_LTE");
   if (!ListNamePtr[0])
      ListNamePtr = CgiLibVar ("WWW_FORM_LSE");
   if (!ListNamePtr[0])
   {
      CgiLibResponseError (FI_LI, 0,
"%s\n\
<P><FORM>\n\
<INPUT TYPE=button VALUE=\"%s\" onClick=\"history.go(-1)\">&nbsp;\n\
<INPUT TYPE=button VALUE=\"%s\" onClick=\"window.close()\">\n\
</FORM>",
         lang_ErrListNotSpecified, lang_BtnGoBack, lang_BtnClose);
      return;
   }

   if (ListNamePtr[0] == '!')
   {
      /* configuration (global) address list */
      UserList = false;

      FileName[0] = '\0';
      cptr = ConfigAddrListPtr;
      while (*cptr)
      {
         while (*cptr && isspace(*cptr)) cptr++;
         if (!*cptr) break;
         sptr = ConfigListName;
         while (*cptr && !isspace(*cptr)) *sptr++ = *cptr++;
         *sptr = '\0';
         if (Debug) fprintf (stdout, "ConfigListName |%s|\n", ConfigListName);
         if (strsame (ListNamePtr+1, ConfigListName, -1))
         {
            while (*cptr && isspace(*cptr)) cptr++;
            sptr = FileName;
            while (*cptr && !isspace(*cptr)) *sptr++ = *cptr++;
            *sptr = '\0';
         }
         while (*cptr && *cptr != '\n') cptr++;
         if (*cptr) cptr++;
      }
   }
   else
   {
      /* user (private) address list */
      UserList = true;

      sptr = ListName;
      for (cptr = ListNamePtr; *cptr; cptr++)
      {
         if (isalnum(*cptr) || *cptr == '-' || *cptr == '_' || *cptr == '$')
            *sptr++ = toupper(*cptr);
         else
            *sptr++ = '_';
      }
      *sptr = '\0';

      strcpy (FileName, VmsMailUserFullDirectory);
      strcat (FileName, ListName);
      strcat (FileName, ".DIS");
   }

   status = ReadFileIntoMemory (FileName, &ListPtr, &ListLength);
   NewListPtr = "";
   if (VMSnok (status))
   {
      if (status == SS$_NOSUCHFILE || status == RMS$_FNF)
      {
         if (UserList)
         {
            NewListPtr =
"&nbsp;&nbsp;&nbsp;<FONT COLOR=\"#ff0000\">%s</FONT>\n";
            ListPtr = "";
            ListLength = 0;
         }
         else
         {
            CgiLibResponseError (FI_LI, 0,
"%s\n\
<P><FORM>\n\
<INPUT TYPE=button VALUE=\"%s\" onClick=\"history.go(-1)\">&nbsp;\n\
<INPUT TYPE=button VALUE=\"%s\" onClick=\"window.close()\">\n\
</FORM>",
               lang_ErrListNotFound, lang_BtnGoBack, lang_BtnClose);
            return;
         }
      }
      else
      {
         CgiLibResponseError (FI_LI, status,
"%s\n\
<P><FORM>\n\
<INPUT TYPE=button VALUE=\"%s\" onClick=\"history.go(-1)\">&nbsp;\n\
<INPUT TYPE=button VALUE=\"%s\" onClick=\"window.close()\">\n\
</FORM>",
         FileName, lang_BtnGoBack, lang_BtnClose);
         return;
      }
   }

   EscapedListPtr = (char*)CgiLibHtmlEscape (ListPtr, -1, NULL, -1);
   free (ListPtr);

   CgiLibResponseHeader (200, "text/html");
   fprintf (stdout,
"<HTML>\n\
<HEAD>\n\
<META NAME=\"generator\" CONTENT=\"%s\">\n\
<META NAME=\"environment\" CONTENT=\"%s\">\n\
<META NAME=\"language\" CONTENT=\"%s %s\">\n\
<TITLE>yahMAIL ... %s: %s / %s</TITLE>\n\
</HEAD>\n\
<BODY %s>\n\
%s\
<B><FONT SIZE=+1><U>%s: &nbsp;%s / %s</FONT></U>%s</B>\n\
<BR>&nbsp;<FONT SIZE=-1>yahMAIL<SUP>*</SUP></FONT>\n\
<FORM ACTION=\"%s\" METHOD=POST>\n\
<INPUT TYPE=hidden NAME=lse VALUE=\"%s\">\n\
<TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>\n\
<TR><TD>\n\
<TEXTAREA NAME=txt ROWS=16 COLS=48 WRAP=off>\n\
%s\
</TEXTAREA>\n\
</TD></TR>\n\
</TABLE>\n",
      SoftwareID, CgiEnvironmentPtr, lang_Language, lang_LanguageAuthor,
      lang_ListAddressSingular, UserList ? MailUserName : "yahMAIL",
                                UserList ? ListName : ListNamePtr+1,
      ConfigBodyTagPtr, PostMasterWarningPtr,
      lang_ListAddressSingular, UserList ? MailUserName : "yahMAIL",
                                UserList ? ListName : ListNamePtr+1,
      NewListPtr, ActionPathInfo, ListName,
      EscapedListPtr);

   if (UserList)
      fprintf (stdout,
"<P><INPUT TYPE=hidden NAME=wha VALUE=\"lis\">\n\
<INPUT TYPE=submit NAME=act VALUE=\"%s\">&nbsp;\n\
<INPUT TYPE=reset VALUE=\"%s\">\
<I> ... %s</I>&nbsp;\n\
<INPUT TYPE=submit NAME=act VALUE=\"%s\">\n",
         lang_BtnUpdate, lang_BtnReset, lang_ResetCaution, lang_BtnRemove);

   fprintf (stdout,
"<P><INPUT TYPE=button VALUE=\"%s\" onClick=\"history.go(-1)\">&nbsp;\n\
<INPUT TYPE=button VALUE=\"%s\" onClick=\"window.close()\">\n\
</FORM>\n\
</BODY>\n\
</HTML>\n",
         lang_BtnGoBack, lang_BtnClose);

   free (EscapedListPtr);
}

/*****************************************************************************/
/*
Write the contents of the address list sent from AddrListAccess() into the
appropriate file in the user's mail directory.
*/

AddrListUpdate ()
       
{
   register char  *cptr, *sptr;

   int  status,
        PostBufferCount;
   char  FileName [256],
         ListName [256];
   char  *ListNamePtr,
         *ListTextPtr,
         *PostBufferPtr;
   FILE  *FilePtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AddrListUpdate()\n");

   CgiLibReadRequestBody (&PostBufferPtr, &PostBufferCount);
   CgiLibFormRequestBody (PostBufferPtr, PostBufferCount);
   free (PostBufferPtr);
   if (Debug)
      while ((cptr = CgiLibVarNull ("*")) != NULL)
         fprintf (stdout, "|%s|\n", cptr);

   ListNamePtr = CgiLibVar ("WWW_FORM_LTE");
   if (!ListNamePtr[0])
      ListNamePtr = CgiLibVar ("WWW_FORM_LSE");

   sptr = ListName;
   for (cptr = ListNamePtr; *cptr; cptr++)
   {
      if (isalnum(*cptr) || *cptr == '_' || *cptr == '_' || *cptr == '$')
         *sptr++ = toupper(*cptr);
      else
         *sptr++ = '_';
   }
   *sptr = '\0';

   strcpy (FileName, VmsMailUserFullDirectory);
   strcat (FileName, ListNamePtr);
   strcat (FileName, ".DIS");

   ListTextPtr = CgiLibVar ("WWW_FORM_TXT");

   FilePtr = fopen (FileName, "w", "shr=nil");
   if (FilePtr == NULL)
   {
      status = vaxc$errno;
      if (Debug) fprintf (stdout, "fopen() %%X%08.08X\n", status);
      return;
   }

   if (fputs (ListTextPtr, FilePtr) == EOF)
   {
      status = vaxc$errno;
      CgiLibResponseError (FI_LI, status,
"%s\n\
<P><FORM>\n\
<INPUT TYPE=button VALUE=\"%s\" onClick=\"history.go(-1)\">&nbsp;\n\
<INPUT TYPE=button VALUE=\"%s\" onClick=\"window.close()\">\n\
</FORM>",
      FileName, lang_BtnGoBack, lang_BtnClose);
   }
   else
   {
      /* purge the versions back to a maximum of 3 */
      strcat (FileName, ";-3");
      while (!remove (FileName));

      CgiLibResponseSuccess (
"%s\n\
<P><FORM>\n\
<INPUT TYPE=button VALUE=\"%s\" onClick=\"history.go(-1)\">&nbsp;\n\
<INPUT TYPE=button VALUE=\"%s\" onClick=\"window.close()\">\n\
</FORM>",
         lang_ListUpdated, lang_BtnGoBack, lang_BtnClose);
   }

   fclose (FilePtr);
}

/*****************************************************************************/
/*
Delete the address list file from the user's mail directory.
*/

AddrListRemove ()
       
{
   register char  *cptr, *sptr;

   int  status;
   char  FileName [256],
         ListName [256];
   char  *ListNamePtr,
         *ListTextPtr;
   FILE  *FilePtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AddrListRemove()\n");

   ListNamePtr = CgiLibVar ("WWW_FORM_LTE");
   if (!ListNamePtr[0])
      ListNamePtr = CgiLibVar ("WWW_FORM_LSE");

   sptr = ListName;
   for (cptr = ListNamePtr; *cptr; cptr++)
   {
      if (isalnum(*cptr) || *cptr == '_' || *cptr == '_' || *cptr == '$')
         *sptr++ = toupper(*cptr);
      else
         *sptr++ = '_';
   }
   *sptr = '\0';

   strcpy (FileName, VmsMailUserFullDirectory);
   strcat (FileName, ListNamePtr);
   strcat (FileName, ".DIS;*");

   if (remove (FileName))
   {
      status = vaxc$errno;
      CgiLibResponseError (FI_LI, status,
"%s\n\
<P><FORM>\n\
<INPUT TYPE=button VALUE=\"%s\" onClick=\"history.go(-1)\">&nbsp;\n\
<INPUT TYPE=button VALUE=\"%s\" onClick=\"window.close()\">\n\
</FORM>",
      FileName, lang_BtnGoBack, lang_BtnClose);
   }
   else
   {
      CgiLibResponseSuccess (
"%s\n\
<P><FORM>\n\
<INPUT TYPE=button VALUE=\"%s\" onClick=\"history.go(-1)\">&nbsp;\n\
<INPUT TYPE=button VALUE=\"%s\" onClick=\"window.close()\">\n\
</FORM>",
         lang_ListRemoved, lang_BtnGoBack, lang_BtnClose);
   }
}

/****************************************************************************/
/*
Read the file contents specified by 'FileName' into memory, set the pointer
at 'FileTextPtr' to the contents and the file size at 'FileSizePtr'.  Returns a
VMS status value that should be checked.
*/ 

int ReadFileIntoMemory
(
char *FileName,
char **FileTextPtr,
int *FileSizePtr
)
{
   int  status,
        Bytes,
        BytesRemaining,
        BufferCount,
        Length;
   char  *BufferPtr,
         *LinePtr;
   FILE  *FilePtr;
   stat_t  FstatBuffer;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ReadFileIntoMemory() |%s|\n", FileName);

   if (FileTextPtr != NULL) *FileTextPtr = NULL;
   if (FileSizePtr != NULL) *FileSizePtr = 0;

   FilePtr = fopen (FileName, "r", "shr=get");
   if (FilePtr == NULL)
   {
      status = vaxc$errno;
      if (Debug) fprintf (stdout, "fopen() %%X%08.08X\n", status);
      return (status);
   }

   if (fstat (fileno(FilePtr), &FstatBuffer) < 0)
   {
      status = vaxc$errno;
      if (Debug) fprintf (stdout, "fstat() %%X%08.08X\n", status);
      fclose (FilePtr);
      return (status);
   }

   Bytes = FstatBuffer.st_size;
   if (Debug) fprintf (stdout, "%d bytes\n", Bytes);
   /* a little margin for error ;^) */
   Bytes += 32;

   BufferPtr = calloc (Bytes, 1);
   if (BufferPtr == NULL)
   {
      CgiLibResponseError (FI_LI, vaxc$errno, "calloc()");
      exit (SS$_NORMAL);
   }
   BytesRemaining = Bytes;
   LinePtr = BufferPtr;

   BufferCount = 0;
   while (fgets (LinePtr, BytesRemaining, FilePtr) != NULL)
   {
      /** if (Debug) fprintf (stdout, "|%s|\n", LinePtr); **/
      Length = strlen(LinePtr);
      LinePtr += Length;
      BufferCount += Length;
      BytesRemaining -= Length;
   }
   fclose (FilePtr);

   if (Debug) fprintf (stdout, "%d |%s|\n", BufferCount, BufferPtr);

   if (FileTextPtr != NULL) *FileTextPtr = BufferPtr;
   if (FileSizePtr != NULL) *FileSizePtr = BufferCount;

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Turn on/off SYSPRV (and optionally, if installed with it) READALL.
*/

EnableAccess (int OnOff)

{
/* extrapolated from bit fields in <prvdef.h> */
#define PRV$M_READALL 0x8

   static unsigned long  ReadAllMask [2] = { 0, PRV$M_READALL };
   static unsigned long  SysPrvMask [2] = { PRV$M_SYSPRV, 0 };
   static boolean  ReadAll = true;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "EnableAccess() %d\n", OnOff);

   if (OnOff)
   {
      status = sys$setprv (1, &SysPrvMask, 0, 0);
      if (Debug) fprintf (stdout, "sys$setprv() %%X%08.08X\n", status);
      if (VMSnok (status) || status == SS$_NOTALLPRIV) exit (status & ~1);
      if (ReadAll)
      {
         status = sys$setprv (1, &ReadAllMask, 0, 0);
         if (Debug) fprintf (stdout, "sys$setprv() %%X%08.08X\n", status);
         if (status == SS$_NOTALLPRIV) ReadAll = false;
      }
   }
   else
   {
      if (VMSnok (status = sys$setprv (0, &SysPrvMask, 0, 0))) exit (status);
      if (Debug) fprintf (stdout, "sys$setprv() %%X%08.08X\n", status);
      if (ReadAll)
      {
         if (VMSnok (status = sys$setprv (0, &ReadAllMask, 0, 0)))
            exit (status);
         if (Debug) fprintf (stdout, "sys$setprv() %%X%08.08X\n", status);
      }
   }
}

/****************************************************************************/
/*
Copies null-terminated string 'SourcePtr' to 'DestinationPtr'.
Returns length of copied string.
Error exit on overflow!
*/ 

boolean strzcpy
(
char *DestinationPtr,
char *SourcePtr,
int SizeOfDestination
)
{
   register char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   zptr = (sptr = DestinationPtr) + SizeOfDestination;
   for (cptr = SourcePtr; *cptr && sptr < zptr; *sptr++ = *cptr++);
   if (sptr >= zptr) exit (SS$_RESULTOVF);
   *sptr = '\0';
   return (sptr - DestinationPtr);
}

/****************************************************************************/
/*
Set an integer reflecting the major and minor version of VMS (e.g. 60, 61, 62,
70, 71, 72, etc.)
*/ 

int GetVmsVersion ()

{
   static char  SyiVersion [16];

   static struct {
      short int  buf_len;
      short int  item;
      unsigned char   *buf_addr;
      unsigned short  *ret_len;
   }
   SyiItems [] =
   {
      { 8, SYI$_VERSION, &SyiVersion, 0 },
      { 0,0,0,0 }
   };

   int  status,
        version;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "GetVmsVersion()\n");

   if (VMSnok (status = sys$getsyiw (0, 0, 0, &SyiItems, 0, 0, 0)))
      exit (status);
   SyiVersion[8] = '\0';
   version = ((SyiVersion[1]-48) * 10) + (SyiVersion[3]-48);
   if (Debug) fprintf (stdout, "|%s| %d\n", SyiVersion, version);
   return (version);
}

/****************************************************************************/
/*
Return true is the string is all digits, false if not.
*/ 

boolean isinteger (register char *sptr)

{
   while (*sptr && isdigit(*sptr)) sptr++;
   if (!*sptr) return (true);
   return (false);
}

/****************************************************************************/
/*
Does a case-insensitive, character-by-character string compare and returns 
true if two strings are the same, or false if not.  If a maximum number of 
characters are specified only those will be compared, if the entire strings 
should be compared then specify the number of characters as 0.
*/ 

boolean strsame
(
register char *sptr1,
register char *sptr2,
register int  count
)
{
   while (*sptr1 && *sptr2)
   {
      if (toupper (*sptr1++) != toupper (*sptr2++)) return (false);
      if (count)
         if (!--count) return (true);
   }
   if (*sptr1 || *sptr2)
      return (false);
   else
      return (true);
}

/*****************************************************************************/

